//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_HAL_HH
#define BMP_HAL_HH

#include <glibmm.h>

#include <config.h>
#include <stdio.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <libhal.h>
#include <libhal-storage.h>

namespace Bmp
{
  namespace Library
  {
    class HAL
    {
      public:

#include "exception.hh"

        EXCEPTION(NotInitializedError)
        EXCEPTION(NoVolumeForUriError)
        EXCEPTION(NoMountPathForUdiError)

        struct Volume
        {
            std::string	    volume_udi;
            std::string	    device_udi;
            std::string	    mount_path;
            std::string	    device_serial;
            std::string	    volume_name;
            std::string	    device_file;
            bool		        is_disc;
            LibHalDriveBus  device_bus;
        };
        typedef std::map < std::string, Volume > Volumes;

        struct Device
        {
            std::string     device_udi;
            std::string     device_name;
            std::string	    device_file;
            bool		        is_disc;
        };
        typedef std::vector   < Device >                        Devices;

        typedef sigc::signal  <void, Volume >                   SignalVolume;
        typedef sigc::signal  <void, std::string, std::string > SignalCDDAInserted;
        typedef sigc::signal  <void, std::string >              SignalBlockDeviceRemoved;

        HAL ();
        ~HAL ();

        Volume
        get_volume_for_uri (Glib::ustring const& uri);

        std::string
        get_mount_path_for_volume_udi (std::string const& volume_udi);

        Devices
        get_optical_devices () const
        {
          return optical;
        }
  
        const Volumes&
        get_volumes () const
        {
          return volumes;
        }

        SignalVolume&
        signal_volume_removed ();

        SignalVolume&
        signal_volume_added   ();

        SignalBlockDeviceRemoved&
        signal_device_removed ();

        SignalCDDAInserted&
        signal_cdda_inserted ();
        
        bool
        volume_is_present (std::string const& udi)
        {
          if (!initialized)
            {
              return true; 
            }
          return (volumes_by_udi.find (udi) != volumes_by_udi.end());
        }

        bool
        is_initialized () const
        {
          return initialized;
        }

      private:

        SignalVolume signal_volume_removed_;
        SignalVolume signal_volume_added_;
        SignalBlockDeviceRemoved signal_device_removed_;
        SignalCDDAInserted signal_cdda_inserted_;

        DBusConnection     *dbus_connection;
        LibHalContext      *ctx;

        Volumes		volumes;
        Volumes		volumes_by_udi;
        Devices		optical;

        bool initialized;
        int  message_domain_id;

        dbus_bool_t hal_mainloop_integration (LibHalContext *ctx, DBusError *error);
        bool hal_init ();
        void cdrom_policy (const char *udi);
        void media_changed (const char *udi, const char *storage_device);
        void block_device_added (const char *udi, const char *capability __attribute__((__unused__)));
        void scan_devices (const char *udi);

        // C Callbacks
        static void device_added (LibHalContext *ctx, const char *udi);
        static void device_removed (LibHalContext *ctx, const char *udi);
        static void property_modified (LibHalContext *ctx, const char *udi, const char *key,
                              dbus_bool_t is_removed, dbus_bool_t is_added);
    };
  }
}

#endif //!BMP_HAL_HH
