#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <boost/format.hpp>

#include "util-file.hh"
#include "util.hh"
#include "paths.hh"
#include "main.hh"
#include "mbxml.hh"

#include "x_library.hh"
#include "x_vfs.hh"

#include "library-ui-modify-import.hh"

namespace
{
  void
  cbe_set_sensitive (Gtk::CheckButton *cb, Gtk::ComboBoxEntry *cbe)
  {
      cbe->set_sensitive (cb->get_active());
  }

  void
  fcb_set_sensitive (Gtk::CheckButton *cb, Gtk::FileChooserButton *fcb)
  {
      fcb->set_sensitive (cb->get_active());
  }
}

namespace Bmp
{
      LibraryUIModifyImport::LibraryUIModifyImport (BaseObjectType                        *cobject,
			                                              const Glib::RefPtr<Gnome::Glade::Xml> &xml) : Gtk::Dialog (cobject), ref_xml (xml),
                                                                                                  disc (Gdk::Pixbuf::create_from_file
                                                                                                  (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "audio.png"))
      {
          ref_xml->get_widget ("merge_cb_artist",   merge_cb_artist);
          ref_xml->get_widget ("merge_cb_album",    merge_cb_album);
          ref_xml->get_widget ("merge_coverart",    merge_coverart);
          ref_xml->get_widget ("merge_album",	      merge_album);
          ref_xml->get_widget ("merge_artist",	    merge_artist);
          ref_xml->get_widget ("merge_artist_id",   merge_artist_id);
          ref_xml->get_widget ("merge_asin",	      merge_asin);
          ref_xml->get_widget ("merge_release_id",  merge_release_id);
          ref_xml->get_widget ("merge_score",	      merge_score);
          ref_xml->get_widget ("b_refresh",	        b_refresh);
          ref_xml->get_widget ("l_tracks",	        l_tracks);
          ref_xml->get_widget ("ok",	              ok);
          ref_xml->get_widget ("cancel",	          cancel);

          merge_cb_artist_entry = merge_cb_artist->get_entry ();
          merge_cb_album_entry  = merge_cb_album->get_entry ();

          merge_cb_artist_entry->set_activates_default (true);
          merge_cb_album_entry->set_activates_default (true);

          merge_cb_artist_entry->set_text (_("<select here>"));
          merge_cb_album_entry->set_text ("");
          merge_coverart->set (disc);
          ok->set_sensitive (false);

          b_refresh->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyImport::on_button_update));
      }

      LibraryUIModifyImport*
      LibraryUIModifyImport::create ()
      {
        const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "library-ui-modify-albums.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        LibraryUIModifyImport *p = 0;
        glade_xml->get_widget_derived ("dialog_modify_albums", p);
        return p;
      }

      void
      LibraryUIModifyImport::populate_store (MusicBrainzXML::MBReleaseV& releases)
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          merge_cb_artist_store = ListStore::create (artists);
          merge_cb_album_store  = ListStore::create (albums);
          merge_artist_release_map.clear ();
          merge_coverart->set (disc);
          merge_artist_id->set_text   ("");
          merge_release_id->set_text  ("");
          merge_asin->set_text        ("");
          merge_score->set_text       ("");
          l_tracks->set_text          ("");

          for (MBReleaseV::const_iterator r_iter  = releases.begin (),
                                          r_end   = releases.end ();
                                          r_iter != r_end; ++r_iter)
          {
            MBRelease const& release = (*r_iter);
            merge_releases.insert (make_pair (release.release_id, release));
            ArtistReleaseMap::iterator map_iter = merge_artist_release_map.find (release.artist_id);

            if (map_iter != merge_artist_release_map.end ())
            {
              ReleaseIds &ids = map_iter->second;
              ids.push_back (release.release_id);
            }
            else
            {
              TreeModel::iterator m_iter  = merge_cb_artist_store->append ();
              (*m_iter)[artists.name]	    = release.artist_name;
              (*m_iter)[artists.id]	      = release.artist_id;

              ReleaseIds ids; 
              ids.push_back (release.release_id);
              merge_artist_release_map.insert (make_pair (release.artist_id, ids)); 
            }
          }

          merge_cb_artist->set_model  (merge_cb_artist_store);
          merge_cb_album->set_model   (merge_cb_album_store);

          merge_cb_artist_entry->set_text (_("<select here>"));
          merge_cb_album_entry->set_text  ("");
          ok->set_sensitive (false);
      }

      void
      LibraryUIModifyImport::on_button_update ()
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          set_sensitive (false);
          MBReleaseV releases; 
          mb_query_releases (merge_cb_artist_entry->get_text (), merge_cb_album_entry->get_text (), releases);
          populate_store (releases);
          set_sensitive (true);
      }

      int
      LibraryUIModifyImport::run (const MusicBrainzXML::MBReleaseV& releases)
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;

          merge_cb_artist_store = ListStore::create (artists);
          merge_cb_album_store  = ListStore::create (albums);

          for (MBReleaseV::const_iterator r_iter  = releases.begin(),
                                          r_end   = releases.end();
                                          r_iter != r_end; ++r_iter)
          {

            MBRelease const& release = (*r_iter);
            merge_releases.insert (std::make_pair (release.release_id, release));
        
            ArtistReleaseMap::iterator map_iter = merge_artist_release_map.find (release.artist_id);
            if (map_iter != merge_artist_release_map.end ())
              {
                ReleaseIds &ids = map_iter->second;
                ids.push_back (release.release_id);
              }
            else
              {
                TreeModel::iterator m_iter  = merge_cb_artist_store->append ();

                (*m_iter)[artists.name]	= release.artist_name;
                (*m_iter)[artists.id]	  = release.artist_id;
              
                ReleaseIds ids; 
                ids.push_back (release.release_id);
                merge_artist_release_map.insert (std::make_pair (release.artist_id, ids)); 
              }
          }

          // Setup 'Artist' ComboBox 
          merge_cb_artist->set_model (merge_cb_artist_store);
          merge_cb_artist->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyImport::artist_changed));
          merge_cb_artist->set_text_column (0);

          // Setup 'Album' ComboBox 
          merge_cb_album->set_model (merge_cb_album_store);
          merge_cb_album->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyImport::album_changed));
          merge_cb_album->set_text_column (0);

          return Gtk::Dialog::run ();
      }

      int
      LibraryUIModifyImport::rerun ()
      {
	      return Gtk::Dialog::run ();
      }

      void
      LibraryUIModifyImport::album_changed ()
      {
          using namespace Gtk;
          using namespace MusicBrainzXML;

          if (merge_cb_album->get_active_row_number () < 0) return;

          set_sensitive (false);
        
          // Get releases from selected artist
          TreeModel::const_iterator m_iter  = merge_cb_album->get_active ();
          ReleaseMap::iterator r_iter       = merge_releases.find ((*m_iter)[albums.id]);
          MBRelease const& release          = r_iter->second;

          try {
              if (release.asin)
                {
                  Glib::RefPtr<Gdk::Pixbuf> cover;
                  Amazon::get_cover_by_asin (release.asin.get(), cover);
                  merge_coverart->set (cover);
                }
              else
                  throw 0;
            }
          catch (Bmp::Amazon::Exception& cxe)
            {
              g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Trackinfo/Amazon: %s", cxe.what());
              merge_coverart->set (disc);
            }
          catch (...)
            { 
              merge_coverart->set (disc);
            }

          merge_album->set_text (release.release_title);
          merge_artist->set_text (release.artist_name);
          merge_artist_id->set_text (release.artist_id);
          merge_release_id->set_text (release.release_id);

          if (release.asin)
            {
              merge_asin->set_text  (release.asin.get());
            }
    
          if (release.score)
            {
              merge_score->set_text ((boost::format ("%d") % release.score.get()).str());
            }
    
          if (release.track_list)
            {
              l_tracks->set_text ((boost::format ("%d") % release.track_list.get()).str());
            }

          selected = release;
          ok->set_sensitive (true);

          set_sensitive (true);
      }

      void
      LibraryUIModifyImport::artist_changed ()
      {
          using namespace Gtk;
          using namespace MusicBrainzXML;

          if (merge_cb_artist->get_active_row_number () < 0) return;

          set_sensitive (false);
        
          // Get releases from selected artist
          TreeModel::const_iterator  m_iter   = merge_cb_artist->get_active ();
          ArtistReleaseMap::iterator map_iter = merge_artist_release_map.find ((*m_iter)[artists.id]);
          ReleaseIds                 &ids     = map_iter->second;

          merge_cb_album_store->clear ();

          for (ReleaseIds::const_iterator r_iter    = ids.begin (),
                                          r_iterEnd = ids.end ();
                                          r_iter != r_iterEnd; ++r_iter)
            {
              // Iter from releases map
              ReleaseMap::iterator map_iter = merge_releases.find (*r_iter);

              // Get the release based on id 
              MBRelease const& release = map_iter->second; 

              TreeModel::iterator c_Iter  = merge_cb_album_store->append ();
              (*c_Iter)[albums.name]      = release.release_title;
              (*c_Iter)[albums.id]        = release.release_id;
            }

          ok->set_sensitive (false);
          set_sensitive (true);
      }
}
