//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <gtk/gtk.h>
#include <gtkmm.h>
#include <glibmm/i18n.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <boost/format.hpp>
#include <boost/optional.hpp>

#include <ne_basic.h>
#include <ne_compress.h>
#include <ne_socket.h>
#include <ne_utils.h>
#include <ne_props.h>
#include <ne_session.h>
#include <ne_request.h>
#include <ne_uri.h>
#include <ne_auth.h>

#include "main.hh"
#include "debug.hh"
#include "paths.hh"
#include "stock.hh"
#include "audio.hh"
#include "util-file.hh"
#include "util.hh"
#include "ui-toolbox.hh"

#include "x_library.hh"
#include "x_vfs.hh"

#include "mbxml.hh"

#include "library-ui-modify.hh"

namespace
{
  void
  cbe_set_sensitive (Gtk::CheckButton *cb, Gtk::ComboBoxEntry *cbe)
  {
      cbe->set_sensitive (cb->get_active());
  }

  void
  fcb_set_sensitive (Gtk::CheckButton *cb, Gtk::FileChooserButton *fcb)
  {
      fcb->set_sensitive (cb->get_active());
  }
  
  bool processed = false;

  void
  set_processed ()
  {
    processed = true;
  }

  void 
  set_progress_bar (int position, Gtk::ProgressBar *pb, Bmp::Audio::ProcessorBase *processor)
  {
    pb->property_fraction() = (1.*position) / 135.;
  }

  void
  set_cbox_sensitive (Gtk::CheckButton *cb, Gtk::ComboBoxEntry *cb_entry)
  {
    cb_entry->set_sensitive (cb->get_active());
  }

  Gtk::Widget*
  get_popup (Glib::RefPtr<Gtk::UIManager> ui_manager,
             Glib::ustring                path) 
  {
      Gtk::Widget *menu_item;
      menu_item = ui_manager->get_widget (path);
      if ( menu_item )
          return dynamic_cast<Gtk::MenuItem*>(menu_item)->get_submenu ();
      else
          return 0;
  }
}

namespace Bmp
{
      LibraryUIModify::LibraryUIModify (BaseObjectType                        *cobject,
                                        const Glib::RefPtr<Gnome::Glade::Xml> &xml)
                    : Gtk::Dialog (cobject), m_ref_xml (xml),
                      disc (Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR, "audio.png")))
      {

          // Widgets
          m_ref_xml->get_widget ("merge_cb_artist", merge_cb_artist);
          m_ref_xml->get_widget ("merge_cb_album", merge_cb_album);

          m_ref_xml->get_widget ("cb_genre", cb_genre);
          m_ref_xml->get_widget ("cb_modify_genre", cb_modify_genre);
          cb_genre_entry = cb_genre->get_entry ();
          cb_modify_genre->signal_toggled().connect (sigc::bind (sigc::ptr_fun (&set_cbox_sensitive), cb_modify_genre, cb_genre));
                                                                               
          m_ref_xml->get_widget ("merge_coverart", merge_coverart);
          m_ref_xml->get_widget ("merge_artist_id", merge_artist_id);
          m_ref_xml->get_widget ("merge_release_id", merge_release_id);
          m_ref_xml->get_widget ("merge_asin", merge_asin);
          m_ref_xml->get_widget ("merge_score",	merge_score);
          m_ref_xml->get_widget ("view_release", view_release);
          m_ref_xml->get_widget ("view_local", view_local);
          m_ref_xml->get_widget ("cancel", cancel);
          m_ref_xml->get_widget ("ok", ok);
          m_ref_xml->get_widget ("progress_processing",	progress_processing);
          m_ref_xml->get_widget ("l_tracks", l_tracks);

          m_ui_manager = Gtk::UIManager::create ();
          m_actions = Gtk::ActionGroup::create ("ActionsLibraryModify"); 

#ifdef ENABLE_NLS
          gtk_action_group_set_translate_func (m_actions->gobj(), Util::menu_translate, NULL, NULL); 
#endif

#define ACTION_UPDATE     "action-update"
#define ACTION_MOVE_UP    "action-move-up"
#define ACTION_MOVE_DOWN  "action-move-down"

          m_actions->add ( Gtk::Action::create ("dummy", "dummy"));

          m_actions->add ( Gtk::Action::create (ACTION_UPDATE, Gtk::StockID (GTK_STOCK_NETWORK), ("Search")),
                                   sigc::mem_fun (this, &Bmp::LibraryUIModify::on_action_update));

          m_actions->add ( Gtk::Action::create (ACTION_MOVE_UP, Gtk::StockID (GTK_STOCK_GO_UP), ("Move Up")),
                                   sigc::mem_fun (this, &Bmp::LibraryUIModify::on_action_up));

          m_actions->add ( Gtk::Action::create (ACTION_MOVE_DOWN, Gtk::StockID (GTK_STOCK_GO_DOWN), ("Move Down")),
                                   sigc::mem_fun (this, &Bmp::LibraryUIModify::on_action_down));

          m_actions->get_action (ACTION_UPDATE)->connect_proxy
                (*(m_ref_xml->get_widget ("b_update")));

          m_actions->get_action (ACTION_MOVE_UP)->connect_proxy
                (*(m_ref_xml->get_widget ("b_move_up")));

          m_actions->get_action (ACTION_MOVE_DOWN)->connect_proxy
                (*(m_ref_xml->get_widget ("b_move_down")));

          m_actions->get_action (ACTION_MOVE_UP)->set_sensitive (false);
          m_actions->get_action (ACTION_MOVE_DOWN)->set_sensitive (false);

#ifdef HAVE_OFA

          n_pending_puids = 0;

#define ACTION_SCAN_RELEASE "action-scan-release"
#define ACTION_SCAN_FILE    "action-scan-file"
#define ACTION_SCAN_ALL     "action-scan-all"
#define ACTION_SUBMIT_PUIDS "action-submit-puids"

          m_actions->add ( Gtk::Action::create (ACTION_SCAN_RELEASE, Gtk::StockID (GTK_STOCK_FIND), ("Find Release based on Selected File")),
                                  Gtk::AccelKey (""), 
                                  sigc::mem_fun (this, &Bmp::LibraryUIModify::on_action_scan));

          m_actions->add ( Gtk::Action::create (ACTION_SCAN_FILE, Gtk::StockID (BMP_STOCK_SCAN), ("Find PUID for selected File")),
                                  Gtk::AccelKey (""), 
                                  sigc::mem_fun (this, &Bmp::LibraryUIModify::on_action_scan_file));

          m_actions->add ( Gtk::Action::create (ACTION_SCAN_ALL, Gtk::StockID (GTK_STOCK_FIND), ("Find PUIDs for all Files")),
                                  Gtk::AccelKey (""), 
                                  sigc::mem_fun (this, &Bmp::LibraryUIModify::on_action_scan_all));

          m_actions->add ( Gtk::Action::create (ACTION_SUBMIT_PUIDS, Gtk::StockID (GTK_STOCK_NETWORK), ("Submit found PUID <-> Track ID Pairs")),
                                  Gtk::AccelKey (""), 
                                  sigc::mem_fun (this, &Bmp::LibraryUIModify::on_action_submit_puids));

          m_actions->get_action (ACTION_SCAN_RELEASE)->set_sensitive (false);
          m_actions->get_action (ACTION_SCAN_FILE)->set_sensitive (false);
          m_actions->get_action (ACTION_SCAN_ALL)->set_sensitive (false);
#endif
      
          // Insert ActionGroup
          m_ui_manager->insert_action_group (m_actions);

#ifdef HAVE_OFA
          m_ui_manager->add_ui_from_file (Glib::build_filename (BMP_UI_DIR, "modify.ui"));
          view_local->signal_event().connect
                  (sigc::mem_fun (this, &Bmp::LibraryUIModify::popup_rhs_menu));
#endif

          merge_cb_artist_entry = merge_cb_artist->get_entry ();
          merge_cb_album_entry  = merge_cb_album->get_entry ();
          merge_cb_artist_entry->set_activates_default (true);
          merge_cb_album_entry->set_activates_default (true);
          merge_cb_artist_entry->set_text (_("<select here>"));
          merge_cb_album_entry->set_text ("");
          merge_coverart->set (disc);

          struct { 
              char    *title;
              double   xalign;
              int	     width;
          } headers_tracks[] = {
              { _("#"),	    0.5, 0,  },
              { _("Title"), 0.0, 100 },
          };

          struct { 
              char    *title;
              double   xalign;
              int	     width;
          } headers_tracks_local[] = {
              { _("#"),	    0.5, 0,  },
              { "",         0.5, 24, },
              { _("Title"), 0.0, 100 },
          };


          // Release
          for (unsigned int n = 0; n < G_N_ELEMENTS (headers_tracks); ++n)
            {
              Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
              cell->property_xalign() = headers_tracks[n].xalign;
              view_release->append_column (headers_tracks[n].title, *cell);
              Gtk::TreeViewColumn *column = view_release->get_column (n);
              column->add_attribute (*cell, "text", n);
              column->set_resizable (true);
            }

          store_release = Gtk::ListStore::create (tracks_cr);
          view_release->set_model (store_release);
          view_release->get_selection()->set_mode (Gtk::SELECTION_NONE);

          // Local
          {
            Gtk::CellRendererToggle *cell = Gtk::manage ( new Gtk::CellRendererToggle() );
            view_local->append_column ("", *cell);
            Gtk::TreeViewColumn *column = view_local->get_column (0);
            column->add_attribute (*cell, "active", 0);
            column->set_resizable (false);
            cell->signal_toggled().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::on_mod_toggled));
          }

          for (unsigned int n = 0; n < G_N_ELEMENTS (headers_tracks_local); ++n)
          {
            if (n == 0)
              {
                Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
                cell->property_xalign() = headers_tracks_local[n].xalign;
                view_local->append_column (headers_tracks_local[n].title, *cell);
                Gtk::TreeViewColumn *column = view_local->get_column (n+1);
                column->set_cell_data_func (*cell, sigc::mem_fun (*this, &Bmp::LibraryUIModify::cell_data_func_number));
              }
#ifdef HAVE_OFA
            else
            if (n == 1)
              {
                Gtk::CellRendererPixbuf *cell = Gtk::manage ( new Gtk::CellRendererPixbuf() );
                view_local->append_column ("", *cell);
                Gtk::TreeViewColumn *column = view_local->get_column (n+1);
                column->set_cell_data_func (*cell, sigc::mem_fun (this, &Bmp::LibraryUIModify::cell_data_func_puid_state));
                column->set_expand (false);
              }
#endif
            else
              {
                Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
                cell->property_xalign() = headers_tracks_local[n].xalign;
                view_local->append_column (headers_tracks_local[n].title, *cell);
                Gtk::TreeViewColumn *column = view_local->get_column (n+1);
                column->add_attribute (*cell, "text", n+1);
              }
          }

          store_local = Gtk::ListStore::create (columns);
          view_local->set_model (store_local);
          view_local->get_selection()->set_mode (Gtk::SELECTION_SINGLE);
          view_local->get_selection()->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::local_selection_changed));
          ok->set_sensitive (false);
      }

      LibraryUIModify*
      LibraryUIModify::create ()
      {
        const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "library-ui-modify.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        LibraryUIModify *p = 0;
        glade_xml->get_widget_derived ("dialog_modify", p);
        return p;
      }

      void
      LibraryUIModify::cell_data_func_number (Gtk::CellRenderer* cell_, const Gtk::TreeModel::iterator& iter)
      {
        Gtk::CellRendererText *cell = reinterpret_cast<Gtk::CellRendererText*>(cell_);
        static boost::format fint ("%d");
        unsigned int nr = (*iter)[columns.tracknumber];
        if (nr == 0)
          cell->property_text() = "";
        else
          cell->property_text() = (fint % nr).str();
      }

      void
      LibraryUIModify::on_mod_toggled (const Glib::ustring& path)
      {
        Gtk::TreeModel::iterator m_iter = store_local->get_iter (path);
        (*m_iter)[columns.modify] = !(*m_iter)[columns.modify];
      }

     void
      LibraryUIModify::local_selection_changed ()
      {
        bool selected = view_local->get_selection()->count_selected_rows ();
        m_actions->get_action (ACTION_MOVE_UP)->set_sensitive (selected);
        m_actions->get_action (ACTION_MOVE_DOWN)->set_sensitive (selected);

#ifdef HAVE_OFA
        m_actions->get_action (ACTION_SCAN_RELEASE)->set_sensitive (selected);
        m_actions->get_action (ACTION_SCAN_FILE)->set_sensitive (selected);
#endif

      }

      void
      LibraryUIModify::on_action_up ()
      {
        Gtk::TreeModel::iterator m_iter = view_local->get_selection()->get_selected ();
        if (m_iter == store_local->children().begin()) return;
        Gtk::TreeModel::iterator d_Iter = m_iter;
        d_Iter--; // yes, this is slow, but it doesn't matter much as a release doesn't have that many tracks
        store_local->iter_swap (m_iter, d_Iter);
      }

      void
      LibraryUIModify::on_action_down ()
      {
        Gtk::TreeModel::iterator m_iter = view_local->get_selection()->get_selected ();
        if (m_iter == store_local->children().end()) return;
        Gtk::TreeModel::iterator d_Iter = m_iter;
        d_Iter++;
        store_local->iter_swap (m_iter, d_Iter);
      }

      void
      LibraryUIModify::populate_data (MusicBrainzXML::MBReleaseV& releases)
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          merge_cb_artist_store = ListStore::create (artists);
          merge_cb_album_store  = ListStore::create (albums);
          merge_artist_release_map.clear ();
          merge_coverart->set (disc);
          merge_artist_id->set_text   ("");
          merge_release_id->set_text  ("");
          merge_asin->set_text        ("");
          merge_score->set_text       ("");
          l_tracks->set_text          ("");
          store_release->clear ();
          store_local->clear ();
          merge_cb_artist_entry->set_text ("");
          merge_cb_album_entry->set_text ("");

          for (MBReleaseV::const_iterator r_iter  = releases.begin (),
                                          r_end   = releases.end ();
                                          r_iter != r_end; ++r_iter)
          {
            MBRelease const& release = (*r_iter);

            merge_releases.insert (make_pair (release.release_id, release));
            ArtistReleaseMap::iterator map_iter = merge_artist_release_map.find (release.artist_id);

            if (map_iter != merge_artist_release_map.end ())
            {
              ReleaseIds &ids = map_iter->second;
              ids.push_back (release.release_id);
            }
            else
            {
              TreeModel::iterator m_iter = merge_cb_artist_store->append ();
              (*m_iter)[artists.name]	= release.artist_name;
              (*m_iter)[artists.id]	= release.artist_id;

              ReleaseIds ids; 
              ids.push_back (release.release_id);
              merge_artist_release_map.insert (make_pair (release.artist_id, ids)); 
            }
          }

          merge_cb_artist->set_model  (merge_cb_artist_store);
          merge_cb_album->set_model   (merge_cb_album_store);

          merge_cb_artist->set_active (0);
          merge_cb_album->set_active (0);
          ok->set_sensitive (true);
      }

#ifdef HAVE_OFA
      bool
      LibraryUIModify::popup_rhs_menu (GdkEvent *ev)
      {
        using namespace Gtk;

        GdkEventButton *event = 0;
        if (ev->type == GDK_BUTTON_PRESS)
          {
            event = reinterpret_cast<GdkEventButton *>(ev);
            if ((event->button == 3) && (event->type == GDK_BUTTON_PRESS))
              {
                Menu *menu = dynamic_cast<Menu *>(get_popup (m_ui_manager, "/popup-modify/menu-modify"));
                menu->popup (event->button, event->time);
                return true;
              }

          } 
        return false;
      }

      void
      LibraryUIModify::cell_data_func_puid_state (Gtk::CellRenderer* cell, const Gtk::TreeModel::iterator& iter)
      {
        if (Glib::ustring((*iter)[columns.location]).empty())
          {
            cell->property_visible() = false;
            return;
          }
        else
          {
            cell->property_visible() = true;
          }

        PuidState state = (*iter)[columns.puid_state];
        Glib::RefPtr<Gdk::Pixbuf> icon;

        switch (state)
          {
            case P_NONE:
                icon = render_icon (Gtk::StockID (BMP_STOCK_ERROR), Gtk::ICON_SIZE_SMALL_TOOLBAR);
                break;

            case P_PUID:
                icon = render_icon (Gtk::StockID (BMP_STOCK_APPROVE_PARTIAL), Gtk::ICON_SIZE_SMALL_TOOLBAR);
                break;

            case P_PUID_METADATA:
                icon = render_icon (Gtk::StockID (BMP_STOCK_APPROVE), Gtk::ICON_SIZE_SMALL_TOOLBAR);
                break;
          }

        Gtk::CellRendererPixbuf *cell_pb = dynamic_cast<Gtk::CellRendererPixbuf *>(cell);
        cell_pb->property_pixbuf () = icon;
      }

      int
      LibraryUIModify::http_auth (gpointer data, const char *realm, 
                                int attempts, char *username, char *password)
      {
        strncpy (username, mcs->key_get<std::string>("musicbrainz", "username").c_str(), NE_ABUFSIZ);
        strncpy (password, mcs->key_get<std::string>("musicbrainz", "password").c_str(), NE_ABUFSIZ);
        return attempts;
      }

      void
      LibraryUIModify::set_local_puid_state (Gtk::TreeModel::iterator& m_iter, PuidState state)
      {
        using namespace Gtk;

        (*m_iter)[columns.puid_state] = state;
        m_actions->get_action (ACTION_SUBMIT_PUIDS)->set_sensitive (true);
      }

      void
      LibraryUIModify::on_action_submit_puids ()
      {
          using namespace Glib;
          using namespace Gtk;
          using namespace Bmp::DB;
          using namespace Bmp::Library;
          using namespace std;

          typedef std::vector<unsigned int> PositionV;

          ne_session *sess = 0;   
          ne_request *reqs = 0;

          TreeNodeChildren const& nodes_release = store_release->children();
          TreeNodeChildren const& nodes_local   = store_local->children();

          unsigned int max_nodes = nodes_release.size();
          if (nodes_local.size() < max_nodes) max_nodes = nodes_local.size();

          TreeNodeChildren::const_iterator iter_r = nodes_release.begin();
          TreeNodeChildren::const_iterator iter_l = nodes_local.begin();

          PositionV posv;
          for (unsigned int n = 0; n < max_nodes; ++n, ++iter_l)
          {
            if (PuidState((*iter_l)[columns.puid_state]) == P_PUID) 
            {
                posv.push_back (n);
            }
          }

          if (!posv.size()) return;

          sess = ne_session_create ("http", "musicbrainz.org", 80);
          if (!sess) return; 

          set_sensitive (false);
          ne_set_server_auth (sess, http_auth, NULL);
          ne_set_useragent (sess, PACKAGE"/"VERSION);

          progress_processing->set_fraction (0.0);
          progress_processing->set_text ("");

          double n_children = posv.size(); 
          double nth_child = 0;

          for (PositionV::iterator v  = posv.begin();
                                   v != posv.end(); ++v)
          {
              Gtk::TreePath path;

              path.append_index (*v);
              Gtk::TreeModel::iterator iter_l = store_local->get_iter (path);

              path = Gtk::TreePath();
              path.append_index (*v);
              Gtk::TreeModel::iterator iter_r = store_release->get_iter (path);

              MusicBrainzXML::MBTrack const& track = (*iter_r)[tracks_cr.track];

              ustring puid = (*iter_l)[columns.puid];
              ustring trackid = track.track_id;
  
              ustring uri ("http://musicbrainz.org/ws/1/track/");
              ustring data ("client=BMP-0.30&puid=");
                      data.append (trackid).append ("+").append(puid);

              debug ("puid-submit", "Data: %s", data.c_str());

              reqs = ne_request_create (sess, "POST", uri.c_str());
#ifdef NE_FEATURE_I18N                  
              ne_set_request_flag (reqs, NE_REQFLAG_IDEMPOTENT, 0);
#endif
              ne_add_request_header (reqs, "Content-type", "application/x-www-form-urlencoded");
              ne_set_request_body_buffer (reqs, data.c_str(), data.size());
              int result = ne_request_dispatch (reqs);
              int status = ne_get_status (reqs)->code;
              ne_request_destroy (reqs);

              std::string error_message = ne_get_error (sess);

              debug ("puid-submit", "NeonError: %s", error_message.c_str());
              debug ("puid-submit", "NeonState: %d", status); 
              debug ("puid-submit", "NeonResult: %d", result); 

              nth_child += 1;
              progress_processing->set_fraction (double((1.*nth_child)/(1.*n_children)));
              progress_processing->set_text ((boost::format ("%d / %d") % int(nth_child) % int(n_children)).str());

              while (gtk_events_pending()) gtk_main_iteration ();

              //FIXME: Deal with errors
              switch (result)
                {
                  case NE_OK:
                    break;

                  case NE_CONNECT:
                    continue;  

                  case NE_TIMEOUT:
                    continue;  

                  case NE_AUTH:
                    return;  
                
                  default: break;
                }

              switch (status)
                {
                  case 200: set_local_puid_state (iter_l, P_PUID_METADATA); break;
                  case 400: break;
                  case 401: break;
                  case 404: break;
                  default: break;
                }
          }

          progress_processing->set_fraction (0.0);
          progress_processing->set_text ("");
          set_sensitive (true);
          ne_session_destroy (sess);
      }

      void
      LibraryUIModify::on_action_scan_all ()
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          m_actions->get_action (ACTION_SCAN_RELEASE)->set_sensitive (false);
          m_actions->get_action (ACTION_SCAN_FILE)->set_sensitive (false);
          m_actions->get_action (ACTION_SCAN_ALL)->set_sensitive (false);
          m_actions->get_action (ACTION_UPDATE)->set_sensitive (false);

          merge_cb_artist->set_sensitive (false);
          merge_cb_album->set_sensitive (false);

          ok->set_sensitive (false);
          cancel->set_sensitive (false);

          set_sensitive (false);

          progress_processing->set_sensitive (true);
          progress_processing->property_fraction() = 0.; 

          view_local->get_selection()->unselect_all();

          for (TreeNodeChildren::iterator m_iter  = store_local->children().begin();
                                          m_iter != store_local->children().end(); ++m_iter)
          {
              if (PuidState((*m_iter)[columns.puid_state]) != P_NONE) continue;
              if (ustring((*m_iter)[columns.location]).empty()) continue;

              view_local->get_selection()->select(m_iter);

              Bmp::Audio::ProcessorPUID *processor = new Bmp::Audio::ProcessorPUID;
              processed = false;
              processor->signal_eos().connect (sigc::ptr_fun (&set_processed));
              processor->signal_position().connect (sigc::bind (sigc::ptr_fun (&set_progress_bar), progress_processing, processor));
              processor->set_uri (ustring((*m_iter)[columns.location]));
              processor->run ();

              while (!processed)
                {
                  while (gtk_events_pending())
                    {
                      gtk_main_iteration();
                    }
                }

              boost::optional<ustring> puid    = processor->get_puid ();
              boost::optional<ustring> artist  = processor->get_artist ();
              boost::optional<ustring> title   = processor->get_title ();

              delete processor;

              if (puid)
                {
                  set_local_puid_state (m_iter, P_PUID);
                  (*m_iter)[columns.puid] = puid.get();
                  MBTrackV tracks;
                  mb_tracks_by_puid (puid.get(), tracks);
                  for (unsigned int n = 0; n < tracks.size(); ++n)
                    {
                      MBTrack const& track = tracks[n];
                      if (track.release_id && track.release_id.get() == m_release.release_id)
                        {
                          set_local_puid_state (m_iter, P_PUID_METADATA);
                          (*m_iter)[columns.title]       = track.track_title; 
                          (*m_iter)[columns.tracknumber] = track.tracknumber; 
                          (*m_iter)[columns.track]       = track;
                          break;
                        }
                    }
                }
          }

          m_actions->get_action (ACTION_SCAN_RELEASE)->set_sensitive (true);
          m_actions->get_action (ACTION_SCAN_ALL)->set_sensitive (true);
          m_actions->get_action (ACTION_UPDATE)->set_sensitive (true);

          merge_cb_artist->set_sensitive (true);
          merge_cb_album->set_sensitive (true);

          ok->set_sensitive (true);
          cancel->set_sensitive (true);

          set_sensitive (true);

          progress_processing->set_sensitive (false);
          progress_processing->property_fraction() = 0.;
      }

      void
      LibraryUIModify::on_action_scan_file ()
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          if (view_local->get_selection()->count_selected_rows() < 1) return;

          m_actions->get_action (ACTION_SCAN_RELEASE)->set_sensitive (false);
          m_actions->get_action (ACTION_SCAN_FILE)->set_sensitive (false);
          m_actions->get_action (ACTION_SCAN_ALL)->set_sensitive (false);
          m_actions->get_action (ACTION_UPDATE)->set_sensitive (false);

          merge_cb_artist->set_sensitive (false);
          merge_cb_album->set_sensitive (false);

          ok->set_sensitive (false);
          cancel->set_sensitive (false);

          set_sensitive (false);

          progress_processing->set_sensitive (true);
          progress_processing->property_fraction() = 0.; 

          Gtk::TreeModel::iterator m_iter = view_local->get_selection()->get_selected ();
          Bmp::Audio::ProcessorPUID *processor = new Bmp::Audio::ProcessorPUID;

          processed = false;
          processor->signal_eos().connect (sigc::ptr_fun (&set_processed));
          processor->signal_position().connect (sigc::bind (sigc::ptr_fun (&set_progress_bar), progress_processing, processor));
          processor->set_uri (ustring((*m_iter)[columns.location]));
          processor->run ();
        
          while (!processed)
          {
            while (gtk_events_pending()) gtk_main_iteration();
          }

          boost::optional<ustring> puid    = processor->get_puid ();
          boost::optional<ustring> artist  = processor->get_artist ();
          boost::optional<ustring> title   = processor->get_title ();

          delete processor;

          if (puid)
            {
              set_local_puid_state (m_iter, P_PUID);
              MBTrackV tracks;
              mb_tracks_by_puid (puid.get(), tracks);
              for (unsigned int n = 0; n < tracks.size(); ++n)
                {
                  MBTrack const& track = tracks[n];
                  if (track.release_id && track.release_id.get() == m_release.release_id)
                    {
                      set_local_puid_state (m_iter, P_PUID_METADATA);
                      (*m_iter)[columns.puid]        = puid.get();
                      (*m_iter)[columns.title]       = track.track_title; 
                      (*m_iter)[columns.tracknumber] = track.tracknumber; 
                      (*m_iter)[columns.track]       = track;
                      break;
                    }
                }
            }

          m_actions->get_action (ACTION_SCAN_RELEASE)->set_sensitive (true);
          m_actions->get_action (ACTION_SCAN_ALL)->set_sensitive (true);
          m_actions->get_action (ACTION_UPDATE)->set_sensitive (true);

          merge_cb_artist->set_sensitive (true);
          merge_cb_album->set_sensitive (true);

          ok->set_sensitive (true);
          cancel->set_sensitive (true);

          set_sensitive (true);

          progress_processing->set_sensitive (false);
          progress_processing->property_fraction() = 0.;
      }

      void
      LibraryUIModify::on_action_scan ()
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          set_sensitive (false);

          progress_processing->set_sensitive (true);
          progress_processing->property_fraction() = 0.;

          Gtk::TreeModel::iterator m_iter = view_local->get_selection()->get_selected ();
          Bmp::Audio::ProcessorPUID *processor = new Bmp::Audio::ProcessorPUID;

          processed = false;
          processor->signal_eos().connect (sigc::ptr_fun (&set_processed));
          processor->signal_position().connect (sigc::bind (sigc::ptr_fun (&set_progress_bar), progress_processing, processor));
          processor->set_uri (ustring((*m_iter)[columns.location]));
          processor->run ();
        
          while (!processed)
          {
              while (gtk_events_pending())
              {
                gtk_main_iteration();
              }
          }

          boost::optional<ustring> puid = processor->get_puid ();

          delete processor;

          progress_processing->set_sensitive (false);
          progress_processing->property_fraction() = 0.;

          if (puid)
            {
              MBTrackV tracks; 
              mb_tracks_by_puid (puid.get(), tracks);
  
              MBReleaseV releases;    
              for (MBTrackV::const_iterator i = tracks.begin(); i != tracks.end(); ++i)
                {
                  if (i->release_id)
                    {
                      mb_releases_by_id (i->release_id.get(), releases);
                    }
                }
              populate_data (releases);
            }
          set_sensitive (true);
      }
#endif

      void
      LibraryUIModify::on_action_update ()
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          set_sensitive (false);
          MBReleaseV releases; 
          mb_query_releases (merge_cb_artist_entry->get_text (), merge_cb_album_entry->get_text (), releases);
          populate_data (releases);
          populate_local_store ();
          set_sensitive (true);
      }

      void
      LibraryUIModify::populate_local_store ()
      {
          using namespace Bmp::DB;
          using namespace Bmp::Library;
          using namespace Bmp::MusicBrainzXML;
          using namespace std;
          using namespace Gtk;
          using namespace Glib;

          //--- Set Up Local 
          store_local->clear ();

#ifdef HAVE_HAL
          for (::Bmp::Library::PUDIVRPV::const_iterator i = m_list.begin (); i != m_list.end(); ++i)
          {
              AttributeV attrs; 
              PUDIVRP const& pair = (*i);
              attrs.push_back (Attribute (DB::EXACT, MetadatumId (DATUM_HAL_VOLUME_UDI), ustring(pair.first))); 
              attrs.push_back (Attribute (DB::EXACT, MetadatumId (DATUM_VOLUME_RELATIVE_PATH), ustring(pair.second)));
              RowV vector;
              library->query (attrs, vector);
#else

          for (::Bmp::VFS::UriList::const_iterator i = m_list.begin (); i != m_list.end(); ++i)
          {
              AttributeV attrs; 
              attrs.push_back (Attribute (DB::EXACT, MetadatumId (DATUM_LOCATION), (*i)));
              RowV vector;
              library->query (attrs, vector);
#endif

              for (RowV::const_iterator v = vector.begin (), end = vector.end (); v != end; ++v) 
                { 
                  Track track (*v);
                  TreeModel::iterator iter_l;

                  try {
                    ustring location = track.location.get(); 
                    iter_l = store_local->append ();
                    (*iter_l)[columns.location] = location; 
                  } catch (...) { continue; }

                  if (track.title) (*iter_l)[columns.title] = track.title.get(); 
                  (*iter_l)[columns.tracknumber] = track.tracknumber ? track.tracknumber.get() : 0;

#ifdef HAVE_HAL
                  (*iter_l)[columns.volume_udi] = track.volume_udi.get();
                  (*iter_l)[columns.volume_relative_path] = track.volume_relative_path.get();
#endif

#ifdef HAVE_OFA
                  set_local_puid_state (iter_l, track.puid ? P_PUID_METADATA : P_NONE); 
#endif
                  (*iter_l)[columns.modify] = true;
                }
            }

          view_local->columns_autosize ();

          TreeNodeChildren const& nodes_release = store_release->children();
          TreeNodeChildren const& nodes_local   = store_local->children();

          if (nodes_local.size () < nodes_release.size())
            {
              while (nodes_local.size () < nodes_release.size ()) store_local->append();
            }
      }

      void
      LibraryUIModify::album_changed ()
      {
          using namespace Bmp::DB;
          using namespace Bmp::Library;
          using namespace Bmp::MusicBrainzXML;
          using namespace std;
          using namespace Gtk;
          using namespace Glib;

          if (merge_cb_album->get_active_row_number () < 0) return;

          set_sensitive (false);

          //--- Set Up Release 
          TreeModel::const_iterator m_iter  = merge_cb_album->get_active ();
          ReleaseMap::iterator r_iter       = merge_releases.find ((*m_iter)[albums.id]);
          MBRelease const& release          = r_iter->second;
          try {
              if (release.asin)
                {
                  Glib::RefPtr<Gdk::Pixbuf> cover;
                  Amazon::get_cover_by_asin (release.asin.get(), cover);
                  merge_coverart->set (cover);
                  merge_asin->set_text (release.asin.get());
                }
              else throw 0;
            }
          catch (...)
            { 
              merge_coverart->set (disc);
            }

          merge_artist_id->set_text   (release.artist_id);
          merge_release_id->set_text  (release.release_id);

          if (release.score)
            {
              merge_score->set_text ((boost::format ("%d") % release.score.get()).str());
            }

          if (release.track_list)
            {
              l_tracks->set_text ((boost::format ("%d") % release.track_list.get()).str());
            }

          MBReleaseV releases;
          mb_releases_by_id (release.release_id, releases);
          m_release = releases[0];
          MBTrackV const& tracks = releases[0].tracks.get();
          store_release->clear ();
          for (MBTrackV::const_iterator iter  = tracks.begin  (),
                                        end   = tracks.end    ();
                                        iter != end; ++iter)
            {
              TreeModel::iterator iter_r        = store_release->append();
              (*iter_r)[tracks_cr.title]        = iter->track_title; 
              (*iter_r)[tracks_cr.tracknumber]  = iter->tracknumber; 
              (*iter_r)[tracks_cr.track]        = (*iter);
            }

          view_release->columns_autosize ();

          populate_local_store ();

#ifdef HAVE_OFA
          m_actions->get_action (ACTION_SCAN_ALL)->set_sensitive (true);
#endif

          ok->set_sensitive (true);
          set_sensitive (true);
      }

      void
      LibraryUIModify::artist_changed ()
      {
          using namespace Gtk;
          using namespace MusicBrainzXML;

          if (merge_cb_artist->get_active_row_number () < 0) return;

          set_sensitive (false);

          // Get releases from selected artist
          TreeModel::const_iterator m_iter = merge_cb_artist->get_active ();
          ArtistReleaseMap::iterator map_iter = merge_artist_release_map.find ((*m_iter)[artists.id]);
          ReleaseIds & ids = map_iter->second;
          merge_cb_album_store->clear ();

          typedef std::set<Glib::ustring> ReleaseSet;
          ReleaseSet release_set;

          for (ReleaseIds::const_iterator r_iter  = ids.begin (),
                                          r_end   = ids.end ();
                                          r_iter != r_end; ++r_iter)
            {
              // Iter from releases map
              ReleaseMap::iterator map_iter = merge_releases.find (*r_iter);

              // Get the release based on id 
              MBRelease const& release = map_iter->second; 

              if (release_set.find(release.release_id) != release_set.end()) continue;
              release_set.insert (release.release_id);

              TreeModel::iterator c_iter = merge_cb_album_store->append ();
              (*c_iter)[albums.name] = release.release_title;
              (*c_iter)[albums.id] = release.release_id;
            }
      
          merge_cb_album_entry->set_text("");
          store_release->clear();
          populate_local_store ();

#ifdef HAVE_OFA
          m_actions->get_action (ACTION_SCAN_FILE)->set_sensitive (false);
          m_actions->get_action (ACTION_SCAN_ALL)->set_sensitive (false);
#endif

          ok->set_sensitive (false);

          if (merge_cb_album_store->children().size() == 1)  
            {
              merge_cb_album->set_active (0);
            }

          set_sensitive (true);
      }

#ifdef HAVE_HAL
      int
      LibraryUIModify::run (MusicBrainzXML::MBReleaseV const& releases,
                            Bmp::Library::PUDIVRPV const& list,
                            ModifyTracks& modify_tracks)
#else
      int
      LibraryUIModify::run (MusicBrainzXML::MBReleaseV const& releases,
                            Bmp::VFS::UriList const& list,
                            ModifyTracks & modify_tracks)
#endif
      {
          using namespace Gtk;
          using namespace Glib;
          using namespace Bmp::MusicBrainzXML;
          using namespace Bmp::Library;
          using namespace std;

          m_list = list;

          merge_cb_artist_store = ListStore::create (artists);
          merge_cb_album_store  = ListStore::create (albums);

          for (MBReleaseV::const_iterator r_iter  = releases.begin(),
                                          r_end   = releases.end();
                                          r_iter != r_end; ++r_iter)
          {
            MBRelease const& release = (*r_iter);
            merge_releases.insert (std::make_pair (release.release_id, release));
        
            ArtistReleaseMap::iterator map_iter = merge_artist_release_map.find (release.artist_id);

            if (map_iter != merge_artist_release_map.end ())
              {
                ReleaseIds &ids = map_iter->second;
                ids.push_back (release.release_id);
              }
            else
              {
                ReleaseIds ids; 
                TreeModel::iterator m_iter  = merge_cb_artist_store->append ();
                (*m_iter)[artists.name]	= release.artist_name;
                (*m_iter)[artists.id]	= release.artist_id;
                ids.push_back (release.release_id);
                merge_artist_release_map.insert (std::make_pair (release.artist_id, ids)); 
              }
          }

          // Setup 'Artist' ComboBox 
          merge_cb_artist->set_model (merge_cb_artist_store);
          merge_cb_artist->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::artist_changed));
          merge_cb_artist->set_text_column (0);
    
          // Setup 'Album' ComboBox 
          merge_cb_album->set_model (merge_cb_album_store);
          merge_cb_album->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::album_changed));
          merge_cb_album->set_text_column (0);

          if (!merge_cb_artist_store->children().empty())
              merge_cb_artist->set_active (0);
          if (!merge_cb_album_store->children().empty())
              merge_cb_album->set_active (0);
          else
              populate_local_store ();

          if (Gtk::Dialog::run ()  == GTK_RESPONSE_OK)
            {
              TreeNodeChildren const& nodes_release = store_release->children();
              TreeNodeChildren const& nodes_local   = store_local->children();

              unsigned int max_nodes = nodes_release.size();
              if (nodes_local.size() < max_nodes) max_nodes = nodes_local.size();

              TreeNodeChildren::const_iterator iter_r = nodes_release.begin();
              TreeNodeChildren::const_iterator iter_l = nodes_local.begin();

              boost::optional<ustring> genre;

              if (cb_modify_genre->get_active()) genre = cb_genre_entry->get_text ();

              for (unsigned int n = 0; n < max_nodes; ++n)
              {
                  if ((*iter_l)[columns.modify]) 
                  {
                      TrackUpdated update_track;
#ifdef HAVE_OFA
                      if (PuidState((*iter_l)[columns.puid_state]) == P_PUID_METADATA) 
                      {
                          MBTrack const& track = (*iter_l)[columns.track];
                          mb_track_to_update_track (track, update_track);

                          update_track.mb_track_id                = track.track_id;
                          update_track.asin                       = m_release.asin; 
                          update_track.puid                       = (*iter_l)[columns.puid];
                          update_track.tracknumber                = (*iter_l)[columns.tracknumber];
                          update_track.location_cur               = (*iter_l)[columns.location];
                          update_track.artist                     = track.artist_name;
                          update_track.title                      = track.track_title.length() ?
                                                                    track.track_title :
                                                                    (*iter_r)[tracks_cr.title];
                          update_track.mb_artist_id               = track.artist_id;
                          update_track.mb_artist_sort_name        = track.artist_sort_name; 
                          update_track.album                      = m_release.release_title; 
                          update_track.mb_album_id                = m_release.release_id;
                          update_track.mb_album_artist            = m_release.artist_name; 
                          update_track.mb_album_artist_sort_name  = m_release.artist_sort_name;
                          update_track.mb_album_artist_id         = m_release.artist_id;
                          update_track.mb_release_date            = m_release.date;
                      }
                      else
#endif
                      {
                          MBTrack const& track = (*iter_r)[tracks_cr.track];
                          mb_track_to_update_track (track, update_track);

                          update_track.mb_track_id                = track.track_id;
                          update_track.asin                       = m_release.asin; 
                          update_track.tracknumber                = (*iter_r)[tracks_cr.tracknumber];
                          update_track.location_cur               = (*iter_l)[columns.location];
                          update_track.artist                     = track.artist_name;
                          update_track.title                      = track.track_title;
                          update_track.mb_artist_id               = track.artist_id;
                          update_track.mb_artist_sort_name        = track.artist_sort_name; 
                          update_track.album                      = m_release.release_title; 
                          update_track.mb_album_id                = m_release.release_id;
                          update_track.mb_album_artist            = m_release.artist_name; 
                          update_track.mb_album_artist_sort_name  = m_release.artist_sort_name;
                          update_track.mb_album_artist_id         = m_release.artist_id;
                          update_track.mb_release_date            = m_release.date;
                      }
 
                      update_track.genre = genre; 
#ifdef HAVE_HAL
                      update_track.volume_udi = (*iter_l)[columns.volume_udi];
                      update_track.volume_relative_path = (*iter_l)[columns.volume_relative_path];
#endif
                      modify_tracks.tracks.push_back (update_track);
                  }
                  ++iter_r;
                  ++iter_l;
              }

              modify_tracks.album.asin        = m_release.asin;
              modify_tracks.album.artist      = m_release.artist_name;
              modify_tracks.album.album       = m_release.release_title;
              modify_tracks.album.release_id  = m_release.release_id;
              modify_tracks.album.artist_id   = m_release.artist_id;
              modify_tracks.album.tracks      = modify_tracks.tracks.size(); 
              modify_tracks.album.sortname    = m_release.artist_sort_name; 

              return GTK_RESPONSE_OK;
            }
          else
            {
              return GTK_RESPONSE_CANCEL;
            }
      }
}
