//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_LIBRARY_CLASS_HH
#define BMP_LIBRARY_CLASS_HH

#include <glibmm.h>
#include <glibmm/i18n.h>
#include <boost/optional.hpp>
#include <set>

#ifndef BMP_PLUGIN_BUILD
#  include <taglib/tag.h>
#endif

#if defined (HAVE_HAL) && !defined (BMP_PLUGIN_BUILD)
#  include "x_hal.hh"
#endif

#include "database.hh"
#include "util.hh"

namespace Bmp
{
  namespace UIPart { class Albums; class CDDA; } 

  namespace Library
  {

#if 0
    enum MB_AudioCD_Status
    {
        MB_NONE     = -1,
        MB_FAILURE,
        MB_OK,
        MB_OFFLINE
    };
#endif

#include "exception.hh"

    EXCEPTION(NoMetadata)
    EXCEPTION(NoMetadataTaglib)
    EXCEPTION(NoMetadataGst)
    EXCEPTION(NoAsin)
    EXCEPTION(InvalidFile)
    EXCEPTION(NonReadableFile)
    EXCEPTION(NonWritableFile)
    EXCEPTION(ConversionError)
    EXCEPTION(NoSortnameForAlbumError)
    EXCEPTION(NoNameForSortnameError)
    EXCEPTION(MetadataWriteError)

    enum DatumTags 
    {
      TS_DATUM_ID,
      TS_DATUM_DESC,  
      TS_DATUM_TABLE
    };

    enum DatumTag 
    {
      TAG_DATUM_ID,
      TAG_DATUM_ATTR_NAME,
      TAG_DATUM_ATTR_TYPE,
      TAG_DATUM_ATTR_PRECISE,
      TAG_DATUM_ATTR_V_BOOL,
      TAG_DATUM_ATTR_V_INT,
      TAG_DATUM_ATTR_V_REAL,
      TAG_DATUM_ATTR_V_STRING,
    };

    enum Datum
    {
      DATUM_LOCATION,
      DATUM_ARTIST,
      DATUM_ALBUM,
      DATUM_TITLE,
      DATUM_TRACK,
      DATUM_TIME,
      DATUM_GENRE,
      DATUM_COMMENT,
      DATUM_RATING,
      DATUM_DATE,
      DATUM_MTIME,
      DATUM_BITRATE,
      DATUM_SAMPLERATE,
      DATUM_COUNT,
      DATUM_PLAYDATE,
      DATUM_ASIN,

#ifdef HAVE_HAL
      DATUM_HAL_VOLUME_UDI,
      DATUM_HAL_DEVICE_UDI,
      DATUM_VOLUME_RELATIVE_PATH,
      DATUM_VOLUME_RELATIVE_BASE_DIR,
#endif

      DATUM_MB_ALBUM_ARTIST,
      DATUM_MB_ALBUM_ARTIST_ID,
      DATUM_MB_ALBUM_ARTIST_SORTNAME,
      DATUM_MB_TRACK_ID,
      DATUM_MB_ALBUM_ID,
      DATUM_MB_ARTIST_ID,
      DATUM_MB_ARTIST_SORTNAME,
      DATUM_MB_RELEASE_DATE,
      DATUM_MUSICIP_PUID,

      DATUM_NEW_ITEM,
      DATUM_HASH,

      DATUM_LAST
    };

    struct DatumDefine
    {
      const char    *title;
      const char    *title_plural;
      const char    *description;
      const char    *id;
      DB::ValueType  type;
    };

    class Album
    {
      public:

        boost::optional <Glib::ustring>   artist;
        boost::optional <Glib::ustring>   album;
        boost::optional <Glib::ustring>   asin;
        boost::optional <Glib::ustring>   release_id;
        boost::optional <Glib::ustring>   artist_id;
        boost::optional <bool>            new_item;
        boost::optional <unsigned int>    tracks;
        boost::optional <Glib::ustring>   sortname;

        static std::string key (Album const& a)
        {
          return ((a.sortname ? a.sortname.get() : a.artist.get())+a.album.get()).casefold_collate_key();
        }

        Album (const Bmp::DB::Row& row);
        Album () {};

    };
    typedef std::vector< Album > AlbumList;

    bool
    operator== (Album const& a,
                Album const& b);

    void
    album_to_attrs (Bmp::Library::Album const& album,
                    Bmp::DB::AttributeV & attrs);

    class Track
    {
      public:

#ifdef HAVE_HAL
        boost::optional <Glib::ustring>   volume_udi;
        boost::optional <Glib::ustring>   device_udi;
        boost::optional <Glib::ustring>   volume_relative_path;
#endif

        boost::optional <Glib::ustring>   location;
        boost::optional <Glib::ustring>   artist;
        boost::optional <Glib::ustring>   album;
        boost::optional <Glib::ustring>   title;
        boost::optional <Glib::ustring>   genre;
        boost::optional <Glib::ustring>   comment;
        boost::optional <Glib::ustring>   asin;
        boost::optional <Glib::ustring>   puid;
        boost::optional <Glib::ustring>   mb_album_artist;
        boost::optional <Glib::ustring>   mb_album_artist_id;
        boost::optional <Glib::ustring>   mb_album_artist_sort_name;
        boost::optional <Glib::ustring>   mb_track_id;
        boost::optional <Glib::ustring>   mb_album_id;
        boost::optional <Glib::ustring>   mb_artist_id;
        boost::optional <Glib::ustring>   mb_artist_sort_name;
        boost::optional <Glib::ustring>   mb_release_date;
        boost::optional <Glib::ustring>   hash;

        boost::optional <int>     tracknumber;
        boost::optional <int>     duration;
        boost::optional <int>     rating;
        boost::optional <int>     date;
        boost::optional <int>     mtime;
        boost::optional <int>     bitrate;
        boost::optional <int>     samplerate;
        boost::optional <int>     count;
        boost::optional <int>     playdate;
        boost::optional <bool>    new_item;

#ifdef HAVE_HAL
        Track (const Bmp::DB::Row& row, bool use_hal = true);
#else
        Track (const Bmp::DB::Row& row);
#endif
        Track () {};
    };

    class TrackUpdated
    {
      public:

        boost::optional <Glib::ustring>   location_cur;
        boost::optional <Glib::ustring>   location_new;
        boost::optional <Glib::ustring>   artist;
        boost::optional <Glib::ustring>   album;
        boost::optional <Glib::ustring>   title;
        boost::optional <Glib::ustring>   genre;
        boost::optional <Glib::ustring>   comment;

        boost::optional <Glib::ustring>   asin;
        boost::optional <Glib::ustring>   puid;

        boost::optional <Glib::ustring>   mb_album_artist;
        boost::optional <Glib::ustring>   mb_album_artist_id;
        boost::optional <Glib::ustring>   mb_album_artist_sort_name;
        boost::optional <Glib::ustring>   mb_track_id;
        boost::optional <Glib::ustring>   mb_album_id;
        boost::optional <Glib::ustring>   mb_artist_id;
        boost::optional <Glib::ustring>   mb_artist_sort_name;
        boost::optional <Glib::ustring>   mb_release_date;

        boost::optional <Glib::ustring>   hash;

#ifdef HAVE_HAL
        boost::optional <Glib::ustring>   volume_udi;
        boost::optional <Glib::ustring>   device_udi;
        boost::optional <Glib::ustring>   volume_relative_path;
#endif

        boost::optional <int>     tracknumber;
        boost::optional <int>     date;
        boost::optional <int>     rating;
        boost::optional <int>     mtime;
        boost::optional <int>     count;
        boost::optional <int>     playdate;
        boost::optional <bool>    new_item;

        TrackUpdated& operator=(Track const& track)
        {
          location_cur                = track.location;
          artist                      = track.artist;
          album                       = track.album;
          title                       = track.title;
          genre                       = track.genre;
          comment                     = track.comment;

          asin                        = track.asin;
          puid                        = track.puid;
#ifdef HAVE_HAL
          volume_udi                  = track.volume_udi;
          device_udi                  = track.device_udi;
          volume_relative_path        = track.volume_relative_path;
#endif
          mb_album_artist             = track.mb_album_artist;
          mb_album_artist_id          = track.mb_album_artist_id;
          mb_album_artist_sort_name   = track.mb_album_artist_sort_name;
          mb_track_id                 = track.mb_track_id;
          mb_album_id                 = track.mb_album_id;
          mb_artist_id                = track.mb_artist_id;
          mb_artist_sort_name         = track.mb_artist_sort_name;

          tracknumber                 = track.tracknumber;
          date                        = track.date;
          rating                      = track.rating;
          mtime                       = track.mtime;
          playdate                    = track.playdate;
          new_item                    = track.new_item;
         
          return *this;
        }
    };

    typedef std::vector < TrackUpdated > TrackUpdatedV;

    struct CollectionAttribute
    {
      Datum             datum;
      bool              precise;
      DB::Variant       value;
    };

    typedef std::vector < CollectionAttribute > CollectionAttributes;

#ifdef HAVE_HAL
    typedef std::pair   < std::string, std::string > PUDIVRP;
    typedef std::vector < PUDIVRP > PUDIVRPV;
#endif

    class Library
      {
        private:

          void    
          get_metadata_row      (Glib::ustring const& uri,
                                 Bmp::DB::Row&        row) const;

        public:

          typedef sigc::signal <void, int> ModifyStart;
          typedef sigc::signal <void>      ModifyStep;
          typedef sigc::signal <void>      ModifyEnd;
          typedef sigc::signal <void>      SignalProcessing;
          typedef sigc::signal <void, Glib::ustring, Bmp::Library::TrackUpdated > SignalRowUpdated;

          SignalProcessing& signal_processing_start ();
          SignalProcessing& signal_processing_end   ();
          SignalRowUpdated& signal_row_updated      ();

          ModifyStart& signal_modify_start ();
          ModifyStep&  signal_modify_step  ();
          ModifyEnd&   signal_modify_end   ();

          Library ();
          ~Library ();

          void
          approve_items ();
          void
          approve_items (Album const& sa);
          void
          drop_items (); 

          void
          del   (DB::AttributeV const& attrs);
          bool
          add   (Glib::ustring const& uri);


          bool
          is_library_track      (Track const& track);

          bool
          modify_tracks         (TrackUpdatedV & tracks,
                                 bool retag     = true,
                                 bool db_update = true,
                                 bool relocate  = true);
          void    
          get_metadata          (Glib::ustring const& uri,
                                 Track & track);
          void
          get_albums            (DB::RowV & vector);

#ifndef BMP_PLUGIN_BUILD

          bool
          album_exists          (Album const& album_in);

          bool
          album_has_new_items   (Album const& album_in);

          bool
          library_has_new_items ();

          Album
          get_album             (Album const& a);
#endif

          GHashTable*         get_hashtable         (Glib::ustring const& uri) const;
          DatumDefine         get_metadatum_define  (Datum datum) const;
          void                increase_play_count   (Glib::ustring const& uri);

          bool collection_exists  (std::string const& name) const;
          void collection_create  (std::string const& name, Glib::ustring const& desc, CollectionAttributes const& attrs); 
                                
          static std::string metadatum_id (Datum datum); 

    protected:

          friend class Bmp::UIPart::Albums;
          friend class Bmp::UIPart::CDDA;

#ifndef BMP_PLUGIN_BUILD
          void
          metadata_set_common   (const TrackUpdated& track, TagLib::Tag* tag);
#endif
          void 
          metadata_set_taglib   (const Glib::ustring &uri, TrackUpdated& track);

          void 
          metadata_get_taglib   (const Glib::ustring &uri, DB::Row &row) const;
          void
          metadata_get_gst      (const Glib::ustring &uri, DB::Row &row) const;

          
    private:

          void recreate_db ();

          SignalProcessing signal_processing_start_;
          SignalProcessing signal_processing_end_;
          SignalRowUpdated signal_row_updated_;

          ModifyStart  signal_modify_start_;
          ModifyStep   signal_modify_step_;
          ModifyEnd    signal_modify_end_;

          DB::ValueMap  map_main;
          DB::ValueMap  map_albums;
          DB::ValueMap  map_va;
          DB::Database *db_main;

          DB::ValueMap  map_tags;
          DB::ValueMap  map_tag;
          DB::Database *db_tags;

          DB::ValueMap  map_cdda;
          DB::Database *db_cdda;

    public:

          void
          query                   (DB::AttributeV const& attrs,
                                   DB::RowV&             vector,
                                   DB::Database*         db    = 0,
                                   std::string           name  = "main") const;
    }; 
  } // !Library:: 

  std::string
  MetadatumId (Bmp::Library::Datum datum);

} // !Bmp:: 

#endif // !BMP_LIBRARY_CLASS_HH
