//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif // HAVE_CONFIG_H

#include <glibmm/i18n.h>
#include <glib/gstdio.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include <fstream>

#include <boost/format.hpp>
#include <boost/optional.hpp>
#include <mcs/mcs.h>

#include "x_lastfm.hh"
#include "x_library.hh"
#include "x_play.hh"
#include "x_service_core.hh"

#include "main.hh"
#include "paths.hh"
#include "debug.hh"
#include "stock.hh"

#include "ui-toolbox.hh"
#include "util.hh"
#include "uri++.hh"

#include "ui-banner-image.hh"

#include "ui-part-albums.hh"
#include "ui-part-cdda.hh"
#include "ui-part-lastfm.hh"
#include "ui-part-playlist.hh"

#include "ui-part-shoutcast-streams.hh"
#include "ui-part-icecast-streams.hh"

#include "main-ui.hh"
#include "popup.hh"

#ifdef HAVE_MOOD
#  include "ui-part-downloads.hh"
#endif // HAVE_MOOD

#ifdef HAVE_GAIM
#  include "gaim-dbus.hh"
#  include "dialog-gaim-send-file.hh"
#endif // HAVE_GAIM

#include <widgets/bmp_tray_icon.h>
#include <widgets/bmp_status_icon.h>

#include "preferences-ui.hh"
#include "dialog-about.hh"

namespace
{
  static boost::format f_time ("<b>%02d:%02d</b>");
  static boost::format f_button_name ("tb-main-%d");
  static boost::format f_image_name  ("i-toolbar-%d");

  const char *source_icons_stock [] =
  {
    BMP_STOCK_LIBRARY,
    BMP_STOCK_SHOUTCAST,
    BMP_STOCK_XIPH,
    BMP_STOCK_CDDA,
    BMP_STOCK_LASTFM,
    BMP_STOCK_PLAYLIST,
#ifdef HAVE_MOOD
    BMP_STOCK_MUSEEK,
#endif
  };

  void
  mcs_volume_changed (MCS_CB_DEFAULT_SIGNATURE, Gtk::Range *range)
  {
    int volume = boost::get<int>(value); 
    play->property_volume() = volume;
    if (int(range->get_value()) != volume)
      {
        range->set_value (volume);
      }
  }

  void
  show_prefs (boost::shared_ptr<Bmp::Preferences>& preferences)
  {  
    preferences->present ();
  }

  bool
  shutdown (GdkEventAny *event)
  {
    core->stop ();
    return true;
  }
}

namespace Bmp
{
  Mcs::Bind *mcs_bind = 0;
}

namespace Bmp
{
      bool
      MainUI::display_popup_delayed ()
      {
        m_popup->tooltip_mode ();
        return false;
      }

      gboolean
      MainUI::status_icon_enter (BmpStatusIcon *icon, GdkEventCrossing *event, gpointer data)
      {
        Bmp::MainUI *main = reinterpret_cast<Bmp::MainUI *>(data);
        if (main->m_active_source != SOURCE_NONE)
          {
            main->delay_popup = Glib::signal_timeout().connect (sigc::mem_fun (main, &Bmp::MainUI::display_popup_delayed), 400);
          }
        return FALSE;
      }

      gboolean
      MainUI::status_icon_leave (BmpStatusIcon *icon, GdkEventCrossing *event, gpointer data) 
      {
        Bmp::MainUI *main = reinterpret_cast<Bmp::MainUI *>(data);
        main->delay_popup.disconnect ();
        main->m_popup->tooltip_mode (false);
        return FALSE;
      }

      void
      MainUI::status_icon_popup_menu (BmpStatusIcon *icon, guint arg1, guint arg2, gpointer data)
      {
        Bmp::MainUI *main = reinterpret_cast<Bmp::MainUI *>(data);

        main->delay_popup.disconnect ();
        if (main->m_popup) main->m_popup->tooltip_mode (false, true);

        GtkWidget *menu = Util::ui_manager_get_popup (main->m_ui_manager_main->gobj(), "/popup-tray/menu-tray");
        gtk_widget_realize (menu);
        gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, arg1, arg2);
      }

      void
      MainUI::status_icon_click (BmpStatusIcon *icon, gpointer data)
      {
        Bmp::MainUI *main = reinterpret_cast<Bmp::MainUI *>(data);

        main->delay_popup.disconnect ();
        if (main->m_popup) main->m_popup->tooltip_mode (false, true);

        if (main->m_visible)
          {
            main->set_skip_taskbar_hint (true);
            main->set_skip_pager_hint (true);
            main->hide ();
            main->m_visible = false;
          }
        else
          {
            main->set_skip_taskbar_hint (false);
            main->set_skip_pager_hint (false);
            main->present ();
            main->m_visible = true;
          }
      }

      void
      MainUI::status_icon_scroll_up (BmpStatusIcon *icon, gpointer data)
      {
        int volume;
        volume = mcs->key_get<int>("bmp", "volume"); 
        volume += 4;
        volume = (volume > 100) ? 100 : volume;
        mcs->key_set<int>("bmp", "volume", volume);
      }

      void
      MainUI::status_icon_scroll_down (BmpStatusIcon *icon, gpointer data)
      {
        int volume;
        volume = mcs->key_get<int>("bmp", "volume"); 
        volume -= 4;
        volume = (volume < 0) ? 0 : volume;
        mcs->key_set<int>("bmp", "volume", volume);
      }

      void
      MainUI::create_popup (GtkPlug *plug, gpointer data)
      {
        Bmp::MainUI *main = reinterpret_cast<Bmp::MainUI *>(data);

        GtkWidget *widget = GTK_WIDGET (plug);

        Glib::ustring text ("");
        Gdk::Color outline, inlay, plate;
        outline.set_rgb (0, 0, 0);
        inlay.set_rgb (65535, 65535, 65535);
        plate.set_rgb_p (0.33, 0.60, 0.84);
        main->m_popup = new Bmp::Popup (widget, outline, inlay, plate, Glib::RefPtr<Gdk::Pixbuf>(0), text);
        g_object_connect (G_OBJECT (plug),
                          "signal::enter-notify-event",
                          G_CALLBACK(MainUI::status_icon_enter),
                          main,
                          "signal::leave-notify-event",
                          G_CALLBACK(MainUI::status_icon_leave),
                          main,
                          NULL);
      }

      void
      MainUI::display_popup (SimpleTrackInfo& sti)
      {
        if (m_popup)
        {
            m_popup->set_info (sti);
            if (mcs->key_get<bool>("bmp","display-notifications"))
            {
              m_popup->show ();
            }
        }
      }

#define MAIN_ACTION_ABOUT     "main-action-about"
#define MAIN_ACTION_PLAY      "main-action-play"
#define MAIN_ACTION_PAUSE     "main-action-pause"
#define MAIN_ACTION_PREV      "main-action-prev"
#define MAIN_ACTION_NEXT      "main-action-next"
#define MAIN_ACTION_STOP      "main-action-stop"
#define MAIN_ACTION_QUIT      "main-action-quit"

#define MAIN_ACTION_RECOMMEND "main-action-recommend"

#ifdef HAVE_GAIM
#  define MAIN_ACTION_GAIM_SEND "main-action-gaim-send"
#endif

      MainUI::MainUI (BaseObjectType                        *cobject,
                      const Glib::RefPtr<Gnome::Glade::Xml> &xml)
          : Gtk::Window       (cobject),
            m_ref_xml         (xml),
            m_active_source   (SOURCE_NONE),
            m_popup           (0),
            m_visible         (true),
            m_seeking         (false),
            m_sent            (false),
            m_error_condition (false),
            m_monitor         (false),
            m_dialog_about    (new Bmp::AboutDialog())
      {
          mcs_bind = new Mcs::Bind(mcs);
          Bmp::register_stock_icons ();

          // Connect to Shutdown
          signal_delete_event().connect (sigc::ptr_fun (&shutdown));

          // Additional UIs
          m_preferences = boost::shared_ptr<Bmp::Preferences>(Bmp::Preferences::create());
          dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_preferences"))->signal_clicked().connect
              (sigc::bind (sigc::ptr_fun (&show_prefs), m_preferences));

#ifdef HAVE_GAIM
          gaim_dbus = boost::shared_ptr<Bmp::Gaim::DBUS>(new Bmp::Gaim::DBUS());
          gaim_dbus->signal_gaim_lost().connect(sigc::mem_fun (this, &Bmp::MainUI::gaim_lost));
          gaim_dbus->signal_gaim_present().connect(sigc::mem_fun (this, &Bmp::MainUI::gaim_present));
#endif

          // Header Banner Image
          {
            Glib::RefPtr<Gdk::Pixbuf> left, right, slice;
            left  = Gdk::Pixbuf::create_from_file (BMP_IMAGE_DIR_MAIN G_DIR_SEPARATOR_S "header-logo-left.png");
            right = Gdk::Pixbuf::create_from_file (BMP_IMAGE_DIR_MAIN G_DIR_SEPARATOR_S "header-logo-right.png");
            slice = Gdk::Pixbuf::create_from_file (BMP_IMAGE_DIR_MAIN G_DIR_SEPARATOR_S "header-logo-slice.png");
            Bmp::UI::BannerImage *banner = Gtk::manage (new Bmp::UI::BannerImage (left, right, slice));
            Gtk::Alignment *a;
            m_ref_xml->get_widget ("header_alignment", a);
            a->add (*banner);
            a->show_all ();
            banner->show_all ();
            banner->set_size_request (-1, slice->get_height());
          }

          Util::window_set_icon_list (GTK_WIDGET (gobj ()), "player");

          // Bottom status row
          Gdk::Color fg, bg;
          fg.set_rgb_p (1., 1., 1.);
          bg.set_rgb_p (0., 0., 0.);

          m_ref_xml->get_widget ("status_image", status_image);
          m_ref_xml->get_widget ("status_label", status_label);
          m_ref_xml->get_widget ("time_label"  , time_label);
          m_ref_xml->get_widget ("status_evbox", status_evbox);

          status_evbox->add_events (Gdk::BUTTON_PRESS_MASK);
          status_evbox->signal_button_press_event().connect
                  (sigc::mem_fun (this, &Bmp::MainUI::clear_error));
          status_evbox->modify_bg (Gtk::STATE_NORMAL, bg);

          m_warning = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_MAIN, "warning.png"));

          m_ref_xml->get_widget ("table_metadata")->modify_bg (Gtk::STATE_NORMAL, bg);
          m_ref_xml->get_widget ("metadata_evbox")->modify_bg (Gtk::STATE_NORMAL, bg);
          m_ref_xml->get_widget ("time_evbox")->modify_bg (Gtk::STATE_NORMAL, bg);
          status_label->modify_fg (Gtk::STATE_NORMAL, fg);
          time_label->modify_fg (Gtk::STATE_NORMAL, fg);

          yy_playing  = Gdk::PixbufAnimation::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_MAIN, "yingyang.gif"));
          yy_paused   = Gdk::PixbufAnimation::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_MAIN, "yingyang_paused.gif"));
          yy_seeking  = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_MAIN, "yingyang_seeking.png"));
          yy_stopped  = yy_playing->get_static_image (); 

          status_image->set (yy_stopped);

          // Volume
          mcs_bind->bind_range (reinterpret_cast<Gtk::Range *>(m_ref_xml->get_widget ("volume")), "bmp", "volume");
          mcs->subscribe ("MainUI", "bmp", "volume", sigc::bind (sigc::ptr_fun (&mcs_volume_changed), 
                  reinterpret_cast<Gtk::Range *>(m_ref_xml->get_widget ("volume"))));

          dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget( "i-lastfm-recommend" ))->set
                  (Gtk::StockID (BMP_STOCK_LASTFM), Gtk::ICON_SIZE_SMALL_TOOLBAR);

#ifdef HAVE_GAIM
          dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget( "i-gaim-send" ))->set
                  (Gtk::StockID (BMP_STOCK_GAIM), Gtk::ICON_SIZE_SMALL_TOOLBAR);
#endif //HAVE_GAIM


          // Tray icon
          m_status_icon_image_default =
            Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_TRAY_ICON_DIR, "tray-icon-default.png"));
          m_status_icon_image_paused =
            Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_TRAY_ICON_DIR, "tray-icon-paused.png"));
          m_status_icon_image_playing =
            Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_TRAY_ICON_DIR, "tray-icon-playing.png"));

          m_status_icon = bmp_status_icon_new_from_pixbuf (m_status_icon_image_default->gobj());
          bmp_status_icon_set_visible (m_status_icon, TRUE);

          g_object_connect (G_OBJECT (m_status_icon),
                            "signal::click",
                            G_CALLBACK(MainUI::status_icon_click),
                            this,
                            "signal::popup-menu",
                            G_CALLBACK(MainUI::status_icon_popup_menu),
                            this,
                            "signal::scroll-up",
                            G_CALLBACK(MainUI::status_icon_scroll_up),
                            this,
                            "signal::scroll-down",
                            G_CALLBACK(MainUI::status_icon_scroll_down),
                            this,
                            NULL);

          GtkWidget *tray_icon = bmp_status_icon_get_tray_icon (m_status_icon);
          gtk_widget_realize (GTK_WIDGET (tray_icon));
          g_object_connect (G_OBJECT (tray_icon),
                            "signal::embedded",
                            G_CALLBACK (Bmp::MainUI::create_popup),
                            this,
                            NULL);

          // Navigation
          m_ref_xml->get_widget ("notebook-main", notebook_main);
          m_ref_xml->get_widget_derived ("main-treeview-sources", treeview_sources);

          ::play->property_status().signal_changed().connect
                (sigc::mem_fun (this, &Bmp::MainUI::playstatus_changed));
          ::play->signal_eos().connect
                (sigc::mem_fun (this, &Bmp::MainUI::next));
          ::play->signal_position().connect
                (sigc::mem_fun (this, &Bmp::MainUI::stream_position));
          ::play->signal_buffering().connect
                (sigc::mem_fun (this, &Bmp::MainUI::buffering));

          // Seek Range
          seekrange = dynamic_cast<Gtk::HScale *>(m_ref_xml->get_widget ("scale-seek"));
          seekrange->add_events (Gdk::BUTTON_PRESS_MASK | Gdk::BUTTON_RELEASE_MASK);
          seekrange->signal_value_changed().connect
                (sigc::mem_fun (this, &Bmp::MainUI::seek_display));
          seekrange->signal_event().connect
                (sigc::mem_fun (this, &Bmp::MainUI::seek_event));
          seekrange->set_sensitive (false);

          
          m_ui_manager_main = Gtk::UIManager::create ();
          m_actions_main = Gtk::ActionGroup::create ("ActionsMain"); 

#ifdef ENABLE_NLS
          gtk_action_group_set_translate_func (m_actions_main->gobj(), Util::menu_translate, NULL, NULL); 
#endif //ENABLE_NLS

          m_actions_main->add ( Gtk::Action::create ("dummy", "dummy"));

          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_ABOUT, Gtk::StockID (BMP_STOCK_ABOUT), ("About BMP")),
                                   sigc::mem_fun (this, &Bmp::MainUI::display_about_dialog));

          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_PLAY, Gtk::StockID (GTK_STOCK_MEDIA_PLAY), _("Play")),
                                   sigc::mem_fun (this, &Bmp::MainUI::play));

          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_PAUSE, Gtk::StockID (GTK_STOCK_MEDIA_PAUSE), _("Pause")),
                                   sigc::mem_fun (this, &Bmp::MainUI::pause));

          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_PREV, Gtk::StockID (GTK_STOCK_MEDIA_PREVIOUS), _("Prev")),
                                   sigc::mem_fun (this, &Bmp::MainUI::prev));

          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_NEXT, Gtk::StockID (GTK_STOCK_MEDIA_NEXT), _("Next")),
                                   sigc::mem_fun (this, &Bmp::MainUI::next));

          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_STOP, Gtk::StockID (GTK_STOCK_MEDIA_STOP), _("Stop")),
                                   sigc::mem_fun (this, &Bmp::MainUI::stop));

          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_RECOMMEND, _("Last.FM Recommend")),
                                   sigc::mem_fun (this, &Bmp::MainUI::recommend));

#ifdef HAVE_GAIM
          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_GAIM_SEND, Gtk::StockID (BMP_STOCK_GAIM), _("Send to Gaim Buddy")),
                                   sigc::mem_fun (this, &Bmp::MainUI::gaim_send));
#endif //HAVE_GAIM

          m_actions_main->add ( Gtk::Action::create (MAIN_ACTION_QUIT, Gtk::StockID (GTK_STOCK_QUIT), _("Quit")),
                                   sigc::mem_fun (core, &Bmp::ServiceCore::stop));

          m_ui_manager_main->insert_action_group (m_actions_main);
          m_ui_manager_main->add_ui_from_file (Glib::build_filename (BMP_UI_DIR, "trayicon.ui"));

          // Connect Bottom EvBox for restoring context
          m_ref_xml->get_widget ("eventbox27", evb);
          evb->add_events (Gdk::BUTTON_PRESS_MASK);
          evb->signal_button_press_event ().connect (sigc::mem_fun (this, &Bmp::MainUI::source_restore_context));

          // Action Proxies 
          m_actions_main->get_action (MAIN_ACTION_PLAY)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_play"))));

          m_actions_main->get_action (MAIN_ACTION_PAUSE)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_pause"))));

          m_actions_main->get_action (MAIN_ACTION_PREV)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_prev"))));

          m_actions_main->get_action (MAIN_ACTION_NEXT)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_next"))));

          m_actions_main->get_action (MAIN_ACTION_STOP)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_stop"))));

          m_actions_main->get_action (MAIN_ACTION_RECOMMEND)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_lastfm_recommend"))));

#ifdef HAVE_GAIM
          m_actions_main->get_action (MAIN_ACTION_GAIM_SEND)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_gaim_send"))));

          m_ref_xml->get_widget ("b_gaim_send")->show_all();
#endif //HAVE_GAIM

          m_actions_main->get_action (MAIN_ACTION_PLAY)->set_sensitive (false);
          m_actions_main->get_action (MAIN_ACTION_PAUSE)->set_sensitive (false);
          m_actions_main->get_action (MAIN_ACTION_PREV)->set_sensitive (false);
          m_actions_main->get_action (MAIN_ACTION_NEXT)->set_sensitive (false);
          m_actions_main->get_action (MAIN_ACTION_STOP)->set_sensitive (false);
          m_actions_main->get_action (MAIN_ACTION_RECOMMEND)->set_sensitive (false);

#ifdef HAVE_GAIM
          m_actions_main->get_action (MAIN_ACTION_GAIM_SEND)->set_sensitive (false);
#endif //HAVE_GAIM

          m_source_caps[SOURCE_ALBUMS]      = PlaybackSource::Caps ( PlaybackSource::CAN_SEEK | PlaybackSource::CAN_PROVIDE_METADATA );
          m_source_caps[SOURCE_PLAYLIST]    = PlaybackSource::Caps ( PlaybackSource::CAN_SEEK | PlaybackSource::CAN_PROVIDE_METADATA );
          m_source_caps[SOURCE_SHOUTCAST]   = PlaybackSource::NONE;
          m_source_caps[SOURCE_ICECAST]     = PlaybackSource::NONE;
          m_source_caps[SOURCE_LASTFM]      = PlaybackSource::Caps ( PlaybackSource::CAN_PROVIDE_METADATA );
          m_source_caps[SOURCE_CDDA]        = PlaybackSource::CAN_SEEK;
#ifdef HAVE_MOOD
          m_source_caps[SOURCE_DOWNLOADS]   = PlaybackSource::NONE;
#endif //HAVE_MOOD


          m_sources[SOURCE_PLAYLIST] = boost::shared_ptr<UIPart::Playlist>(new UIPart::Playlist (m_ref_xml));
          m_sources[SOURCE_PLAYLIST]->signal_caps().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_caps), SOURCE_PLAYLIST));
          m_sources[SOURCE_PLAYLIST]->signal_track_info().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_track_info), SOURCE_PLAYLIST));
          m_sources[SOURCE_PLAYLIST]->signal_playback_request().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_playback_request), SOURCE_PLAYLIST));


          m_sources[SOURCE_SHOUTCAST] = boost::shared_ptr<UIPart::ShoutcastStreams>(new UIPart::ShoutcastStreams (m_ref_xml));
          m_sources[SOURCE_SHOUTCAST]->signal_caps().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_caps), SOURCE_SHOUTCAST));
          m_sources[SOURCE_SHOUTCAST]->signal_track_info().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_track_info), SOURCE_SHOUTCAST));


          m_sources[SOURCE_ICECAST] = boost::shared_ptr<UIPart::IcecastStreams>(new UIPart::IcecastStreams (m_ref_xml));
          m_sources[SOURCE_ICECAST]->signal_caps().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_caps), SOURCE_ICECAST));
          m_sources[SOURCE_ICECAST]->signal_track_info().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_track_info), SOURCE_ICECAST));


          m_sources[SOURCE_LASTFM] = boost::shared_ptr<UIPart::LASTFM>(new UIPart::LASTFM (m_ref_xml));
          m_sources[SOURCE_LASTFM]->signal_caps().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_caps), SOURCE_LASTFM));
          m_sources[SOURCE_LASTFM]->signal_track_info().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_track_info), SOURCE_LASTFM));
          m_sources[SOURCE_LASTFM]->signal_playback_request().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_playback_request), SOURCE_LASTFM));
          m_sources[SOURCE_LASTFM]->signal_stop_request().connect
                (sigc::mem_fun (this, &Bmp::MainUI::stop));


          m_sources[SOURCE_CDDA] = boost::shared_ptr<UIPart::CDDA>(new UIPart::CDDA (m_ref_xml));
          m_sources[SOURCE_CDDA]->signal_caps().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_caps), SOURCE_CDDA));
          m_sources[SOURCE_CDDA]->signal_track_info().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_track_info), SOURCE_CDDA));
          reinterpret_cast<Bmp::UIPart::CDDA*>(m_sources[SOURCE_CDDA].get())->signal_request_import().connect
                (sigc::mem_fun (this, &Bmp::MainUI::import_request));


#ifdef HAVE_MOOD
          m_sources[SOURCE_DOWNLOADS] = boost::shared_ptr<UIPart::Downloads>(new UIPart::Downloads (m_ref_xml));
          m_sources[SOURCE_DOWNLOADS]->signal_caps().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_caps), SOURCE_DOWNLOADS));
          m_sources[SOURCE_DOWNLOADS]->signal_track_info().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_track_info), SOURCE_DOWNLOADS));
          reinterpret_cast<Bmp::UIPart::Downloads*>(m_sources[SOURCE_DOWNLOADS].get())->signal_request_import().connect
                (sigc::mem_fun (this, &Bmp::MainUI::import_request));
#endif // HAVE_MOOD
 
          set_sensitive (false);
          while (gtk_events_pending()) gtk_main_iteration ();

          m_sources[SOURCE_ALBUMS] = boost::shared_ptr<UIPart::Albums>(new UIPart::Albums (m_ref_xml));
          m_sources[SOURCE_ALBUMS]->signal_caps().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_caps), SOURCE_ALBUMS));
          m_sources[SOURCE_ALBUMS]->signal_track_info().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::source_track_info), SOURCE_ALBUMS));
          reinterpret_cast<Bmp::UIPart::Albums*>(m_sources[SOURCE_ALBUMS].get())->signal_uri_enqueue_request().connect
                (sigc::bind (sigc::mem_fun (this, &Bmp::MainUI::play_uris), false));

          treeview_sources->select_source (SOURCE_ALBUMS);
          reinterpret_cast<Bmp::UIPart::Albums*>(m_sources[SOURCE_ALBUMS].get())->albums_update (false);

          notebook_main->set_current_page (0);
          m_sources[SOURCE_ALBUMS]->send_caps ();

          treeview_sources->source_selected().connect (sigc::mem_fun (this, &Bmp::MainUI::source_changed));
          set_sensitive (true);
          while (gtk_events_pending()) gtk_main_iteration ();
      }

      bool
      MainUI::clear_error (GdkEventButton*)
      {
        if (!m_error_condition) return false;

        status_label->set_text("");
        status_image->set (yy_stopped);
        m_error_condition = false;

        return false;
      }

      void
      MainUI::set_error (Glib::ustring message, PlaybackSourceID source)
      {
        m_error_condition = true;
        status_label->set_markup (message);
        status_image->set (m_warning);
      }

#ifdef HAVE_GAIM
      void
      MainUI::gaim_present ()
      {
        Playstatus status = Playstatus (::play->property_status().get_value());
        try {
            Bmp::URI u (::play->property_stream().get_value());
            if (status != PLAYSTATUS_STOPPED)
            {
              m_actions_main->get_action (MAIN_ACTION_GAIM_SEND)->set_sensitive (u.get_protocol() == Bmp::URI::PROTOCOL_FILE);
              return;
            }
          }
        catch (...)
          {
              m_actions_main->get_action (MAIN_ACTION_GAIM_SEND)->set_sensitive (false);
          }
      }

      void
      MainUI::gaim_lost ()
      {
        m_actions_main->get_action (MAIN_ACTION_GAIM_SEND)->set_sensitive (false);
      }

      void
      MainUI::gaim_send ()
      {
        Glib::ustring uri (::play->property_stream().get_value());

        // Make sure
        try {
          Bmp::URI u (uri);
          if (u.get_protocol () != Bmp::URI::PROTOCOL_FILE)
            {
              Gtk::MessageDialog dialog (_("You can only send locally accessible files trough Gaim!"), false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
              dialog.set_title (_("Send File: Error - BMP"));
              dialog.run ();
              return;
            }
          }
        catch (...) {} // FIXME: 

        std::string filename = Glib::filename_from_uri (uri);

        try {
          boost::shared_ptr<Bmp::Gaim::SendFile> gaim_dialog (Bmp::Gaim::SendFile::create());
          Bmp::Gaim::DBUS::OnlineBuddies buddies;
          gaim_dbus->get_online_buddies (buddies);
          if ( gaim_dialog->run( buddies , uri ) == GTK_RESPONSE_OK )
          {
            gaim_dbus->send_file( gaim_dialog->get_current_buddy () , filename );
          }
          gaim_dialog->hide ();
        }
        catch (Bmp::Gaim::DBUS::Exception& cxe)
          {
              Glib::ustring message (_("An Error occured while communicating with Gaim: "));
              message.append (cxe.what());
              Gtk::MessageDialog dialog (message, false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
              dialog.set_title (_("Send File: Error - BMP"));
              dialog.run ();
          }
        catch (...) {} //FIXME: What else could happen from our side..?!
      }
#endif

      void
      MainUI::recommend ()
      {
        boost::shared_ptr<Bmp::LastFM::RecommendDialog> dialog (Bmp::LastFM::RecommendDialog::create());
        dialog->run (m_sti.get().artist.get(), m_sti.get().album.get(), m_sti.get().title.get());
      }

      void
      MainUI::display_about_dialog ()
      {
        m_dialog_about->present ();
      }

      void
      MainUI::seek_display ()
      {
        double value = seekrange->get_value();
        int position = int (value);
        int minutes = position / 60;
        int seconds = position % 60;
        time_label->set_markup ((f_time % minutes % seconds).str());
      }

      bool
      MainUI::seek_event (GdkEvent *event)
      {
        if (event->type == GDK_BUTTON_PRESS)
          {
            m_seeking = true;
            status_image->set (yy_seeking);
          }
        else if (event->type == GDK_BUTTON_RELEASE)
          {
            double value = seekrange->get_value();
            ::play->seek(int(value));
            m_seeking = false;
            status_image->set (yy_playing);
          }
        return false;
      }

      void
      MainUI::buffering (double buffered)
      {
        if (buffered == 0.)
            status_image->set (yy_seeking);
        else
        if (buffered == 1.)
            status_image->set (yy_playing);

        status_label->set_markup ((boost::format ("Prebuffering: %d%%") % int(buffered * 100.)).str());
      }

      void
      MainUI::stream_position (int position)
      {
        using namespace Glib;

        if (m_seeking) return;

        if (m_popup && m_popup->is_visible()) m_popup->queue_draw ();

        int minutes = position / 60;
        int seconds = position % 60;
        int length  = ::play->property_length().get_value();

        time_label->set_markup ((f_time % minutes % seconds).str());
  
        if (seekrange->is_sensitive()) 
          {
            seekrange->set_range (0., (length > 0) ? double(length) : 1.); 
            seekrange->set_value ((position > 0) ? double(position): 0.);
          }
        else
          {
            return;
          }

        if (m_monitor && ((position > 240) || (position >= (length/2))) && (!m_sent) && (length > 30))
          {
            if (mcs->key_get<bool>("lastfm", "queue-enable") &&
                  ((m_active_source == SOURCE_ALBUMS) || (m_active_source == SOURCE_PLAYLIST)))
            {
              Glib::ustring uri (::play->property_stream().get_value());
              try {
                ::lastfm_service->send_song_information (uri);
                m_sent = true;
                }
              catch (...) {}
            }

            if (m_active_source == SOURCE_ALBUMS)
            {
              Glib::ustring uri (::play->property_stream().get_value());
              try {
                library->increase_play_count (uri);
                }
              catch (...) {}
            }
          }
      }

      bool
      MainUI::source_restore_context (GdkEventButton *event)
      {
        if (m_active_source == SOURCE_NONE) return false;
        treeview_sources->select_source (m_active_source);
        notebook_main->set_current_page (int(m_active_source));
        boost::shared_ptr<Bmp::PlaybackSource> &source = m_sources.find (m_active_source)->second;
        source->restore_context ();
        return false;
      }

      void
      MainUI::source_track_info (Glib::ustring title, SimpleTrackInfo sti, PlaybackSourceID source)
      {
        status_label->set_markup (title);
        display_popup (sti);
        sti.image.clear ();
        m_sti = sti;
        m_actions_main->get_action (MAIN_ACTION_RECOMMEND)->set_sensitive ((m_sti->artist && m_sti->album && m_sti->title));
      }

      void
      MainUI::source_caps (Bmp::PlaybackSource::Caps caps, PlaybackSourceID source)
      {
        m_source_caps[source] = caps;

        core->source_caps_changed (source, caps);

        if ((m_active_source == SOURCE_NONE) || (int(m_active_source) == int(notebook_main->get_current_page())))
          {
            m_actions_main->get_action (MAIN_ACTION_PAUSE)->set_sensitive (caps & PlaybackSource::CAN_PAUSE);
          }

        m_actions_main->get_action (MAIN_ACTION_PREV)->set_sensitive (caps & PlaybackSource::CAN_GO_PREV);
        m_actions_main->get_action (MAIN_ACTION_NEXT)->set_sensitive (caps & PlaybackSource::CAN_GO_NEXT);
        m_actions_main->get_action (MAIN_ACTION_PLAY)->set_sensitive (caps & PlaybackSource::CAN_PLAY);
      }

      void
      MainUI::source_playback_request (PlaybackSourceID source)
      {
        if (m_active_source != SOURCE_NONE)
          {
            m_sources.find (m_active_source)->second->stop();
          }

        m_active_source = source; 
        treeview_sources->set_active_source (source);
        ::play->property_status () = PLAYSTATUS_WAITING;
        ::play->property_stream() = m_sources.find (source)->second->get_uri();
        ::play->property_status () = PLAYSTATUS_PLAYING;
        m_actions_main->get_action (MAIN_ACTION_STOP)->set_sensitive (true);
      }

      void
      MainUI::source_changed (int n)
      {
        notebook_main->set_current_page (n);
        Sources::iterator s_iter = m_sources.find (PlaybackSourceID (n));
        if (s_iter != m_sources.end())
          {
            s_iter->second->send_caps ();
          }
        else
          {
            m_actions_main->get_action (MAIN_ACTION_PLAY)->set_sensitive (false);
            m_actions_main->get_action (MAIN_ACTION_PREV)->set_sensitive (false);
            m_actions_main->get_action (MAIN_ACTION_NEXT)->set_sensitive (false);
          }
      }

      void
      MainUI::play ()
      {
        PlaybackSourceID new_source = PlaybackSourceID (notebook_main->get_current_page());

        if (m_active_source != SOURCE_NONE)
          {
            m_sources.find (m_active_source)->second->stop();
            if (new_source != m_active_source)
              {
                status_image->set (yy_stopped);
                status_label->set_text (_("Changing Source.."));
                ::play->property_status () = PLAYSTATUS_STOPPED;
              }
          }
        else
          {
            status_image->set (yy_stopped);
            status_label->set_text (_("Changing Source.."));
          }

        m_active_source = new_source;
        treeview_sources->set_active_source (m_active_source);

        try {
          m_sources.find (m_active_source)->second->play();
          ::play->property_status () = PLAYSTATUS_WAITING;
          ::play->property_stream() = m_sources.find (m_active_source)->second->get_uri();
          ::play->property_status () = PLAYSTATUS_PLAYING;
          m_actions_main->get_action (MAIN_ACTION_STOP)->set_sensitive (true);
        } catch (Bmp::PlaybackSource::UnableToInitiatePlaybackError& cxe) {
          set_error (cxe.what(), m_active_source);
          m_active_source = SOURCE_NONE;
          treeview_sources->set_active_source (m_active_source);
          ::play->property_status () = PLAYSTATUS_STOPPED;
        }
      }

      void
      MainUI::pause ()
      {
        ::play->property_status () = PLAYSTATUS_PAUSED;
      }

      void
      MainUI::prev ()
      {
        boost::shared_ptr<Bmp::PlaybackSource> &source = m_sources.find (m_active_source)->second;
        PlaybackSource::Caps caps = m_source_caps.find (m_active_source)->second;

        if (caps & PlaybackSource::CAN_GO_PREV)
        {
            if (source->go_prev())
            {
              ::play->property_status () = PLAYSTATUS_WAITING;
              ::play->property_stream() = source->get_uri ();
              ::play->property_status () = PLAYSTATUS_PLAYING;
              m_sent = false;
              return;
            }
        }
        stop ();
      }
  
      void
      MainUI::next ()
      {
        boost::shared_ptr<Bmp::PlaybackSource> &source = m_sources.find (m_active_source)->second;
        PlaybackSource::Caps caps = m_source_caps.find (m_active_source)->second;
  
        if (caps & PlaybackSource::CAN_GO_NEXT)
        {
            if (source->go_next())
            {
              if (!(caps & PlaybackSource::PHONY_NEXT))
              {
                ::play->property_status () = PLAYSTATUS_WAITING;
                ::play->property_stream() = source->get_uri ();
                ::play->property_status () = PLAYSTATUS_PLAYING;
              }
              m_sent = false;
              return;
            }
        }
        stop ();
      }

      void
      MainUI::stop ()
      {
        if (m_active_source != SOURCE_NONE)
          {
            ::play->property_status () = PLAYSTATUS_WAITING;
            m_sources.find (m_active_source)->second->stop();
            ::play->property_status () = PLAYSTATUS_STOPPED;

            Sources::iterator s_iter = m_sources.find (PlaybackSourceID (notebook_main->get_current_page()));
            if (s_iter != m_sources.end())
              {
                s_iter->second->send_caps ();
              }
            else
              {
                m_actions_main->get_action (MAIN_ACTION_PLAY)->set_sensitive (false);
                m_actions_main->get_action (MAIN_ACTION_PAUSE)->set_sensitive (false);
                m_actions_main->get_action (MAIN_ACTION_PREV)->set_sensitive (false);
                m_actions_main->get_action (MAIN_ACTION_NEXT)->set_sensitive (false);
              }

            m_actions_main->get_action (MAIN_ACTION_RECOMMEND)->set_sensitive (false);
            m_sti.reset ();
          }
      }

      void
      MainUI::playstatus_changed ()
      {
        Playstatus status = Playstatus (::play->property_status().get_value());
        PlaybackSource::Caps caps = m_source_caps.find (m_active_source)->second;

        Bmp::URI u;

        switch (status)
        {
            case PLAYSTATUS_STOPPED:

              if (!m_error_condition)
                {
                  status_label->set_text ("");
                  status_image->set (yy_stopped);
                }

              if (m_popup)
                {
                  m_popup->hide ();
                  m_popup->set_source_icon (Glib::RefPtr<Gdk::Pixbuf>(0));
                }

              if (m_active_source != SOURCE_NONE)
                {
                  m_sources.find (m_active_source)->second->stop();
                }

              treeview_sources->set_active_source (SOURCE_NONE);
              m_active_source = SOURCE_NONE;

              bmp_status_icon_set_from_pixbuf (m_status_icon, m_status_icon_image_default->gobj());

              seekrange->set_sensitive (false);
              seekrange->set_range (0., 1.); 
              seekrange->set_value (double(0.));

              time_label->set_markup ("<b>--:--</b>");

              m_actions_main->get_action (MAIN_ACTION_RECOMMEND)->set_sensitive (false);
              m_actions_main->get_action (MAIN_ACTION_STOP)->set_sensitive (false);
              m_sti.reset ();

              m_sent = false;
#ifdef HAVE_GAIM
              m_actions_main->get_action (MAIN_ACTION_GAIM_SEND)->set_sensitive (false);
#endif // HAVE_GAIM
              break;

            case PLAYSTATUS_WAITING:
              bmp_status_icon_set_from_pixbuf (m_status_icon, m_status_icon_image_default->gobj());
              status_image->set (yy_stopped);
              break;

            case PLAYSTATUS_PLAYING:

              m_error_condition = false;

              if (caps & PlaybackSource::CAN_SEEK)
                {
                  seekrange->set_sensitive (true);
                }
              else
                {
                  seekrange->set_sensitive (false);
                  seekrange->set_range (0., 1.); 
                  seekrange->set_value (double(0.));
                }

              bmp_status_icon_set_from_pixbuf (m_status_icon, m_status_icon_image_playing->gobj());
              status_image->set (yy_playing);

              if (m_popup && (m_active_source != SOURCE_NONE))
                {
                  m_popup->set_source_icon 
                          (m_popup->render_icon (Gtk::StockID (source_icons_stock[m_active_source]),
                                                 Gtk::ICON_SIZE_SMALL_TOOLBAR));
                }

#ifdef HAVE_GAIM
              try {
                  u = Bmp::URI (::play->property_stream().get_value());
                  m_actions_main->get_action (MAIN_ACTION_GAIM_SEND)->set_sensitive
                          ((u.get_protocol() == Bmp::URI::PROTOCOL_FILE) && gaim_dbus->gaim_is_present()); 
                }
              catch (...) 
                {
                  m_actions_main->get_action (MAIN_ACTION_GAIM_SEND)->set_sensitive
                          (false);
                }
#endif
              u  = Bmp::URI (::play->property_stream().get_value());
              m_monitor = (u.get_protocol() == Bmp::URI::PROTOCOL_FILE);

              break;

            case PLAYSTATUS_PAUSED:
              if (caps & PlaybackSource::CAN_SEEK)
                {
                  seekrange->set_sensitive (false);
                }
              bmp_status_icon_set_from_pixbuf (m_status_icon, m_status_icon_image_paused->gobj());
              status_image->set (yy_paused);
              break;

            default: ;
        }
      }

      bool
      MainUI::on_window_state_event (GdkEventWindowState *event)
      {
        if ((event->changed_mask & GDK_WINDOW_STATE_ICONIFIED)
              && (event->new_window_state & GDK_WINDOW_STATE_ICONIFIED))
          {
            if (mcs->key_get<bool>("bmp", "ui-esc-trayconify"))
              {
                set_skip_taskbar_hint (true);
                set_skip_pager_hint (true);
                Gtk::Widget::hide ();
                m_visible = false;
              }
          }
          return false;
      }

      bool
      MainUI::on_key_press_event (GdkEventKey *event)
      {
        if (event->keyval == GDK_Escape)
          {
            if (mcs->key_get<bool>("bmp", "ui-esc-trayconify"))
              {
                set_skip_taskbar_hint (true);
                set_skip_pager_hint (true);
                Gtk::Widget::hide ();
                m_visible = false;
              }
            else
              {
                Gtk::Window::iconify ();
              }
            return true;
          }

        return Gtk::Widget::on_key_press_event (event);
      }

      MainUI*
      MainUI::create ()
      {
        const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "main-ui.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        MainUI *p = 0;
        glade_xml->get_widget_derived ("main-ui", p);
        return p; 
      }

      MainUI::~MainUI ()
      {
        g_object_unref (m_status_icon);
        delete mcs_bind;
      }

      GHashTable*
      MainUI::get_metadata_from_source (PlaybackSourceID source)
      {
        Sources::iterator s_iter = m_sources.find (source);
        if (s_iter == m_sources.end()) throw InvalidSourceError();

        SourceCaps::iterator c_iter = m_source_caps.find (source);
        if (c_iter == m_source_caps.end()) throw InvalidSourceError();


        if (c_iter->second & PlaybackSource::CAN_PROVIDE_METADATA)
        {
          GHashTable *m = s_iter->second->get_metadata ();
          if (!m) throw NoMetadataError();
          return m;
        }
        else
        {
          throw NoMetadataError();
        }
      }

      PlaybackSource::Caps
      MainUI::get_source_caps (PlaybackSourceID source)
      {
        SourceCaps::iterator s_iter = m_source_caps.find (source);
        if (s_iter == m_source_caps.end()) throw InvalidSourceError();
        return s_iter->second; 
      }

      PlaybackSourceID
      MainUI::get_current_source ()
      {
        return m_active_source;
      }

      void
      MainUI::play_uris (VFS::UriList& list, bool playback)
      {
        if (playback)
          {
            stop ();
            m_active_source = SOURCE_PLAYLIST;
            treeview_sources->set_active_source (SOURCE_PLAYLIST);
            treeview_sources->select_source (SOURCE_PLAYLIST);
            notebook_main->set_current_page (int(SOURCE_PLAYLIST));
            //FIXME: This code rather sucks ^^^
          }

        reinterpret_cast<Bmp::UIPart::Playlist*>(m_sources[SOURCE_PLAYLIST].get())->add_uris (list, playback);

        if (playback) 
          {
            m_sources[SOURCE_PLAYLIST]->play_requested();
            ::play->property_status () = PLAYSTATUS_WAITING;
            ::play->property_stream() = m_sources[SOURCE_PLAYLIST]->get_uri();
            ::play->property_status () = PLAYSTATUS_PLAYING;
            m_actions_main->get_action (MAIN_ACTION_STOP)->set_sensitive (true);
          } 
      }

      void
      MainUI::play_lastfm_uri (Glib::ustring const& lastfm_uri)
      {
        reinterpret_cast<Bmp::UIPart::LASTFM*>(m_sources[SOURCE_LASTFM].get())->set_uri (lastfm_uri);
        treeview_sources->select_source (SOURCE_LASTFM);
        notebook_main->set_current_page (int(SOURCE_LASTFM));
      }

      void
      MainUI::import_request (Bmp::VFS::UriList& list)
      {
        notebook_main->set_current_page (int(SOURCE_ALBUMS));
        reinterpret_cast<Bmp::UIPart::Albums*>(m_sources[SOURCE_ALBUMS].get())->import_external (list);
      }

      void
      MainUI::raise_ui ()
      {
        if (!m_visible)
        {
          m_visible = true;
          set_skip_taskbar_hint (false);
          set_skip_pager_hint (false);
        }
        Gtk::Window::present ();
      }

      /////////////////// Sources TreeView

      MainUI::TreeViewSources::TreeViewSources  (BaseObjectType                        *cobject,
                                                 const Glib::RefPtr<Gnome::Glade::Xml> &xml)
            : Gtk::TreeView (cobject),
              m_active_source (SOURCE_NONE)
      {
        store = Gtk::ListStore::create (columns);

        Gtk::TreeViewColumn *column = 0;
        Gtk::CellRendererPixbuf *cell_pixbuf = 0;
        Gtk::CellRendererText *cell_text = 0;

        cell_pixbuf = Gtk::manage (new Gtk::CellRendererPixbuf ());
        cell_pixbuf->property_ypad () = 2;
        cell_pixbuf->property_xalign () = 0.90;
        append_column ("", *cell_pixbuf);
        column = get_column (0);
        column->set_sizing (Gtk::TREE_VIEW_COLUMN_FIXED);
        column->add_attribute (*cell_pixbuf, "pixbuf", 0);
        column->set_resizable (false);
        column->set_expand (false);
        column->set_fixed_width (26);

        cell_text = Gtk::manage (new Gtk::CellRendererText ());
        append_column ("", *cell_text); 
        cell_text->property_ypad () = 2;
        column = get_column (1);
        column->add_attribute (*cell_text, "markup", 1);
        column->set_resizable (false);
        column->set_expand (false);
        column->set_cell_data_func (*cell_text, sigc::mem_fun (this, &Bmp::MainUI::TreeViewSources::playing_cell_data_func));

        get_selection()->set_mode (Gtk::SELECTION_BROWSE);  

        const char *names [] =
        {
          N_("Albums"),
          N_("Shoutcast"),
          N_("Icecast"),
          N_("Audio CD"),
          N_("Last.FM"),
          N_("Playlist"),
#ifdef HAVE_MOOD
          N_("Downloads"),
#endif
        };

        for (unsigned int n = 0; n < G_N_ELEMENTS(source_icons_stock); ++n)
        {
          using namespace Gtk;
          TreeModel::iterator m_iter = store->append (); 
          (*m_iter)[columns.icon] = render_icon (Gtk::StockID (source_icons_stock[n]), Gtk::ICON_SIZE_SMALL_TOOLBAR); 
          (*m_iter)[columns.name] = _(names[n]); 
        }

        set_model (store);
        conn = get_selection()->signal_changed().connect
            (sigc::mem_fun (this, &Bmp::MainUI::TreeViewSources::selection_changed));
      }

      MainUI::TreeViewSources::~TreeViewSources ()
      {
      }

      void
      MainUI::TreeViewSources::set_active_source (PlaybackSourceID source)
      {
        m_active_source = source;
        queue_draw ();
      }

      void
      MainUI::TreeViewSources::playing_cell_data_func (Gtk::CellRenderer* cell_, const Gtk::TreeModel::iterator& m_iter)
      {
        Gtk::CellRendererText *d = dynamic_cast<Gtk::CellRendererText *>(cell_);
        Gtk::TreePath path (store->get_path (m_iter));
        if ((m_active_source != SOURCE_NONE) && (path.get_indices().data()[0] == int (m_active_source)))
          {
            d->property_weight() = Pango::WEIGHT_BOLD; 
            return;
          }
        d->property_weight() = Pango::WEIGHT_NORMAL;
      }
     
      void
      MainUI::TreeViewSources::select_source (PlaybackSourceID source)
      {
        using namespace Gtk;

        TreeModel::Path path;
        path.append_index (int(source));
        conn.disconnect ();
        get_selection()->select (path);
        conn = get_selection()->signal_changed().connect (sigc::mem_fun (this, &Bmp::MainUI::TreeViewSources::selection_changed));
      }

      void
      MainUI::TreeViewSources::selection_changed ()
      {
        using namespace Gtk;

        TreeModel::iterator m_iter = get_selection()->get_selected ();
        TreeModel::Path path (m_iter);
        source_selected_.emit (path.get_indices().data()[0]);
      }


}
