//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_MAIN_UI_HH 
#define BMP_MAIN_UI_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtkmm.h>
#include <libglademm.h>
#include <boost/shared_ptr.hpp>
#include <boost/optional.hpp>
#include <mcs/mcs.h>

#include "playback_source.hh"
#include "popup.hh"
#include "vfs.hh"

#ifdef HAVE_GAIM
#  include "gaim-dbus.hh"
#endif

#include <widgets/bmp_tray_icon.h>
#include <widgets/bmp_status_icon.h>

#include "dialog-about.hh"
#include "preferences-ui.hh"

namespace Bmp
{
  class MainUI : public Gtk::Window
  {
    protected:

      virtual bool
      on_key_press_event (GdkEventKey*);

      virtual bool
      on_window_state_event (GdkEventWindowState *event);

    private:

      Glib::RefPtr<Gnome::Glade::Xml> m_ref_xml;
      Gtk::Notebook *notebook_main;

      Glib::RefPtr<Gtk::UIManager>   m_ui_manager_main;
      Glib::RefPtr<Gtk::ActionGroup> m_actions_main;

      typedef sigc::signal<void, int> SignalSourceSelected;
    
      class TreeViewSources : public Gtk::TreeView
      {
        public:

      	  TreeViewSources (BaseObjectType                       *cobject,
                       		 const Glib::RefPtr<Gnome::Glade::Xml> &xml);
          ~TreeViewSources ();

          void select_source (PlaybackSourceID source);
          void set_active_source (PlaybackSourceID source);
         
          SignalSourceSelected& 
          source_selected () { return source_selected_; }

        private:
          void selection_changed ();
          void playing_cell_data_func (Gtk::CellRenderer* cell_, const Gtk::TreeModel::iterator& m_iter);

          class Columns : public Gtk::TreeModel::ColumnRecord
          {
            public:

              Gtk::TreeModelColumn<Glib::RefPtr<Gdk::Pixbuf> > icon; 
              Gtk::TreeModelColumn<Glib::ustring> name;

              Columns ()
              {
                  add (icon); 
                  add (name);
              }
          };
          Columns columns;
          Glib::RefPtr<Gtk::ListStore> store;

          SignalSourceSelected      source_selected_;
          sigc::connection          conn; 

          PlaybackSourceID          m_active_source;
      };

      typedef std::map <PlaybackSourceID, Bmp::PlaybackSource::Caps> SourceCaps;
      typedef std::map <PlaybackSourceID, boost::shared_ptr<Bmp::PlaybackSource> > Sources;
      TreeViewSources                  *treeview_sources;
      PlaybackSourceID                  m_active_source;
      SourceCaps                        m_source_caps;
      Sources                           m_sources;
      boost::optional<SimpleTrackInfo>  m_sti;

      void display_about_dialog ();

      // Playback Controls
      void playstatus_changed ();
      void play ();
      void pause ();
      void prev ();
      void next ();
      void stop ();

      void stream_position (int position);
      void seek_display ();
      bool seek_event (GdkEvent *event);
    
      void buffering (double buffered);

      // Source Handling
      bool source_restore_context (GdkEventButton *event);
      void source_caps (Bmp::PlaybackSource::Caps caps, PlaybackSourceID source);
      void source_track_info (Glib::ustring title, SimpleTrackInfo sti, PlaybackSourceID source);
      void source_playback_request (PlaybackSourceID source);
      void source_changed (int n);
      bool clear_error (GdkEventButton*);
      void set_error (Glib::ustring message, PlaybackSourceID source);
  
      Glib::RefPtr<Gdk::Pixbuf>           yy_stopped;
      Glib::RefPtr<Gdk::Pixbuf>           yy_seeking;
      Glib::RefPtr<Gdk::PixbufAnimation>  yy_playing;
      Glib::RefPtr<Gdk::PixbufAnimation>  yy_paused;

      Glib::RefPtr<Gdk::Pixbuf>           m_warning;

      Gtk::Image    *status_image;
      Gtk::Label    *status_label;
      Gtk::Label    *time_label;
      Gtk::HScale   *seekrange;
      Gtk::EventBox *evb;
      Gtk::EventBox *status_evbox;

      Bmp::Popup                    *m_popup;

      BmpStatusIcon                 *m_status_icon;
      Glib::RefPtr<Gdk::Pixbuf>      m_status_icon_image_default;
      Glib::RefPtr<Gdk::Pixbuf>      m_status_icon_image_paused;
      Glib::RefPtr<Gdk::Pixbuf>      m_status_icon_image_playing;

      static void create_popup  (GtkPlug *plug, gpointer userdata);
      static void status_icon_click (BmpStatusIcon *icon, gpointer data);
      static void status_icon_scroll_up (BmpStatusIcon *icon, gpointer data);
      static void status_icon_scroll_down (BmpStatusIcon *icon, gpointer data);
      static void status_icon_popup_menu (BmpStatusIcon *icon, guint arg1, guint arg2, gpointer data);
      static gboolean status_icon_enter (BmpStatusIcon *icon, GdkEventCrossing *event, gpointer data); 
      static gboolean status_icon_leave (BmpStatusIcon *icon, GdkEventCrossing *event, gpointer data); 
      sigc::connection delay_popup;
      bool display_popup_delayed ();

      void display_popup (SimpleTrackInfo& info);
      void recommend ();

#ifdef HAVE_GAIM
      void gaim_send ();
      void gaim_present ();
      void gaim_lost ();

      boost::shared_ptr<Bmp::Gaim::DBUS> gaim_dbus;
#endif

      bool m_visible;
      bool m_seeking;
      bool m_sent;
      bool m_error_condition;
      bool m_monitor;

      // Additional UIs
      boost::shared_ptr<Bmp::Preferences> m_preferences;
      boost::shared_ptr<Bmp::AboutDialog> m_dialog_about;

      void import_request (Bmp::VFS::UriList& list);
      void enqueue_request (Bmp::VFS::UriList& list);

    public:

#include "exception.hh"

      EXCEPTION(InvalidSourceError)
      EXCEPTION(NoMetadataError)

      MainUI (BaseObjectType                       *cobject,
              const Glib::RefPtr<Gnome::Glade::Xml> &xml);
      ~MainUI ();
      static MainUI* create ();

      // Inquiries
      PlaybackSourceID      get_current_source ();
      PlaybackSource::Caps  get_source_caps (PlaybackSourceID source);
      GHashTable*           get_metadata_from_source (PlaybackSourceID source);

      // Actions
      void play_uris (VFS::UriList& list, bool playback);
      void play_lastfm_uri (Glib::ustring const& lastfm_uri);
      void raise_ui ();
  };
}
#endif
