//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "main.hh"
#include "vfs.hh"

#include "ui-toolbox.hh"
#include "main-ui.hh"
#include "service_core.hh"

#include <bmp/dbus.h>

#include <glibmm.h>
#include <cstring>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#define TYPE_C_SERVICE_CORE (CServiceCore::get_type ())
#define C_SERVICE_CORE(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), TYPE_C_SERVICE_CORE, CServiceCore))

namespace Bmp
{
  char const* mpris_identity = "BMP2";

  struct CServiceCoreClass
  {
    GObjectClass parent;
  };

  struct ServiceCore::CServiceCore
  {
    GObject parent;
    ServiceCore *core;

    enum
      {
        SIGNAL_STARTUP_COMPLETE,
        SIGNAL_SOURCE_CAPS_CHANGED,
        N_SIGNALS,
      };

    static guint signals[N_SIGNALS];

    static gpointer parent_class;

    static GType
    get_type ();

    static CServiceCore *
    create (ServiceCore & core);

    static void
    class_init (gpointer klass,
                gpointer class_data);

    static GObject *
    constructor (GType                   type,
                 guint                   n_construct_properties,
                 GObjectConstructParam * construct_properties);

    static gboolean
    identity (CServiceCore * self,
              char **        identity,
              GError **      error);

    static gboolean
    ui_raise (CServiceCore * self,
              GError **      error);

    static gboolean
    volume_set (CServiceCore * self,
                int            volume,
                GError **      error);

    static gboolean
    volume_get (CServiceCore * self,
                int *          volume,
                GError **      error);

    static gboolean
    startup (CServiceCore * self,
             GError **      error);

    static gboolean
    playlist_add_uri_list (CServiceCore * self,
                           char **        uris,
                           gboolean       playback,
                           GError **      error);

    static gboolean
    last_fm_play_uri (CServiceCore * self,
                      char *         uri,
                      GError **      error);

    static gboolean
    get_current_source (CServiceCore * self,
                        int *          source,
                        GError **      error);

    static gboolean
    get_source_caps (CServiceCore * self,
                     int            source,
                     int *          caps,
                     GError **      error);

    static gboolean
    get_metadata_from_source (CServiceCore * self,
                              int            source,
                              GHashTable **  metadata,
                              GError **      error);


    // Used Internally by ServiceCore
    static void
    initialize_dbus (CServiceCore * self);

    static void
    startup_complete (CServiceCore * self);

    static void
    source_caps_changed (CServiceCore *       self,
                         PlaybackSourceID     id,
                         PlaybackSource::Caps caps);
  };

  gpointer ServiceCore::CServiceCore::parent_class       = 0;
  guint    ServiceCore::CServiceCore::signals[N_SIGNALS] = { 0 };


  // HACK: Hackery to rename functions in glue
#define c_service_core_identity                 identity
#define c_service_core_ui_raise                 ui_raise
#define c_service_core_volume_get               volume_get
#define c_service_core_volume_set               volume_set
#define c_service_core_get_current_source       get_current_source
#define c_service_core_get_source_caps          get_source_caps
#define c_service_core_get_metadata_from_source get_metadata_from_source
#define c_service_core_playlist_add_uri_list    playlist_add_uri_list
#define c_service_core_last_fm_play_uri         last_fm_play_uri

#include "service_core_glue.h"
#include "bmp-marshalers.h"

  void
  ServiceCore::CServiceCore::class_init (gpointer klass,
                                         gpointer class_data)
  {
    parent_class = g_type_class_peek_parent (klass);

    GObjectClass *gobject_class = reinterpret_cast<GObjectClass*>(klass);
    gobject_class->constructor  = &CServiceCore::constructor;

    signals[SIGNAL_STARTUP_COMPLETE] =
      g_signal_new ("startup-complete",
                    G_OBJECT_CLASS_TYPE (G_OBJECT_CLASS (klass)),
                    GSignalFlags (G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED),
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

    signals[SIGNAL_SOURCE_CAPS_CHANGED] =
      g_signal_new ("source-caps-changed",
                    G_OBJECT_CLASS_TYPE (G_OBJECT_CLASS (klass)),
                    GSignalFlags (G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED),
                    0,
                    NULL, NULL,
                    bmpx_VOID__INT_INT, G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);
  }

  GObject *
  ServiceCore::CServiceCore::constructor (GType                  type,
                                          guint                  n_construct_properties,
                                          GObjectConstructParam *construct_properties)
  {
    GObject *object = G_OBJECT_CLASS (parent_class)->constructor (type, n_construct_properties, construct_properties);

    return object;
  }

  ServiceCore::CServiceCore *
  ServiceCore::CServiceCore::create (ServiceCore &core)
  {
    dbus_g_object_type_install_info (TYPE_C_SERVICE_CORE, &dbus_glib_c_service_core_object_info);

    CServiceCore *self = C_SERVICE_CORE (g_object_new (TYPE_C_SERVICE_CORE, NULL));
    self->core = &core;

    // Set up DBus connection
    ServiceCore::CServiceCore::initialize_dbus (self);

    return self;
  }

  GType
  ServiceCore::CServiceCore::get_type ()
  {
    static GType type = 0;

    if (G_UNLIKELY (type == 0))
      {
        static GTypeInfo const type_info =
          {
            sizeof (CServiceCoreClass),
            NULL,
            NULL,
            &class_init,
            NULL,
            NULL,
            sizeof (CServiceCore),
            0,
            NULL
          };

        type = g_type_register_static (G_TYPE_OBJECT, "CServiceCore", &type_info, GTypeFlags (0));
      }

    return type;
  }

  gboolean
  ServiceCore::CServiceCore::identity (CServiceCore * self,
                                       char **        identity,
                                       GError **      error)
  {
    *identity = g_strdup (mpris_identity);
    return TRUE;
  }

  gboolean
  ServiceCore::CServiceCore::ui_raise (CServiceCore * self,
                                       GError **      error)
  {
    self->core->main_ui->raise_ui();
    return TRUE;
  }

  gboolean
  ServiceCore::CServiceCore::volume_set (CServiceCore * self,
                                         int            volume,
                                         GError **      error)
  {
    mcs->key_set<int> ("bmp", "volume", volume);
    return TRUE;
  }

  gboolean
  ServiceCore::CServiceCore::volume_get (CServiceCore  *self,
                                         int *          volume,
                                         GError **      error)
  {
    *volume = mcs->key_get<int> ("bmp", "volume");
    return TRUE;
  }

  gboolean
  ServiceCore::CServiceCore::startup  (CServiceCore * self,
                                       GError **      error)
  {
    return TRUE;
  }

  gboolean
  ServiceCore::CServiceCore::last_fm_play_uri (CServiceCore * self,
                                               char *         uri,
                                               GError **      error)
  {
    self->core->main_ui->play_lastfm_uri (uri);
    return TRUE;
  }

  gboolean
  ServiceCore::CServiceCore::playlist_add_uri_list (CServiceCore * self,
                                                    char **        uris,
                                                    gboolean       playback,
                                                    GError **      error)
  {
    VFS::UriList list;
    VFS::VFS::strv_to_uri_list (list, uris, false);
    self->core->main_ui->play_uris (list, playback);
    return TRUE;
  }

  gboolean
  ServiceCore::CServiceCore::get_current_source (CServiceCore * self,
                                                 int *          source,
                                                 GError **      error)
  {
    *source = self->core->main_ui->get_current_source ();
    return TRUE;
  }

  gboolean
  ServiceCore::CServiceCore::get_source_caps (CServiceCore * self,
                                              int            source,
                                              int *          caps,
                                              GError **      error)
  {
      try
        {
          *caps = int (self->core->main_ui->get_source_caps (PlaybackSourceID (source)));
          return TRUE;
        }
      catch (MainUI::InvalidSourceError& cxe)
        {
          *error = g_error_new (g_quark_from_static_string ("BMP-CServiceCore"), 0, "Invalid Source Specified");
          return FALSE;
        }
      catch (...)
        {
          *error = g_error_new (g_quark_from_static_string ("BMP-CServiceCore"), 0, "Unknown Error");
          return FALSE;
        }
  }

  gboolean
  ServiceCore::CServiceCore::get_metadata_from_source (CServiceCore * self,
                                                       int            source,
                                                       GHashTable **  metadata,
                                                       GError **      error)
  {
      try
        {
          *metadata = self->core->main_ui->get_metadata_from_source (PlaybackSourceID (source));
          return TRUE;
        }
      catch (MainUI::InvalidSourceError & cxe)
        {
          *error = g_error_new (g_quark_from_static_string ("BMP-CServiceCore"), 0, "Invalid Source Specified");
          return FALSE;
        }
      catch (MainUI::NoMetadataError & cxe)
        {
          *error = g_error_new (g_quark_from_static_string ("BMP-CServiceCore"), 0, "No Metadata Available");
          return FALSE;
        }
      catch (...)
        {
          *error = g_error_new (g_quark_from_static_string ("BMP-CServiceCore"), 0, "Unknown Error");
          return FALSE;
        }
  }

  void
  ServiceCore::CServiceCore::startup_complete (CServiceCore *self)
  {
    g_signal_emit (self, signals[SIGNAL_STARTUP_COMPLETE], 0);
  }

  void
  ServiceCore::CServiceCore::source_caps_changed (CServiceCore *       self,
                                                  PlaybackSourceID     id,
                                                  PlaybackSource::Caps caps)
  {
    g_signal_emit (self, signals[SIGNAL_SOURCE_CAPS_CHANGED], 0, int (id), int (caps));
  }

  void
  ServiceCore::CServiceCore::initialize_dbus (CServiceCore *self)
  {
    DBusGConnection * session_bus;
    DBusGProxy *      session_bus_proxy;
    gboolean          connected = FALSE;
    guint             request_name_result;
    GError *          error = NULL;

    session_bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (!session_bus)
      {
        g_message ("Running: %s (DBUS Error: '%s')", G_OBJECT_TYPE_NAME(self), error->message);
        g_error_free (error);
        return;
      }

    session_bus_proxy = dbus_g_proxy_new_for_name (session_bus,
                                                   "org.freedesktop.DBus",
                                                   "/org/freedesktop/DBus",
                                                   "org.freedesktop.DBus");

    if (!dbus_g_proxy_call (session_bus_proxy, "RequestName", &error,
                            G_TYPE_STRING, BMP_DBUS_SERVICE,
                            G_TYPE_UINT, 0,
                            G_TYPE_INVALID,
                            G_TYPE_UINT, &request_name_result,
                            G_TYPE_INVALID))
      {
        g_error ("%s: Failed RequestName request: %s", G_STRFUNC, error->message);
        g_error_free (error);
        error = NULL;
      }

    switch (request_name_result)
      {
        case DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER:
          {
            dbus_g_connection_register_g_object (session_bus, BMP_DBUS_PATH, G_OBJECT(self));
            connected = true;
            break;
          }

        case DBUS_REQUEST_NAME_REPLY_EXISTS:
          {
            connected = false;
            break;
          }
        }

    if (connected)
      {
        g_message ("Running: %s (DBus:%s)", G_OBJECT_TYPE_NAME(self), BMP_DBUS_SERVICE);
      }
    else
      {
        g_message ("Running: %s", G_OBJECT_TYPE_NAME(self));
      }
  }

  ServiceCore::ServiceCore ()
    : Glib::ObjectBase (typeid(ServiceCore)),
      c_service_core   (0),
      main_ui          (0),
      msg_domain_id    (0)
  {
    c_service_core = CServiceCore::create (*this);
  }

  ServiceCore::~ServiceCore ()
  {
    if (c_service_core)
      g_object_unref (c_service_core);
  }

  void
  ServiceCore::ui_stop ()
  {
    delete main_ui;
  }

  void
  ServiceCore::ui_start ()
  {
    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Starting UI");
    main_ui = MainUI::create ();
    gtk_widget_set_style (GTK_WIDGET (main_ui->gobj()), Util::style_get());
    main_ui->present ();
    while (gtk_events_pending()) gtk_main_iteration ();
    CServiceCore::startup_complete (c_service_core);
    s_startup_complete_.emit ();
  }

  ServiceCore::MsgDomainId
  ServiceCore::message_domain_register (std::string const&   name,
                                        Glib::ustring const& desc)
  {
      UserMsgDomain domain;
      domain.name = name;
      domain.description = desc;

      msg_domains[msg_domain_id] = domain;

      return msg_domain_id++;
  }

  int
  ServiceCore::message_dispatch (MsgDomainId          id,
                                 Gtk::MessageType     type,
                                 Glib::ustring const& text)
  {
    MsgDomainTable::const_iterator match = msg_domains.find (id);

    if (match != msg_domains.end ())
      {
        UserMsgDomain const& domain = match->second;

        Gtk::ButtonsType buttons = (type == Gtk::MESSAGE_QUESTION) ? Gtk::BUTTONS_YES_NO : Gtk::BUTTONS_OK;

        Gtk::MessageDialog dialog (text, true, type, buttons, true);
        dialog.set_title (domain.description);
        dialog.set_position (Gtk::WIN_POS_CENTER);

        return dialog.run ();
      }
    else
      {
        return -1;
      }
  }

  void
  ServiceCore::source_caps_changed (PlaybackSourceID     id,
                                    PlaybackSource::Caps caps)
  {
    CServiceCore::source_caps_changed (c_service_core, id, caps);
  }

} // Bmp namespace
