//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm/i18n.h>
#include <glib/gstdio.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include <fstream>

#include <boost/format.hpp>
#include <boost/optional.hpp>
#include <boost/shared_ptr.hpp>

#include <mcs/mcs.h>

#include "main.hh"
#include "paths.hh"
#include "debug.hh"

#include "xds.hh"
#include "uri++.hh"
#include "mbxml.hh"
#include "stock.hh"
#include "util.hh"
#include "util-file.hh"

#include "x_library.hh"
#include "x_play.hh"
#include "x_vfs.hh"
#include "x_mcsbind.hh"

#ifdef HAVE_HAL
#  include "x_hal.hh"
#endif

#include "library-ui-modify.hh"
#include "library-ui-modify-import.hh"
#include "library-ui-relocate-album.hh"
#include "dialog-filelist.hh"

#include "ui-toolbox.hh"
#include "ui-part-albums.hh"
#include "taskdialog.hh"

using namespace Gtk;
using namespace Gdk;
using namespace Glib;

using namespace boost;
using namespace std;

using namespace Bmp::DB;
using namespace Bmp::Library;
using namespace Bmp::MusicBrainzXML;
using namespace Bmp::Util;
using namespace Bmp::VFS;

namespace
{
    const Datum search_data[] =
    {
      DATUM_ARTIST,
      DATUM_ALBUM,
      DATUM_TITLE,
      DATUM_GENRE,
    };

    const int search_default[] =
    {
      0, 1, 2,
    };

    static boost::format fint ("%d");
    static boost::format progress_text ("%d / %d");

    int n_mod_tracks = 0, n_mod_current = 0;

    void
    on_modify_start (int n_tracks, ProgressBar* bar)
    {
      n_mod_tracks = n_tracks;
      n_mod_current = 0;
      bar->show_all ();
      bar->set_fraction (0.0);
      bar->set_text ((progress_text % 0 % n_tracks).str());
    }

    void
    on_modify_step  (ProgressBar *bar)
    {
      n_mod_current++;
      bar->set_fraction ((n_mod_current*1.)/(n_mod_tracks*1.));
      bar->set_text ((progress_text % n_mod_current % n_mod_tracks).str());
      while (gtk_events_pending()) gtk_main_iteration ();
    }

    void
    on_modify_end   (ProgressBar *bar)
    {
      bar->set_fraction (0.0);
      bar->set_text ("");
      n_mod_tracks = 0;
      n_mod_current = 0;
    }

    void
    clear_entry (Gtk::Entry *entry)    
    {
      entry->set_text ("");
    }

    Gtk::Widget*
    get_popup (Glib::RefPtr<Gtk::UIManager> ui_manager,
               Glib::ustring                path) 
    {
      Gtk::Widget *menu_item;
      menu_item = ui_manager->get_widget (path);
      if ( menu_item )
          return dynamic_cast<Gtk::MenuItem*>(menu_item)->get_submenu ();
      else
          return 0;
    }
}

namespace Bmp
{
      AlbumView::AlbumView (BaseObjectType                  *cobject,
                            RefPtr<Gnome::Glade::Xml> const& xml)

            : TreeView                  (cobject),
              m_viewsize                (ViewSize (mcs->key_get<int>("bmp","library-view-size"))),
              m_selection_list_modified (false),
              m_selection_change_block  (false),
              m_show_only_new           (false),
              m_uid (0)
      {
          m_emblem_new  = Gdk::Pixbuf::create_from_file
                            (Glib::build_filename (BMP_IMAGE_DIR, "emblem-new.png"));
          m_disc_normal = Gdk::Pixbuf::create_from_file
                            (Glib::build_filename (BMP_IMAGE_DIR, "audio.png"))->scale_simple (64, 64, Gdk::INTERP_BILINEAR);
          m_disc_small  = Gdk::Pixbuf::create_from_file
                            (Glib::build_filename (BMP_IMAGE_DIR, "audio.png"))->scale_simple (32, 32, Gdk::INTERP_BILINEAR);

          enum Renderer
          {
            R_TEXT,
            R_PIXBUF,
          };

         struct { 
              char    *title;
              double   xalign;
              int      width;
              Renderer renderer;
          } headers_albums[] = {
              { _("New"),        0.5, -1,  R_PIXBUF },
              { _("Cover"),      0.5, -1,  R_PIXBUF },
              { _("Album"),      0.0, 250, R_TEXT   },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS(headers_albums); ++n)
          {
            switch (n)
            {
              case 0:
              {
                Gtk::CellRendererPixbuf *cell = manage (new CellRendererPixbuf());
                cell->property_xpad() = 2; 
                cell->property_ypad() = 2; 
                cell->property_yalign() = 0.0;
                cell->property_xalign() = 0.0; 
                append_column (headers_albums[n].title, *cell);
                TreeViewColumn *column = get_column (n);
                column->set_cell_data_func (*cell, sigc::mem_fun
                          (this, &Bmp::AlbumView::album_new_cell_data_func));
                column->set_resizable (false);
                column->set_expand (false);
                column->set_sizing (TREE_VIEW_COLUMN_FIXED);
                column->set_fixed_width (22);
                break;
              }

              case 1:
              {
                Gtk::CellRendererPixbuf *cell = manage (new CellRendererPixbuf());
                cell->property_xalign() = headers_albums[n].xalign;
                cell->property_xpad() = 0; 
                cell->property_ypad() = 2; 
                append_column (headers_albums[n].title, *cell);
                TreeViewColumn *column = get_column (n);
                column->set_cell_data_func (*cell, sigc::mem_fun
                          (this, &Bmp::AlbumView::album_cover_cell_data_func));
                column->set_resizable (false);
                column->set_expand (false);
                column->set_sizing (TREE_VIEW_COLUMN_FIXED);
                column->set_fixed_width (m_viewsize?36:68);
                break;
              }

              case 2:
              {
                m_cell_album_text = manage (new CellRendererText());
                m_cell_album_text->property_xalign() = 0.0; 
                m_cell_album_text->property_yalign() = 0.5; 
                m_cell_album_text->property_size() = (m_viewsize?8:14) * PANGO_SCALE;
                m_cell_album_text->property_width() = 700;
                append_column (headers_albums[n].title, *m_cell_album_text);
                TreeViewColumn *column  = get_column (n);
                column->set_cell_data_func (*m_cell_album_text, sigc::mem_fun
                          (this, &Bmp::AlbumView::album_title_cell_data_func));
                column->set_resizable (false);
                column->set_expand (false);
                column->set_sizing (TREE_VIEW_COLUMN_FIXED);
                column->set_fixed_width(headers_albums[n].width);
                break;
              }
            }
          }

          m_store_base = ListStore::create (albums);
          m_store_base->set_default_sort_func
                (sigc::mem_fun (*this, &Bmp::AlbumView::default_sort_func));
          m_store_base->set_sort_column (-1, SORT_ASCENDING);

          m_store_filter = TreeModelFilter::create (m_store_base);
          m_store_filter->set_visible_func
                (sigc::mem_fun (*this, &Bmp::AlbumView::row_is_visible));

          get_selection()->set_mode (SELECTION_MULTIPLE);
          get_selection()->set_select_function
                (sigc::mem_fun (*this, &Bmp::AlbumView::record_selection_change));

          set_model (m_store_filter);
      }

      void
      AlbumView::set_viewsize (ViewSize viewsize)
      {
        m_viewsize = viewsize;
        m_cell_album_text->property_size() = (m_viewsize?8:14) * PANGO_SCALE;
        TreeViewColumn *column = get_column (1);
        std::vector<Gtk::CellRenderer*> renderers = column->get_cell_renderers();
        renderers[0]->set_fixed_size (m_viewsize?36:68, m_viewsize?36:68);
        column->set_fixed_width (m_viewsize?36:68);
        column->queue_resize ();
        queue_draw ();
      }

      void
      AlbumView::new_approve_one (guint64 uid)
      {
        NewItemsMap::iterator i = m_new_items_map.find (uid);
        if (i != m_new_items_map.end())
        {
          Gtk::TreeModel::iterator m_iter = m_store_base->get_iter (i->second.get_path());
          Library::Album a = (*m_iter)[albums.album];
          a.new_item = false;
          (*m_iter)[albums.album] = a;
          m_new_items_map.erase (i);
        }
      }

      void
      AlbumView::new_remove_one (guint64 uid)
      {
        NewItemsMap::iterator i = m_new_items_map.find (uid);
        if (i != m_new_items_map.end())
        {
          Gtk::TreeModel::iterator m_iter = m_store_base->get_iter (i->second.get_path());
          Gtk::TreeModel::iterator p_iter = m_store_filter->convert_child_iter_to_iter (m_iter);
          if (p_iter)
            {
              get_selection()->unselect (p_iter);
            }
          m_new_items_map.erase (i);
        }
      }

      void
      AlbumView::new_approve_all ()
      {
        for (NewItemsMap::const_iterator i = m_new_items_map.begin(), e = m_new_items_map.end(); i !=e; ++i)
        {
          Gtk::TreeModel::iterator m_iter = m_store_base->get_iter (i->second.get_path());
          Library::Album a = (*m_iter)[albums.album];
          a.new_item = false;
          (*m_iter)[albums.album] = a;
        }
        m_new_items_map.clear();
      }

      void
      AlbumView::new_drop_all ()
      {
        for (NewItemsMap::const_iterator i = m_new_items_map.begin(), e = m_new_items_map.end(); i !=e; ++i)
        {
          Gtk::TreeModel::iterator m_iter = m_store_base->get_iter (i->second.get_path());
          TreeModel::Path const& c_path = m_store_base->get_path (m_iter); 
          TreeModel::Path const& m_path = m_store_filter->convert_child_path_to_path (c_path);
          if (m_path.gobj())
          {
            if (get_selection()->is_selected (m_path))
              {
                get_selection()->unselect (m_path);
              }
          }
          m_store_base->erase (m_iter);
        }
        m_new_items_map.clear();
      }

      void
      AlbumView::restore_selection ()
      {
        m_selection_change_block = true;

        for (SelectionList::iterator p = m_selection_list_albums.begin(),
                                     e = m_selection_list_albums.end(); p != e; ++p)
        {
          TreeModel::Path const& c_path = p->second.get_path();
          TreeModel::Path const& m_path = m_store_filter->convert_child_path_to_path (c_path);
          if (m_path.gobj())
          {
            if (!get_selection()->is_selected (m_path))
              {
                get_selection()->select (m_path);
              }
          }
        }
        m_selection_change_block = false;
      }

      void
      AlbumView::set_only_new_filter (bool show_only_new)
      {
        m_show_only_new = show_only_new;
        m_store_filter->refilter ();      
        restore_selection ();
      }

      void
      AlbumView::set_sort_filter (ustring const& sort_text)
      {
        m_sort = sort_text;
        m_store_filter->refilter ();      
        restore_selection ();
      }

      bool
      AlbumView::row_is_visible (TreeModel::const_iterator const& m_iter)
      {
        Album const& a = (*m_iter)[albums.album];

        if (m_show_only_new)
          {
            if (!a.new_item || !a.new_item.get()) return false;
          }

        return (m_sort.empty() || (a.artist && a.album && match_keys (a.artist.get() + " " + a.album.get(), m_sort)));
      }

      int
      AlbumView::default_sort_func (TreeModel::iterator const& iter_a, TreeModel::iterator const& iter_b)
      {
        return std::string((*iter_a)[albums.key]).compare(std::string((*iter_b)[albums.key]));
      }

      void
      AlbumView::album_new_cell_data_func (Gtk::CellRenderer* cell_, Gtk::TreeModel::iterator const& m_iter)
      {
        using namespace Gtk;
        using namespace Gdk;
        using namespace Glib;

        CellRendererPixbuf *cell = dynamic_cast<CellRendererPixbuf *>(cell_);

        Album const& a = ((*m_iter)[albums.album]);

        if (a.new_item && a.new_item.get())
          {
            cell->property_pixbuf() = m_emblem_new;
          }
        else
          {
            cell->property_pixbuf() = Glib::RefPtr<Gdk::Pixbuf>(0); 
          }
      }

      void
      AlbumView::album_cover_cell_data_func (Gtk::CellRenderer* cell_, Gtk::TreeModel::iterator const& m_iter)
      {
        Gtk::CellRendererPixbuf *cell =dynamic_cast<Gtk::CellRendererPixbuf *>(cell_);

        Glib::RefPtr<Gdk::Pixbuf> const& p = ((*m_iter)[albums.cover]);
        if (!p) return;

        cell->property_pixbuf() = m_viewsize ? p->scale_simple (32, 32, Gdk::INTERP_BILINEAR) : p;
      }

      void
      AlbumView::album_title_cell_data_func (Gtk::CellRenderer* cell_, Gtk::TreeModel::iterator const& m_iter)
      {
        Gtk::CellRendererText *cell =dynamic_cast<Gtk::CellRendererText *>(cell_);

        Album const& a = ((*m_iter)[albums.album]);

        if (!a.album) return;
      
        Glib::ustring title;
        title .append ("<b>")
              .append (Markup::escape_text (a.sortname ? a.sortname.get().c_str() : a.artist.get().c_str()))
              .append ("</b>\n")
              .append ("<i>") 
              .append (Markup::escape_text (a.album.get().c_str()))
              .append ("</i> <small>(")
              .append ((fint % a.tracks.get()).str())
              .append ((a.tracks.get() > 1) ? _(" Tracks") : _(" Track"))
              .append (")</small>");
        cell->property_markup() = title;
      }

      bool
      AlbumView::record_selection_change  (const RefPtr<TreeModel>& model,
                                           const TreeModel::Path&   path,
                                           bool                     selected)
      {
        if (m_selection_change_block) return true;

        RefPtr<TreeModelFilter> model_filter = RefPtr<TreeModelFilter>::cast_static (model);
        RefPtr<TreeModel> model_child = model_filter->get_model(); 
        TreeModel::Path const& c_path = model_filter->convert_path_to_child_path (path); 
        TreeModel::iterator const& m_iter = model_child->get_iter (c_path);

        if (!selected) // we're about to select
        {
          m_selection_list_albums.push_back (make_pair ((*m_iter)[albums.rowuid],
                                                        TreeModel::RowReference (model_child, c_path)));
        }
        else
        {
          unsigned int rowuid = (*m_iter)[albums.rowuid];
          for (AlbumView::SelectionList::iterator p = m_selection_list_albums.begin(),
                                                  e = m_selection_list_albums.end(); p != e; ++p)
          {
            if (p->first == rowuid) 
              {
                m_selection_list_albums.erase (p);
                break;
              }
          }
        }
        m_selection_list_modified = true;
        return true;
      }

      bool
      AlbumView::find_matching_album (Album const& album)
      {
        TreeModel::Children nodes = m_store_base->children ();
        TreeModel::iterator m_iter = nodes.begin (); 
        for (TreeModel::const_iterator m_end = nodes.end(); m_iter != m_end; ++m_iter)
          {
            if ((*m_iter)[albums.album] == album) return true;
          }
        return false;
      }

      bool
      AlbumView::find_matching_album (Album const& album,
                                      TreeModel::iterator & m_iter)
      {
        TreeModel::Children nodes = m_store_base->children ();
        m_iter = nodes.begin (); 
        for (TreeModel::const_iterator m_end = nodes.end(); m_iter != m_end; ++m_iter)
          {
            if ((*m_iter)[albums.album] == album) return true;
          }
        return false;
      }

      void
      AlbumView::update_new ()
      {
          RowV vector;
          library->get_albums (vector);
          for (RowV::iterator i = vector.begin(), e = vector.end(); i != e; ++i)
            {
              Album a (*i);

              if (find_matching_album (a)) continue;

              if (a.artist && a.album)
                {
                  AlbumFlags flags = AlbumView::A_NONE; 
                  TreeModel::iterator m_iter = m_store_base->append ();

                  (*m_iter)[albums.key] = ((a.sortname ? a.sortname.get() : a.artist.get())+a.album.get()).casefold_collate_key();
                  (*m_iter)[albums.album] = a;

                  m_new_items_map[m_uid] = Gtk::TreeModel::RowReference (m_store_base, m_store_base->get_path (m_iter));

                  try {
                        if (!a.asin) throw 0;
                        RefPtr<Pixbuf> cover;
                        Amazon::get_cover_by_asin (a.asin.get(), cover, true);
                        (*m_iter)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
                        flags = AlbumFlags (flags | AlbumView::A_HAS_COVER);
                    }
                  catch (...)
                    {
                        (*m_iter)[albums.cover] = m_disc_normal;
                    }

                  (*m_iter)[albums.flags]   = AlbumView::AlbumFlags (flags | (a.sortname ? AlbumView::A_HAS_SORTNAME : 0));
                  (*m_iter)[albums.rowuid]  = m_uid++;

                  while (gtk_events_pending ()) gtk_main_iteration ();
                }
            }
      }

      void
      AlbumView::update ()
      {
          m_selection_list_albums.clear();
          unset_model ();
          m_store_base->clear ();

          RowV vector;
          library->get_albums (vector);
          for (RowV::iterator i = vector.begin(), e = vector.end(); i != e; ++i)
            {
              Album a (*i);

              if (a.artist && a.album)
                {
                  AlbumFlags flags = AlbumView::A_NONE; 
                  TreeModel::iterator m_iter = m_store_base->append ();

                  (*m_iter)[albums.key] = ((a.sortname ? a.sortname.get() : a.artist.get())+a.album.get()).casefold_collate_key();
                  (*m_iter)[albums.album] = a;

                  m_new_items_map[m_uid] = Gtk::TreeModel::RowReference (m_store_base, m_store_base->get_path (m_iter));

                  try {
                        if (!a.asin) throw 0;
                        RefPtr<Pixbuf> cover;
                        Amazon::get_cover_by_asin (a.asin.get(), cover, true);
                        (*m_iter)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
                        flags = AlbumFlags (flags | AlbumView::A_HAS_COVER);
                    }
                  catch (...)
                    {
                        (*m_iter)[albums.cover] = m_disc_normal;
                    }

                  (*m_iter)[albums.flags]   = AlbumView::AlbumFlags (flags | (a.sortname ? AlbumView::A_HAS_SORTNAME : 0));
                  (*m_iter)[albums.rowuid]  = m_uid++;

                  while (gtk_events_pending ()) gtk_main_iteration ();
                }
            }

          set_model (m_store_filter);
      }

      bool
      AlbumView::albums_relocate ()
      {
        TreeSelection::ListHandle_Path paths = get_selection()->get_selected_rows ();
        if (paths.size () == 1)
        {
            TreeModel::iterator m_iter = m_store_filter->get_iter (*(paths.begin())); 
            shared_ptr<LibraryUIRelocateAlbum> dialog (LibraryUIRelocateAlbum::create ());

            Album const& a = (*m_iter)[albums.album];
            int response = dialog->run (a);

            dialog->hide ();

            if (response == RESPONSE_OK)
            {
                string path_new;
                ustring uri = dialog->fcb_move->get_current_folder_uri ();
                if (dialog->cb_append_albumname->get_active())
                {
                  string path = filename_from_uri (uri);
                  string append = filename_from_utf8 (a.album.get());
                  path_new = build_filename (path, append);
                  g_mkdir (path_new.c_str(), mode_t(S_IRWXU)); 
                }
                else
                {
                  path_new = filename_from_uri (uri);
                }

                TrackUpdatedV update_tracks;
                for (RowV::const_iterator i = dialog->vector.begin (), end = dialog->vector.end(); i != end; ++i)
                  {
                    try {
                        Track track (*i);

                        string filename = filename_from_uri (track.location.get());
                        string basename = path_get_basename (filename);
                        string pathname = path_get_dirname  (filename);

                        if (pathname != path_new) 
                          {
                            TrackUpdated update_track;
                            update_track.location_cur = track.location.get(); 
                            update_track.location_new = filename_to_uri (build_filename (path_new, basename));
#ifdef HAVE_HAL
                            update_track.volume_relative_path = track.volume_relative_path;
                            update_track.volume_udi = track.volume_udi;
#endif
                            update_tracks.push_back (update_track); 
                          }
                      }
                    catch (ConvertError& cxe)
                      {
                        //FIXME: We should convert all names beforehand so this will result in a mostly atomic procedure
                        MessageDialog message_dialog (_("Unable to proceed: A conversion error occured during processing"),
                                                         false, MESSAGE_ERROR, BUTTONS_OK, true);
                        message_dialog.run ();
                        return false;
                      }
                  }
                return library->modify_tracks (update_tracks, false);
              }
          }
          return false;
      }

      void
      AlbumView::get_album_list (AlbumList &list, ustring& label_text, bool& multiple)
      {
        if (get_selection()->count_selected_rows() == 0)
          {
            label_text = (_("(None Selected)"));
            return;
          }

        if (get_selection()->count_selected_rows() == 1)
          {
            std::vector<Gtk::TreeModel::Path> paths = get_selection()->get_selected_rows();
            TreeModel::iterator const& m_iter = m_store_filter->get_iter (paths[0]);
            Album const& a = (*m_iter)[albums.album];
            label_text  .append ("<b>")
                        .append (Markup::escape_text (a.sortname ? a.sortname.get() : a.artist.get()))
                        .append (":</b> <i>")
                        .append (Markup::escape_text (a.album.get()))
                        .append ("</i>");
            multiple = false;
          }
        else
          {
            label_text = (_("(Multiple Albums)"));
            multiple = true;
          }

        for (SelectionList::const_iterator p  = m_selection_list_albums.begin(),
                                           e  = m_selection_list_albums.end();
                                           p != e; ++p)
          {
            TreeModel::Path const& c_path = p->second.get_path ();
            TreeModel::Path m_path = m_store_filter->convert_child_path_to_path (c_path);
            if (m_path.gobj())
              {
                TreeModel::iterator const& m_iter = m_store_filter->get_iter (m_path);
                Album const& a = (*m_iter)[albums.album];
                list.push_back (a);
              }
          }
      }

#ifdef HAVE_HAL
      void
      AlbumView::albums_modify (PUDIVRPV const& list)
#else
      void
      AlbumView::albums_modify (VFS::UriList const& list)
#endif
      {
        MBReleaseV releases; 
        PathList paths = get_selection()->get_selected_rows ();
        for (PathList::const_iterator path = paths.begin(), end = paths.end(); path != end; ++path)
          {
            TreeModel::iterator const& m_iter = m_store_filter->convert_iter_to_child_iter (m_store_filter->get_iter (*path)); 
            Album const& a = (*m_iter)[albums.album];
            mb_query_releases (a.artist.get(), a.album.get(), releases);
          }

        shared_ptr<LibraryUIModify> modify (LibraryUIModify::create ());
        LibraryUIModify::ModifyTracks modify_tracks;

        if (modify->run (releases, list, modify_tracks) == GTK_RESPONSE_OK)
          {

            modify->hide ();

            if (modify_tracks.album.tracks.get() != list.size())
              {
                static boost::format msg_lt (_("The number of tracks in the selected release is only <b>'%s'</b>, while you "
                                                 "have selected <b>'%s'</b> "
                                                 "from your collection. Are you sure you want to proceed?\n\n<small>Note: This "
                                                 "might be actually valid "
                                                 "if you e.g. have duplicate tracks, but you might want to check again if this "
                                                 "is the correct release.</small>"));

                static boost::format msg_gt (_("The number of tracks in the selected release is <b>'%s'</b>, while you have "
                                                 "selected only <b>'%s'</b> from "
                                                 "your collection. Are you sure you want to proceed?\n\n<small>Note: This might "
                                                 "be actually valid if you e.g. "
                                                 "don't have all tracks from this album currently.</small>"));

                 MessageDialog dialog ((modify_tracks.album.tracks.get() < list.size())
                                        ? (msg_lt % modify_tracks.album.tracks % list.size()).str()
                                        : (msg_gt % modify_tracks.album.tracks % list.size()).str(),
                                        true, MESSAGE_QUESTION, BUTTONS_YES_NO, true);

                if (dialog.run() == RESPONSE_NO) return;
            }

            library->modify_tracks (modify_tracks.tracks);

            ReferenceList reference_list;
            for (PathList::const_iterator p = paths.begin (), e = paths.end (); p != e; ++p)
              {
                reference_list.push_back (TreeModel::RowReference (m_store_base, m_store_filter->convert_path_to_child_path (*p)));
              }

            for (ReferenceList::const_iterator r = reference_list.begin (), e = reference_list.end (); r != e; ++r)
              {
                TreeModel::iterator m_iter = m_store_base->get_iter (r->get_path());
                Album const& a ((*m_iter)[albums.album]); 
                if (library->album_exists (a))
                  {
                    Album b (library->get_album (a));
                    (*m_iter)[albums.album] = b; 
                    (*m_iter)[albums.key]   = Album::key (b); 
                  }
                else
                  {
                    get_selection()->unselect (m_store_filter->convert_child_iter_to_iter (m_iter));
                    new_remove_one ((*m_iter)[albums.rowuid]);
                    m_store_base->erase (m_iter);
                  }
              }

            TreeModel::iterator m_iter;
            AlbumFlags flags;

            Album a (library->get_album (modify_tracks.album));

            if (find_matching_album (a, m_iter))
              {
                flags = AlbumFlags ((*m_iter)[albums.flags]);
              }
            else
              {
                m_iter = m_store_base->append ();
                m_new_items_map[m_uid] = Gtk::TreeModel::RowReference (m_store_base, m_store_base->get_path (m_iter));
                (*m_iter)[albums.rowuid] = m_uid++;
                flags = AlbumFlags (0);
              }

            (*m_iter)[albums.album] = a;
            (*m_iter)[albums.key]   = Album::key (a); 
            (*m_iter)[albums.flags] = AlbumFlags (flags | (a.sortname ? A_HAS_SORTNAME : 0));

            if (!(flags & A_HAS_COVER))
              {
                try {
                    if (!modify_tracks.album.asin) throw 0;
                    RefPtr<Pixbuf> cover;
                    Amazon::get_cover_by_asin (modify_tracks.album.asin.get(), cover, false);
                    (*m_iter)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
                    (*m_iter)[albums.flags] = AlbumFlags (flags | A_HAS_COVER);
                  }
                catch (...)
                  {
                    (*m_iter)[albums.cover] = m_disc_normal;
                    (*m_iter)[albums.flags] = AlbumFlags (flags & ~A_HAS_COVER);
                  }
              }

            m_store_filter->clear_cache ();
            m_store_filter->refilter ();

            get_selection()->select (m_store_filter->convert_child_iter_to_iter (m_iter));
            scroll_to_row (m_store_filter->get_path (m_store_filter->convert_child_iter_to_iter (m_iter)), 0.5);
          }
      }

      void
      AlbumView::albums_remove (bool force)
      {
        ReferenceList reference_list;
        PathList paths = get_selection()->get_selected_rows ();

        ustring message;
        message.append (_("Do You Really Want to Remove the following Album(s) from the Library:\n"));
        for (PathList::const_iterator path = paths.begin (), end = paths.end (); path != end; ++path)
          {
            reference_list.push_back (TreeRowReference (m_store_base, m_store_filter->convert_path_to_child_path (*path)));
            TreeModel::iterator const& m_iter = m_store_filter->get_iter (*path);
            Album const & a ((*m_iter)[albums.album]); 
            message.append ("\n\""+Glib::Markup::escape_text (a.album.get())+"\"");
          }

        if (!force)
          {
            MessageDialog dialog (message, true, MESSAGE_ERROR, BUTTONS_YES_NO, true);
            dialog.set_title (_("Remove Albums - BMP"));
            if (dialog.run () != RESPONSE_YES) return;
          }
  
        set_sensitive (false);

        for (ReferenceList::const_iterator ref  = reference_list.begin (), end = reference_list.end (); ref != end; ++ref)
          {
            TreeModel::iterator const& m_iter = m_store_base->get_iter (ref->get_path());
            Album const& a ((*m_iter)[albums.album]); 

            AttributeV attrs; 
            album_to_attrs (a, attrs);

            RowV vector;
            library->query (attrs, vector);

            if (force)
            {
              if (!vector.empty())
              {
                Album b (library->get_album (a));
                (*m_iter)[albums.album] = b; 
                (*m_iter)[albums.key]   = Album::key (b); 
              }
              else
              {
                get_selection()->unselect (m_store_filter->convert_child_iter_to_iter (m_iter));
                new_remove_one ((*m_iter)[albums.rowuid]);
                m_store_base->erase (m_iter);
              }
            }
            else
            {
              get_selection()->unselect (m_store_filter->convert_child_iter_to_iter (m_iter));
              new_remove_one ((*m_iter)[albums.rowuid]);
              m_store_base->erase (m_iter);
              if (!vector.empty()) library->del (attrs);
            }
            while (gtk_events_pending()) gtk_main_iteration();
          }

        set_sensitive (true);
      }

      void
      AlbumView::scan_approved_sel ()
      {
        std::vector<TreeModel::Path> const& paths = get_selection()->get_selected_rows ();
        TreeModel::iterator const& m_iter = m_store_filter->get_iter (paths[0]);
        Album const& a = (*m_iter)[albums.album];
        library->approve_items (a);
        new_approve_one ((*m_iter)[albums.rowuid]);
        m_store_filter->row_changed (paths[0], m_iter);
      }
}

namespace Bmp
{
    TrackView::TrackView (BaseObjectType                  *cobject,
                          RefPtr<Gnome::Glade::Xml> const& xml)

          : TreeView  (cobject),
            playing   (Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_MAIN, "playing.png")))
    {
      const unsigned int N_RATINGS = 6;
      for (unsigned int n = 0; n < N_RATINGS; ++n)
        {
          static boost::format star ("stars_%d.png");
          rating[n] = Gdk::Pixbuf::create_from_file (build_filename (BMP_IMAGE_DIR_RATING, (star % n).str()));
        }

      enum Renderer
      {
        R_TEXT,
        R_PIXBUF,
      };

      struct { 
          char    *title;
          double   xalign;
          int      width;
          Renderer renderer;
          int      column;
      } headers_tracks[] = {
          { "",              0.5, 30,   R_PIXBUF,  -1},
          { "",              0.5, 30,   R_PIXBUF,  -1},
          { _("Track"),      1.0, 30,   R_TEXT,     0},
          { _("Title"),      0.0, 100,  R_TEXT,     1},
          { _("Rating"),     0.0, 80,   R_PIXBUF,   2},
          { _("Artist"),     0.0, 80,   R_TEXT,     3},
          { _("Album"),      0.0, 80,   R_TEXT,     4},
          { _("Genre"),      0.0, 80,   R_TEXT,     5},
          { _("Bitrate"),    0.0, 30,   R_TEXT,     6},
          { _("Samplerate"), 0.0, 30,   R_TEXT,     7},
      };

      for (unsigned int n = 0; n < G_N_ELEMENTS(headers_tracks); ++n)
      {
          TreeViewColumn *column = 0;
          CellRenderer *cell_ = 0;

          if (headers_tracks[n].renderer == R_PIXBUF)
          {
              CellRendererPixbuf *cell = manage (new CellRendererPixbuf());
              cell->property_xalign() = headers_tracks[n].xalign;
              append_column (headers_tracks[n].title, *cell);
              column = get_column (n);
              column->set_resizable (false);
              column->set_expand (false);
              column->property_min_width() = headers_tracks[n].width; 
              column->property_max_width() = headers_tracks[n].width; 
              cell_ = cell;

              switch (n)
                {
                    case 0: cell_new = cell; break;
                    case 1: cell_playing = cell; break;
                    case 4: cell_rating = cell; break;
                }
          }
          else 
          {
              CellRendererText *cell = manage ( new CellRendererText() );
              cell->property_xalign() = headers_tracks[n].xalign;
              append_column (headers_tracks[n].title, *cell);
              column = get_column (n);
              column->add_attribute (*cell, "text", headers_tracks[n].column);
              column->set_resizable (true);
              column->set_sort_column (headers_tracks[n].column);
              column->set_expand (true);
              column->property_min_width() = headers_tracks[n].width;
              cell_ = cell;
          }
        column->set_cell_data_func (*cell_, sigc::mem_fun (this, &Bmp::TrackView::cell_data_func_tracks));
      }

      store_tracks = ListStore::create (tracks);
      set_model (store_tracks);

      get_selection()->set_mode (SELECTION_MULTIPLE);

      get_selection()->set_select_function
            (sigc::mem_fun (this, &Bmp::TrackView::select_func_tracks));

      library->signal_row_updated().connect
            (sigc::mem_fun (this, &Bmp::TrackView::check_update));
    }

    void
    TrackView::notify_remove (ustring& uri)
    {
      if (playing_uri && playing_uri.get() == uri)
        {
          playing_uri.reset ();
          playing_row.reset ();
        }
    }

    void
    TrackView::check_update (ustring const&      location,
                             TrackUpdated track)
    {
      RowMapTracks::iterator rmt_iter = m_row_map_tracks.find (location); 
      if (rmt_iter != m_row_map_tracks.end())
        {

          set_sensitive (false);

          TreeModel::iterator m_iter = store_tracks->get_iter (rmt_iter->second.get_path());

          if (track.title) (*m_iter)[tracks.title] = track.title.get (); 
          if (track.artist) (*m_iter)[tracks.artist] = track.artist.get (); 
          if (track.album) (*m_iter)[tracks.album] = track.album.get (); 
          if (track.tracknumber) (*m_iter)[tracks.track] = track.tracknumber.get (); 
          if (track.genre) (*m_iter)[tracks.genre] = track.genre.get (); 
          if (track.location_new) (*m_iter)[tracks.location] = track.location_new.get (); 

          (*m_iter)[tracks.rating] = track.rating ? track.rating.get () : 0;

#ifdef HAVE_HAL
          if (track.volume_relative_path) (*m_iter)[tracks.volume_relative_path] = track.volume_relative_path.get (); 
          if (track.volume_udi) (*m_iter)[tracks.volume_udi] = track.volume_udi.get (); 
#endif

          if (track.location_new) 
          {
            (*m_iter)[tracks.present] = (((track.location_new && 
                                         file_test (filename_from_uri (track.location_new.get()),
                                         FILE_TEST_EXISTS)))); 

            TreeModel::Path path (store_tracks->get_path (m_iter)); 
            m_row_map_tracks.erase (location);
            m_row_map_tracks.insert (std::make_pair (track.location_new.get(),
                TreeRowReference (store_tracks, TreePath (m_iter))));
          }

          set_sensitive (true);
        }
    }

    bool
    TrackView::select_func_tracks (RefPtr<TreeModel> const& model,
                                   TreeModel::Path const& path,
                                   bool selected)
    {
      TreeModel::iterator m_iter = model->get_iter (path);
      return ((*m_iter)[tracks.present]);
    }

    void
    TrackView::cell_data_func_tracks (CellRenderer * cell_, TreeModel::iterator const& m_iter)
    {
      cell_->property_sensitive() = ((*m_iter)[tracks.present]); 

      if (cell_ == cell_rating)
        {
          CellRendererPixbuf *cell = dynamic_cast<CellRendererPixbuf *>(cell_);
          cell->property_pixbuf() = rating[((*m_iter)[tracks.rating])];
          return;
        }

      else

      if (cell_ == cell_playing)
        {
          CellRendererPixbuf * cell = dynamic_cast<CellRendererPixbuf *>(cell_);
          if (playing_uri && ((*m_iter)[tracks.location] == playing_uri.get()))
          {
            cell->property_pixbuf() = playing; 
            return;
          }
          cell->property_pixbuf() = RefPtr<Gdk::Pixbuf>(0);
        }

      else

      if (cell_ == cell_new)
        {
          CellRendererPixbuf * cell = dynamic_cast<CellRendererPixbuf *>(cell_);
          if ((*m_iter)[tracks.new_item])
          {
            cell->property_pixbuf() = render_icon (Gtk::StockID (BMP_STOCK_NEW), Gtk::ICON_SIZE_SMALL_TOOLBAR); 
            return;
          }
          cell->property_pixbuf() = RefPtr<Gdk::Pixbuf>(0);
        }
    }

    bool
    TrackView::on_event (GdkEvent * ev)
    {
      GdkEventButton *event = 0;

      if (ev->type == GDK_BUTTON_PRESS)
        {
          event = reinterpret_cast<GdkEventButton *>(ev);
          TreeViewColumn *column, *column_rating = get_column (4);
          TreeModel::Path path;
          int cell_x, cell_y;

          if (get_path_at_pos (int(event->x),
                               int(event->y),
                               path,
                               column,
                               cell_x,
                               cell_y)) 
            {
              if (column == column_rating)
                {
                  TreeModel::iterator m_iter = store_tracks->get_iter (path);

                  int n = (int(cell_x+7)/14);

                  TrackUpdatedV update_tracks;
                  TrackUpdated  update_track;

#ifdef HAVE_HAL
                  update_track.volume_udi = ((*m_iter)[tracks.volume_udi]); 
                  update_track.volume_relative_path = ((*m_iter)[tracks.volume_relative_path]); 
#endif
                  update_track.location_cur = ((*m_iter)[tracks.location]); 

                  update_track.rating = n;
                  update_tracks.push_back (update_track);
                  try {
                      library->modify_tracks (update_tracks, false, true, false);
                      (*m_iter)[tracks.rating] = n; 
                    }
                  catch (...) {}
                  return true;
                }
            } 
        }
      return false;
    }

    bool
    TrackView::remove ()
    {
      bool all_missing = true;
      ReferenceList refs;
      TreeSelection::ListHandle_Path paths = get_selection()->get_selected_rows ();

      for (TreeSelection::ListHandle_Path::const_iterator path = paths.begin (); path != paths.end (); ++path)
        {
          TreeModel::iterator m_iter = store_tracks->get_iter (*path);
          refs.push_back (TreeModel::RowReference (store_tracks, (*path)));
          if (bool((*m_iter)[tracks.present])) all_missing = false;
        }

      bool physically_delete_files = mcs->key_get<bool>("bmp", "physically-delete-files");
      if (physically_delete_files && !all_missing)
      {
        TaskDialog dialog (_("Music Library - BMP"),
                           _("Remove selected Files"),
                           Gtk::MESSAGE_QUESTION,
                           _("<b>You are about to delete the selected files also <i>physically</i> from "
                             "the filesystem. "
                             "If you do <i>NOT</i> wish that, please use 'Keep Files' now.</b>\n\n<small>NOTE: You "
                             "can change this "
                             "behaviour in Preferences, section 'Library'.</small>"));
                                                  
        dialog.add_button ("Remove Files", "Removes the selected files <b>irreversibely</b>", Gtk::Stock::GO_FORWARD, GTK_RESPONSE_OK);
        dialog.add_button ("Keep Files", "Keeps the files at this time", Gtk::Stock::CANCEL, GTK_RESPONSE_CANCEL);
        dialog.set_default_response (GTK_RESPONSE_CANCEL);

        if (dialog.run() != GTK_RESPONSE_OK)
        {
          dialog.hide ();
          return false;
        }
      }

      for (ReferenceList::const_iterator m_ref  = refs.begin (),
                                         m_end  = refs.end (); m_ref != m_end; ++m_ref)
        {
          TreeModel::iterator m_iter = store_tracks->get_iter (m_ref->get_path());
          ustring uri = (*m_iter)[tracks.location];
          notify_remove (uri);

#ifdef HAVE_HAL
          AttributeV attrs;
          attrs.push_back (Attribute (Bmp::DB::EXACT,
                                      Bmp::MetadatumId (DATUM_HAL_VOLUME_UDI), ustring((*m_iter)[tracks.volume_udi])));
          attrs.push_back (Attribute (Bmp::DB::EXACT,
                                      Bmp::MetadatumId (DATUM_VOLUME_RELATIVE_PATH), ustring((*m_iter)[tracks.volume_relative_path])));

          library->del (attrs);
          if (physically_delete_files && bool((*m_iter)[tracks.present]))
            {
              ustring uri = (*m_iter)[tracks.location];
              string filename = filename_from_uri (uri);
              if (file_test (filename, FILE_TEST_EXISTS))
                g_unlink (filename.c_str());    
            }
          store_tracks->erase (m_iter);
#else
          AttributeV attrs;
          attrs.push_back (Attribute (Bmp::DB::EXACT, Bmp::MetadatumId (DATUM_LOCATION), uri)); 

          library->del (attrs);
          if (physically_delete_files && bool((*m_iter)[tracks.present]))
            {
              string filename = filename_from_uri (uri);
              if (file_test (filename, FILE_TEST_EXISTS))
                g_unlink (filename.c_str());    
            }
          store_tracks->erase (m_iter);
#endif
        }
      return true;
    }

    void
    TrackView::add_tracks (Bmp::DB::RowV const& vector)
    {
      m_row_map_tracks.clear ();
  
      for (RowV::const_iterator v = vector.begin (), end = vector.end (); v != end; ++v) 
      { 

        try {

            Track track (*v);
            TreeModel::iterator m_iter = store_tracks->append ();
            TreeModel::Row row (*m_iter);

            if (track.title)        row[tracks.title]       = track.title.get(); 
            if (track.artist)       row[tracks.artist]      = track.artist.get(); 
            if (track.album)        row[tracks.album]       = track.album.get(); 
            if (track.tracknumber)  row[tracks.track]       = track.tracknumber.get(); 
            if (track.genre)        row[tracks.genre]       = track.genre.get(); 
            if (track.bitrate)      row[tracks.bitrate]     = track.bitrate.get(); 
            if (track.samplerate)   row[tracks.samplerate]  = track.samplerate.get(); 
            if (track.new_item)     row[tracks.new_item]    = track.new_item.get(); 

            if (track.location)
              {
                row[tracks.location] = track.location.get (); 
                if (playing_uri && (playing_uri.get() == track.location.get()))
                 {
                    TreeModel::Path path (store_tracks->get_path (m_iter));                    
                    playing_row = TreeModel::RowReference (store_tracks, path);
                  }
              }


#ifdef HAVE_HAL
            if (track.volume_relative_path)
              {
                row[tracks.volume_relative_path] = track.volume_relative_path.get(); 
              }

            if (track.volume_udi)
              {
                row[tracks.volume_udi] = track.volume_udi.get(); 
              }

            if (hal->is_initialized())
              {
                row[tracks.present] = (((track.location && file_test (filename_from_uri (track.location.get()), FILE_TEST_EXISTS)))); 
              }
#else
            row[tracks.present] = file_test (filename_from_uri (track.location.get()), FILE_TEST_EXISTS); 
                                       
#endif
            if (track.location)
              {
                m_row_map_tracks.insert (std::make_pair (track.location.get(), TreeModel::RowReference (store_tracks, TreePath (m_iter))));
              }

            row[tracks.rating] = track.rating ? track.rating.get () : 0;
          }
        catch (...) { continue; }
      }
    }

    ////////////////////////////////// Public API    

    bool
    TrackView::all_tracks_present ()
    {
      bool all = true;
      TreeModel::Children nodes = store_tracks->children ();
      for (TreeModel::Children::const_iterator m_iter = nodes.begin(), end = nodes.end(); m_iter != end; ++m_iter)
        {
          if (!((*m_iter)[tracks.present]))
            {
              all = false;
              break;
            }
        }
      return all;
    }

    void
    TrackView::get_files (Util::FileList& list)
    {
      TreeModel::Children nodes = store_tracks->children ();
      for (TreeModel::Children::const_iterator node  = nodes.begin (),
                                               end   = nodes.end (); 
                                               node != end; ++node)
        {
          ustring uri = (*node)[tracks.location];
          list.push_back (filename_from_uri (uri));
        }
    }

    void
    TrackView::clear ()
    {
      store_tracks->clear ();
      m_row_map_tracks.clear ();
      playing_row.reset ();
    }

    void
    TrackView::unsort ()
    {
      store_tracks->set_sort_column (TreeSortable::DEFAULT_UNSORTED_COLUMN_ID, SORT_ASCENDING);
    }

#ifdef HAVE_HAL
    void
    TrackView::get_tracks (PUDIVRPV& pairs)
    {
      TreeModel::Children nodes = store_tracks->children ();
      for (TreeModel::Children::const_iterator m_iter = nodes.begin(), end = nodes.end(); m_iter != end; ++m_iter)
        {
          ustring volume_udi = (*m_iter)[tracks.volume_udi];
          ustring volume_relative_path = (*m_iter)[tracks.volume_relative_path];
          pairs.push_back (std::make_pair (volume_udi, volume_relative_path));
        } 
    }
#else
    void
    TrackView::get_tracks (Bmp::VFS::UriList& locations)
    {
      TreeModel::Children nodes = store_tracks->children ();
      for (TreeModel::Children::const_iterator m_iter = nodes.begin(), end = nodes.end(); m_iter != end; ++m_iter)
        {
          locations.push_back ((*m_iter)[tracks.location]);
        }
    }
#endif

    ustring
    TrackView::playback_uri ()
    {
      return playing_uri.get();
    }

    void
    TrackView::notify_stop ()
    {
      if (playing_uri) playing_uri.reset ();
      if (playing_row) playing_row.reset ();
      queue_draw ();
    }

    bool
    TrackView::notify_next (unsigned int& position, unsigned int& size)
    {
      if (!playing_row) return false;

      TreeModel::Children const& children = store_tracks->children();
      TreeModel::Path path (playing_row.get().get_path());      

      size = children.size();

      if ((unsigned int)(path.get_indices().data()[0] + 1) != size) 
        {
          position = path.get_indices().data()[0] + 1;

          TreeModel::Path path_old (1, position-1);
          TreeModel::Path path_new (1, position);

          TreeModel::iterator m_iter = store_tracks->get_iter (path_new);
          playing_row = TreeModel::RowReference (store_tracks, path_new);
          playing_uri = (*m_iter)[tracks.location]; 

          store_tracks->row_changed (path_old, m_iter);
          store_tracks->row_changed (path_new, m_iter);

          return true;
        }
      return false;
    }

    bool
    TrackView::notify_prev (unsigned int& position, unsigned int& size)
    {
      if (!playing_row) return false;

      TreeModel::Children const& children = store_tracks->children();
      TreeModel::Path path (playing_row.get().get_path());      

      size = children.size();

      if ((path.get_indices().data()[0] - 1) >= 0) 
        {
          position = path.get_indices().data()[0] - 1;

          TreeModel::Path path_old (1, position+1);
          TreeModel::Path path_new (1, position);

          TreeModel::iterator m_iter = store_tracks->get_iter (path_new);
          playing_row = TreeModel::RowReference (store_tracks, path_new);
          playing_uri = (*m_iter)[tracks.location]; 

          store_tracks->row_changed (path_old, m_iter);
          store_tracks->row_changed (path_new, m_iter);
          return true;
        }
      return false;
    }

    void
    TrackView::notify_next_prev (bool &next, bool &prev)
    {
      next = false;
      prev = false;

      if (playing_row && playing_row.get().is_valid())
      {
          TreeModel::Children const& children = store_tracks->children();
          TreeModel::Path path (playing_row.get().get_path());      

          unsigned int position;
          position = path.get_indices().data()[0];
          prev = (position > 0);
          next = (position+1 != children.size());
      }
      else if (!playing_row || (playing_row && !playing_row.get().is_valid()))
      {
          playing_row.reset ();
          return;
      }
    }

    bool
    TrackView::notify_play (unsigned int& position, unsigned int& size)
    {
      if (get_selection()->count_selected_rows() == 0)
      {
        TreeModel::Path path; 
        path.append_index (0);
        get_selection()->select (path);
      }

      TreeSelection::ListHandle_Path paths = get_selection()->get_selected_rows ();
      TreeModel::Path const& path = *(paths.begin());
      TreeModel::iterator m_iter = store_tracks->get_iter (path);
      TreeModel::Children const& children = store_tracks->children();

      size = children.size();

      if (!playing_row) 
        {
          playing_row = TreeModel::RowReference (store_tracks, path);
          playing_uri = ((*m_iter)[tracks.location]);
          store_tracks->row_changed (path, m_iter);
          position = path.get_indices().data()[0]; 
          return true; 
        }
      else
        {
          TreeModel::iterator l_iter = store_tracks->get_iter (path);
          if ((*l_iter)[tracks.location] != playing_uri.get())
          {
            playing_row = TreeModel::RowReference (store_tracks, path);
            playing_uri = ((*l_iter)[tracks.location]);
            store_tracks->row_changed (path, l_iter);
            position = path.get_indices().data()[0]; 
            return true; 
          }
        }
      return false;
    }
}

namespace Bmp
{
  namespace UIPart
  {
      Albums::~Albums () {}

      Albums::Albums (RefPtr<Gnome::Glade::Xml> const& xml)
          : PlaybackSource (PlaybackSource::CAN_SEEK),
            m_ref_xml      (xml),
            m_updated      (false),
            m_busy         (false),
            m_init         (true)
      {
          struct ButtonActionProxyI
          {
            char  *widget;
            char  *action;
          };

          struct ImageStockIdI
          {
            char  *widget;
            char  *stockid;
          };

#define ACTION_IMPORT_MUSIC       "library-action-import-music"
#define ACTION_IMPORT_RETAG       "library-action-retag-import"

#define ACTION_EDIT_ALBUMS        "library-action-edit-album"
#define ACTION_IMPORT_TRACKS      "library-action-import-tracks"
#define ACTION_REMOVE_ALBUMS      "library-action-remove-albums"
#define ACTION_RELOCATE_ALBUM     "library-action-relocate-album"

#define ACTION_REMOVE_TRACKS      "library-action-remove-tracks"
#define ACTION_ENQUEUE_TRACKS     "library-action-enqueue-tracks"

#define ACTION_APPROVE_ITEMS      "library-action-approve-items"
#define ACTION_APPROVE_SELECTED   "library-action-approve-selected"
#define ACTION_DROP_ALL           "library-action-drop-all"
#define ACTION_SHOW_ONLY_NEW      "library-action-show-only-new"

#ifdef HAVE_HAL
          bool enabled = hal->is_initialized(); 
#else
          bool enabled = true;
#endif

          m_ui_manager            = UIManager::create ();
          m_actions[IMPORT]       = ActionGroup::create ("LibraryActionsImport"); 
          m_actions[ALBUM]        = ActionGroup::create ("LibraryActionsAlbum"); 
          m_actions[TRACKS]       = ActionGroup::create ("LibraryActionsPlayback"); 
          m_actions[APPROVE]      = ActionGroup::create ("LibraryActionsApprove"); 


          // IMPORT actiongroup
          m_actions[IMPORT]->add ( Action::create (ACTION_IMPORT_MUSIC,
                                           _("Import Music")),
                                           (sigc::bind (sigc::mem_fun (this, &Bmp::UIPart::Albums::import_tracks),
                                                       m_ref_xml->get_widget ("main-ui"))));

          m_actions[IMPORT]->add ( Action::create (ACTION_IMPORT_RETAG,
                                           _("Retag Files and Import")),
                                           (sigc::bind (sigc::mem_fun (this, &Bmp::UIPart::Albums::import_tracks_as_album),
                                                       m_ref_xml->get_widget ("main-ui"))));

          // ALBUM actiongroup
          m_actions[ALBUM]->add ( Action::create (ACTION_EDIT_ALBUMS,
                                          _("Edit Album Metadata")),
                                          sigc::mem_fun (this, &Bmp::UIPart::Albums::albums_modify));

          m_actions[ALBUM]->add ( Action::create (ACTION_REMOVE_ALBUMS,
                                          _("Remove Albums")),
                                          sigc::mem_fun (this, &Bmp::UIPart::Albums::on_albums_remove));

          m_actions[ALBUM]->add ( Action::create (ACTION_RELOCATE_ALBUM,
                                          _("Relocate Album")),
                                          sigc::mem_fun (this, &Bmp::UIPart::Albums::albums_relocate));

          // TRACKS actiongroup
          m_actions[TRACKS]->add ( Action::create ("dummy", "dummy", "dummy"));
          m_actions[TRACKS]->add ( Action::create (ACTION_REMOVE_TRACKS, Gtk::StockID (BMP_STOCK_DELETE),
                                           _("Remove Selected Tracks")),
                                           sigc::mem_fun (this, &Bmp::UIPart::Albums::remove_tracks));
          m_actions[TRACKS]->add ( Action::create (ACTION_ENQUEUE_TRACKS, Gtk::StockID (BMP_STOCK_ADD),
                                           _("Enqueue Selected Tracks To Playlist")),
                                           sigc::mem_fun (this, &Bmp::UIPart::Albums::enqueue_tracks));

          // APPROVE actiongroup
          m_actions[APPROVE]->add ( Action::create (ACTION_APPROVE_ITEMS,
                                            _("Approve All")),
                                            sigc::mem_fun (this, &Bmp::UIPart::Albums::approve_all));
          m_actions[APPROVE]->add ( Action::create (ACTION_APPROVE_SELECTED,
                                            _("Approve Selected")),
                                            sigc::mem_fun (this, &Bmp::UIPart::Albums::approve_sel));
          m_actions[APPROVE]->add ( Action::create (ACTION_DROP_ALL,
                                            _("Drop Remaining")),
                                            sigc::mem_fun (this, &Bmp::UIPart::Albums::drop_items));

          m_actions[APPROVE]->add ( ToggleAction::create (ACTION_SHOW_ONLY_NEW, ""));

          m_ui_manager->insert_action_group (m_actions[IMPORT]);
          m_ui_manager->insert_action_group (m_actions[ALBUM]);
          m_ui_manager->insert_action_group (m_actions[TRACKS]);
          m_ui_manager->insert_action_group (m_actions[APPROVE]);
          m_ui_manager->add_ui_from_file (Glib::build_filename (BMP_UI_DIR, "albums-tracklist.ui"));

          m_actions[ALBUM]->set_sensitive (false);

          m_actions[APPROVE]->get_action (ACTION_SHOW_ONLY_NEW)->connect_proxy
                (*(dynamic_cast<Gtk::ToggleButton *>(m_ref_xml->get_widget ("tb_show_only_new"))));

          m_actions[APPROVE]->get_action (ACTION_APPROVE_ITEMS)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_approve"))));

          m_actions[APPROVE]->get_action (ACTION_APPROVE_SELECTED)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_approve_selected"))));

          m_actions[APPROVE]->get_action (ACTION_DROP_ALL)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_drop"))));


          ////////////////////////////////////////////////////////////////////////////////////////////////////////////

          const ButtonActionProxyI _import[] =
          {
            { "b_import_tracks", ACTION_IMPORT_MUSIC },
            { "b_import_tracks_as_album", ACTION_IMPORT_RETAG },
          };
          for (unsigned int n = 0; n < G_N_ELEMENTS (_import); ++n)
          {
            m_actions[IMPORT]->get_action (_import[n].action)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget (_import[n].widget))));
            m_actions[IMPORT]->get_action (_import[n].action)->set_sensitive (enabled);
          }

          ////////////////////////////////////////////////////////////////////////////////////////////////////////////

          const ButtonActionProxyI _album[] =
          {
            { "b_albums_modify",  ACTION_EDIT_ALBUMS    },
            { "b_albums_remove",   ACTION_REMOVE_ALBUMS  },
            { "b_albums_relocate", ACTION_RELOCATE_ALBUM },
          };
          for (unsigned int n = 0; n < G_N_ELEMENTS (_album); ++n)
          {
            m_actions[ALBUM]->get_action (_album[n].action)->connect_proxy
                (*(dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget (_album[n].widget))));
            m_actions[ALBUM]->get_action (_album[n].action)->set_sensitive (enabled);
          }

          ////////////////////////////////////////////////////////////////////////////////////////////////////////////

          // Throbbers
          {
            dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("i_throbber-1"))->set
            (Glib::build_filename (BMP_IMAGE_DIR, BMP_THROBBER));

            dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("i_throbber-2"))->set
            (Glib::build_filename (BMP_IMAGE_DIR, BMP_THROBBER));
          }

          const ImageStockIdI _images[] =
          {
            { "i_add",
              BMP_STOCK_ADD_MUSIC    },
            { "i_add_tracks_album",
              BMP_STOCK_ADD_MUSIC    },
            { "i_remove_album",
              BMP_STOCK_REMOVE_MUSIC },
            { "i_import_tracks",
              BMP_STOCK_ADD_MUSIC    },
            { "i_tachikoma1",
              BMP_STOCK_EDIT_ALBUM   },
            { "i_tachikoma2",
              BMP_STOCK_MOVE_ALBUM   },
            { "i_display_new",
              BMP_STOCK_NEW          },
            { "i_approve",
              BMP_STOCK_APPROVE      },
            { "i_approve_selected",
              BMP_STOCK_APPROVE      },
            { "i_drop",
              BMP_STOCK_DELETE       },
            { "i_filter_clear",
              BMP_STOCK_ENTRY_CLEAR  },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS (_images); ++n)
          {
            Gtk::Image *image;
            m_ref_xml->get_widget (_images[n].widget, image);
            image->set (StockID (_images[n].stockid), ICON_SIZE_SMALL_TOOLBAR);
          }

          m_ref_xml->get_widget ("label_current_album", label_current_album);
          label_current_album->set_text("");

          m_ref_xml->get_widget ("cb_view_size", cb_view_size);
          mcs_bind->bind_cbox (cb_view_size, "bmp", "library-view-size"); 

          // Progress for modifications
          m_ref_xml->get_widget ("progress_modify", progress_modify);
          library->signal_modify_start (). connect (sigc::bind (sigc::ptr_fun (&on_modify_start), progress_modify));
          library->signal_modify_step  (). connect (sigc::bind (sigc::ptr_fun (&on_modify_step), progress_modify));
          library->signal_modify_end   (). connect (sigc::bind (sigc::ptr_fun (&on_modify_end), progress_modify));

          m_ref_xml->get_widget ("notebook_albums", notebook_albums);
          m_ref_xml->get_widget ("notebook_add", notebook_add);

          m_ref_xml->get_widget_derived ("view_tracks", m_view_tracks);
          m_view_tracks->signal_row_activated().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::activate_default));
          m_view_tracks->get_selection()->signal_changed().connect
            (sigc::mem_fun (this, &Bmp::UIPart::Albums::on_tracks_selection_changed));
          m_view_tracks->signal_event().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::popup_menu));

          m_ref_xml->get_widget_derived ("view_albums", m_view_albums); 
          m_view_albums->get_selection()->signal_changed().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::track_view_update));
          m_view_albums->get_selection()->signal_changed().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::on_albums_selection_changed));
          m_view_albums->signal_row_activated().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::activate_default));

          m_ref_xml->get_widget ("albums_filter", albums_filter);
          albums_filter->signal_changed().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::on_sort_entry_changed));

          m_action_show_only_new = RefPtr<ToggleAction>::cast_static(m_actions[APPROVE]->get_action (ACTION_SHOW_ONLY_NEW));
          m_action_show_only_new->signal_toggled().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::on_show_only_new_toggled));

#ifdef HAVE_HAL
          // Connect Bmp::HAL
          if (hal->is_initialized())
            {
              hal->signal_volume_added().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::hal_volume_added));
              hal->signal_volume_removed().connect
                (sigc::mem_fun (this, &Bmp::UIPart::Albums::hal_volume_removed));
            }
#endif

          dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_filter_clear"))->signal_clicked().connect
                  (sigc::bind (sigc::ptr_fun (&clear_entry), albums_filter));

          m_ref_xml->get_widget ("main-ui")->show ();

          mcs->subscribe ("AlbumsView", "bmp", "library-view-size",
                  sigc::mem_fun (this, &Bmp::UIPart::Albums::mcs_view_size_changed));

          m_init = false;
      }

#ifdef HAVE_HAL
      void
      Albums::hal_volume_removed (HAL::Volume volume)
      {
        debug ("library-ui", "HAL Volume Removed: %s", volume.volume_udi.c_str());
        track_view_update (); 
      }

      void
      Albums::hal_volume_added (HAL::Volume volume)
      {
        debug ("library-ui", "HAL Volume Added: %s", volume.volume_udi.c_str());
        track_view_update (); 
      }
#endif

      void
      Albums::activate_default (TreeModel::Path const& path, TreeViewColumn* column)
      {
        dynamic_cast<Gtk::Window *>(m_ref_xml->get_widget ("main-ui"))->activate_default ();
      }

      void
      Albums::mcs_view_size_changed (MCS_CB_DEFAULT_SIGNATURE)
      {
        m_view_albums->set_viewsize (AlbumView::ViewSize (boost::get<int>(value)));
      }
      
      void
      Albums::track_view_update ()
      {
        label_current_album->set_text ("");
        m_view_tracks->clear ();

        m_actions[ALBUM]->set_sensitive (false);
        m_actions[ALBUM]->get_action (ACTION_REMOVE_ALBUMS)->set_sensitive (false);
        m_actions[APPROVE]->get_action (ACTION_APPROVE_SELECTED)->set_sensitive (false);
  
        if (m_busy) return;

        AlbumList  list;
        ustring    label_text;
        bool       multiple;

        m_view_albums->get_album_list (list, label_text, multiple);
        label_current_album->set_markup (label_text);

        if (list.empty()) return;

        if (multiple) m_view_tracks->unsort ();

        for (AlbumList::const_iterator n  = list.begin(), e  = list.end(); n != e; ++n)
          {
            AttributeV a; 
            album_to_attrs (*n, a);

            RowV v;
            library->query (a, v);
            m_view_tracks->add_tracks (v);
          }

          m_actions[ALBUM]->set_sensitive (true);
          m_actions[ALBUM]->get_action (ACTION_REMOVE_ALBUMS)->set_sensitive (true);

          // Check State Of Tracks
          bool all_tracks_present = m_view_tracks->all_tracks_present();
          m_actions[ALBUM]->get_action (ACTION_EDIT_ALBUMS)->set_sensitive (all_tracks_present);

          if (list.size() == 1)
            {
#ifdef HAVE_HAL
              m_actions[ALBUM]->get_action(ACTION_RELOCATE_ALBUM)->set_sensitive
                        (hal->is_initialized() && all_tracks_present); 
#else
              m_actions[ALBUM]->get_action(ACTION_RELOCATE_ALBUM)->set_sensitive
                        (all_tracks_present);
#endif
            }
          else
            {
              m_actions[ALBUM]->get_action (ACTION_RELOCATE_ALBUM)->set_sensitive (false);
            }

            check_state_approve_selected ();
            check_nextprev_caps ();
      }

      void
      Albums::check_nextprev_caps ()
      {
          bool next, prev;
          m_view_tracks->notify_next_prev (next, prev);

          if (!next)
            m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);
          else
            m_caps = Caps (m_caps |  PlaybackSource::CAN_GO_NEXT);

          if (!prev)
            m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_PREV);
          else
            m_caps = Caps (m_caps |  PlaybackSource::CAN_GO_PREV);

          m_caps = Caps (m_caps & ~PlaybackSource::CAN_PLAY);
          s_caps_.emit (m_caps);
      }

      void
      Albums::drop_items ()
      {
        TaskDialog dialog (_("Music Library - BMP"),
                           _("Remove <i>all</i> remaining new items ?"),
                           Gtk::MESSAGE_QUESTION,
                           _("If you just want to remove one or several "
                             "albums use the regular 'Remove Album' option."));
                                                  
        dialog.add_button ("Remove Items", "Removes all remaining 'new' items", Gtk::Stock::GO_FORWARD, GTK_RESPONSE_OK);
        dialog.add_button ("Keep Items", "Keeps the items at this time", Gtk::Stock::CANCEL, GTK_RESPONSE_CANCEL);
        dialog.set_default_response (GTK_RESPONSE_CANCEL);

        if (dialog.run() == GTK_RESPONSE_OK)
          {
            m_ref_xml->get_widget ("main-ui")->set_sensitive (false);
            library->drop_items ();
            m_view_albums->new_drop_all ();
            m_ref_xml->get_widget ("main-ui")->set_sensitive (true);
          }
      }

      void
      Albums::approve_sel ()
      {
        m_ref_xml->get_widget ("main-ui")->set_sensitive (false);
        m_view_albums->scan_approved_sel ();

        m_actions[APPROVE]->get_action (ACTION_APPROVE_ITEMS)->set_sensitive (library->library_has_new_items()); 
        m_actions[APPROVE]->get_action (ACTION_DROP_ALL)->set_sensitive (library->library_has_new_items()); 

        track_view_update (); 
        m_ref_xml->get_widget ("main-ui")->set_sensitive (true);
      }

      void
      Albums::approve_all ()
      {
        m_ref_xml->get_widget ("main-ui")->set_sensitive (false);
        library->approve_items ();
        m_view_albums->new_approve_all ();
        m_action_show_only_new->set_active (false);

        m_actions[APPROVE]->get_action (ACTION_APPROVE_ITEMS)->set_sensitive (library->library_has_new_items()); 
        m_actions[APPROVE]->get_action (ACTION_DROP_ALL)->set_sensitive (library->library_has_new_items()); 

        m_ref_xml->get_widget ("main-ui")->set_sensitive (true);
      }

      void
      Albums::albums_remove (bool force)
      {
        m_view_albums->albums_remove (force);
        check_nextprev_caps ();
      }

      void
      Albums::albums_relocate ()
      {
        if (m_view_albums->albums_relocate()) track_view_update ();
      }

      void
      Albums::check_state_approve_selected ()
      {
        if (m_view_albums->get_selection()->count_selected_rows() == 1)
          {
            std::vector<TreeModel::Path> paths = m_view_albums->get_selection()->get_selected_rows ();
            Glib::RefPtr<TreeModel> const& model = m_view_albums->get_model();
            TreeModel::iterator const& m_iter = model->get_iter (paths[0]);
            Album const& a = (*m_iter)[m_view_albums->albums.album]; 
            m_actions[APPROVE]->get_action (ACTION_APPROVE_SELECTED)->set_sensitive (a.new_item.get());
          }
        else
          {
            m_actions[APPROVE]->get_action (ACTION_APPROVE_SELECTED)->set_sensitive (false);
          }
      }
  
      void
      Albums::albums_modify ()
      {
        Util::FileList list;
        m_view_tracks->get_files (list);
  
        Util::FileList non_writable;
        Util::files_writable (list, non_writable);
        if (!non_writable.empty ())
          {
            shared_ptr<DialogFilelist> dialog (DialogFilelist::create ());
            return void(dialog->run (non_writable));
          }

        if (!Util::files_taggable (list))
          {
            MessageDialog dialog (_("Unable to modify: Some of the selected/specified files are "
                                    "currently not taggable with BMP"),
                                    false, MESSAGE_ERROR, BUTTONS_OK, true);
            return void(dialog.run());
          }


        m_busy = true;
        m_actions[ALBUM]->set_sensitive (false);
        m_ref_xml->get_widget ("main-ui")->set_sensitive (false);

#ifdef HAVE_HAL
        Bmp::Library::PUDIVRPV l;
#else
        Bmp::VFS::UriList l;
#endif

        m_view_tracks->get_tracks (l);
        m_view_albums->albums_modify (l);

        m_actions[ALBUM]->set_sensitive (true);
        m_ref_xml->get_widget ("main-ui")->set_sensitive (true);
        m_busy = false;

        track_view_update ();
      }

      bool
      Albums::popup_menu (GdkEvent *ev)
      {
        using namespace Gtk;

        GdkEventButton *event = 0;
        if (ev->type == GDK_BUTTON_PRESS)
          {
            event = reinterpret_cast<GdkEventButton *>(ev);
            if ((event->button == 3) && (event->type == GDK_BUTTON_PRESS))
              {
                Menu *menu = dynamic_cast<Menu *>
                      (get_popup(m_ui_manager, "/popup-albums-tracklist/menu-albums-tracklist"));
                m_actions[TRACKS]->get_action (ACTION_REMOVE_TRACKS)->set_sensitive
                      (m_view_tracks->get_selection()->count_selected_rows() > 0);
                m_actions[TRACKS]->get_action (ACTION_ENQUEUE_TRACKS)->set_sensitive
                      (m_view_tracks->get_selection()->count_selected_rows() > 0);
                menu->popup (event->button, event->time);
                return true;
              }
          } 
        return false;
      }

      void
      Albums::remove_tracks ()
      {
        if (m_view_tracks->remove ())
          {
            albums_remove (true);
          }
      }

      void
      Albums::enqueue_tracks ()
      {
        using namespace Gtk;
        using namespace Glib;

        PathList const& paths (m_view_tracks->get_selection()->get_selected_rows());
        VFS::UriList list;
        for (PathList::const_iterator p = paths.begin(), e = paths.end(); p != e; ++p)
          {
            TreeModel::iterator m_iter = m_view_tracks->get_model()->get_iter (*p);
            list.push_back ((*m_iter)[m_view_tracks->tracks.location]);
          }
        s_uri_enqueue_request_.emit (list);
      }

      void
      Albums::import_external (VFS::UriList& urilist)
      {
        if (urilist.empty()) return;

        while (m_busy)
          {
            while (gtk_events_pending()) gtk_main_iteration();
          }

        // Check for writability
        Util::FileList list, non_writable;
        for (UriList::const_iterator iter  = urilist.begin (), end = urilist.end ();
                                     iter != end; ++iter)
          {
            list.push_back (filename_from_uri(*iter));
          }

        Util::files_writable (list, non_writable);
        if (!non_writable.empty ())
          {
            shared_ptr<DialogFilelist> dialog (DialogFilelist::create ());
            dialog->run (non_writable);
            return;
          }

        // Check for taggability
        if (!Util::files_taggable (list))
          {
            MessageDialog dialog ((*(dynamic_cast<Gtk::Window *>(m_ref_xml->get_widget ("main-ui")))),
                                  _("Unable to modify: Some of the selected/specified files are currently "
                                    "not taggable with BMP"),
                                     true, MESSAGE_ERROR, BUTTONS_OK, true);
            return void(dialog.run ());
          }

        shared_ptr<LibraryUIModifyImport> dialog (LibraryUIModifyImport::create());
        MBReleaseV releases; 

        if (dialog->run (releases) == RESPONSE_OK)
          {
            MBRelease const& selected = dialog->selected;
            dialog->hide();

            int n_mod_tracks = urilist.size();
            int n_mod_current = 0;
            progress_modify->set_fraction (0.0);
            progress_modify->set_text ((progress_text % 0 % n_mod_tracks).str());

            unsigned int new_items = 0;

            for (UriList::const_iterator iter = urilist.begin (), end = urilist.end (); iter != end; ++iter)
              {
                TrackUpdated update_track;

                update_track.mb_artist_id = selected.artist_id;
                update_track.artist = selected.artist_name;
                update_track.mb_album_id = selected.release_id;
                update_track.album = selected.release_title;

                if (selected.asin) update_track.asin = selected.asin;
            
                try {
                    library->metadata_set_taglib ((*iter), update_track); 
                    new_items |=library->add (*iter);
                  }
                catch (Bmp::Library::Exception& cxe)
                  {
                    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Library Error: %s", cxe.what());
                  }

                n_mod_current++;
                progress_modify->set_fraction ((n_mod_current*1.)/(n_mod_tracks*1.));
                progress_modify->set_text ((progress_text % n_mod_current % n_mod_tracks).str());

                while (gtk_events_pending()) gtk_main_iteration ();
              }

            progress_modify->set_fraction (0.0);
            progress_modify->set_text ("");

            if (!new_items) return;

            m_action_show_only_new->set_active ();
            albums_update ();
          }
      }

      void
      Albums::import_tracks_as_album (Gtk::Widget *widget)
      {
        FileChooserDialog dialog (_("Select Tracks to Import as Album - BMP"), FILE_CHOOSER_ACTION_OPEN);
        dialog.set_select_multiple (true);
        dialog.add_button (StockID (Stock::CANCEL), RESPONSE_CANCEL);
        dialog.add_button (StockID (Stock::OPEN), RESPONSE_OK);
        dialog.set_default_response (RESPONSE_CANCEL);
        dialog.set_current_folder (mcs->key_get<string>("bmp", "file-chooser-path"));

        if (dialog.run () == RESPONSE_OK)
          {
            dialog.hide ();
            mcs->key_set<string>("bmp", "file-chooser-path", dialog.get_current_folder());

            SListHandle<ustring> handle = dialog.get_uris (); 
            UriList urilist (handle.begin(), handle.end());

            if (urilist.empty()) 
              {
                return;
              }

            // Check for writability
            Util::FileList list, non_writable;
            for (UriList::const_iterator iter  = urilist.begin (), end = urilist.end ();
                                         iter != end; ++iter)
              {
                list.push_back (filename_from_uri(*iter));
              }

            Util::files_writable (list, non_writable);
            if (!non_writable.empty ())
              {
                shared_ptr<DialogFilelist> dialog (DialogFilelist::create ());
                dialog->run (non_writable);
                return;
              }

            // Check for taggability
            if (!Util::files_taggable (list))
              {
                MessageDialog dialog (_("Unable to modify: Some of the selected/specified files are currently "
                                             "not taggable with BMP"),
                                             true, MESSAGE_ERROR, BUTTONS_OK, true);
                dialog.run ();
                return;
              }

            shared_ptr<LibraryUIModifyImport> dialog (LibraryUIModifyImport::create());
            MBReleaseV releases; 

            if (dialog->run (releases) == RESPONSE_OK)
              {
                MBRelease const& selected = dialog->selected;
                dialog->hide();
                widget->set_sensitive (false);

                int n_mod_tracks = urilist.size();
                int n_mod_current = 0;
                progress_modify->set_fraction (0.0);
                progress_modify->set_text ((progress_text % 0 % n_mod_tracks).str());

                unsigned int new_items = 0;

                for (UriList::const_iterator  iter  = urilist.begin (), end = urilist.end ();
                                              iter != end; ++iter)
                  {
                    TrackUpdated update_track;

                    update_track.mb_artist_id = selected.artist_id;
                    update_track.artist = selected.artist_name;
                    update_track.mb_album_id = selected.release_id;
                    update_track.album = selected.release_title;

                    if (selected.asin) update_track.asin = selected.asin;
                
                    try {
                        library->metadata_set_taglib ((*iter), update_track); 
                        new_items |=library->add (*iter);
                      }
                    catch (Bmp::Library::Exception& cxe)
                      {
                        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Library Error: %s", cxe.what());
                      }

                    n_mod_current++;
                    progress_modify->set_fraction ((n_mod_current*1.)/(n_mod_tracks*1.));
                    progress_modify->set_text ((progress_text % n_mod_current % n_mod_tracks).str());

                    while (gtk_events_pending()) gtk_main_iteration ();
                  }

                progress_modify->set_fraction (0.0);
                progress_modify->set_text ("");

                widget->set_sensitive (true);

                if (!new_items) return;

                m_action_show_only_new->set_active ();
                albums_update ();
              }
          }
      }

      void
      Albums::import_tracks (Gtk::Widget *widget)
      {
        FileChooserDialog dialog (_("Select Path To Add - BMP"),
                                  FILE_CHOOSER_ACTION_SELECT_FOLDER);

        dialog.add_button (Stock::CANCEL, RESPONSE_CANCEL);
        dialog.add_button (Stock::ADD, RESPONSE_OK);
        dialog.set_current_folder (mcs->key_get<string>("bmp", "file-chooser-path"));
        dialog.set_default_response (RESPONSE_CANCEL);
        dialog.set_default_size (750, 570);
        dialog.set_position (WIN_POS_CENTER);

        if (dialog.run () == RESPONSE_OK)
          {
            mcs->key_set<string>("bmp", "file-chooser-path", dialog.get_current_folder());
            notebook_add->set_current_page (1);
            ustring folder = dialog.get_filename ();
            dialog.hide ();
            widget->set_sensitive (false);

            Util::FileList list;
            Util::collect_path (folder, list);
            unsigned int new_items = 0;
            for (Util::FileList::const_iterator iter  = list.begin (),
                                                end   = list.end ();
                                                iter != end; ++iter)
              {
                try {
                    debug ("library-ui", "Adding URI '%s' to Library", (*iter).c_str()); 
                    new_items |= library->add (filename_to_uri (*iter));
                    while (gtk_events_pending ()) gtk_main_iteration ();
                  }
                catch (ConvertError& cxe) {} // FIXME:
              }

            notebook_add->set_current_page (0);
            widget->set_sensitive (true);
            
            if (!new_items)
              {
                track_view_update ();
              }
            else
              {
                m_action_show_only_new->set_active ();
                albums_update ();
              }
          }
      }

      void
      Albums::albums_update (bool update_new)
      {
          if (m_init) return;

          notebook_albums->set_current_page (1);
          cb_view_size->set_sensitive (false);
          albums_filter->set_sensitive (false);

          if (!update_new)
            {
              m_view_tracks->clear ();
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_PREV);
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_PLAY);
              s_caps_.emit  (m_caps);
            }

          m_actions[IMPORT]->set_sensitive (false);
          m_actions[APPROVE]->set_sensitive (false);
          m_actions[ALBUM]->set_sensitive (false);

          while (gtk_events_pending()) gtk_main_iteration ();

          m_busy = true;
          if (update_new)
            m_view_albums->update_new();
          else
            m_view_albums->update();
          m_busy = false;

          m_actions[APPROVE]->set_sensitive (true);
          m_actions[APPROVE]->get_action (ACTION_APPROVE_ITEMS)->set_sensitive (library->library_has_new_items()); 
          m_actions[APPROVE]->get_action (ACTION_DROP_ALL)->set_sensitive (library->library_has_new_items()); 
          m_actions[APPROVE]->get_action (ACTION_APPROVE_SELECTED)->set_sensitive (false);

#ifdef HAVE_HAL
          m_actions[IMPORT]->set_sensitive (hal->is_initialized());
#else
          m_actions[IMPORT]->set_sensitive (true);
#endif

          m_actions[ALBUM]->set_sensitive (true);
          m_actions[ALBUM]->get_action (ACTION_EDIT_ALBUMS)->set_sensitive (m_view_albums->get_selection()->count_selected_rows());
          m_actions[ALBUM]->get_action (ACTION_REMOVE_ALBUMS)->set_sensitive (m_view_albums->get_selection()->count_selected_rows());
          m_actions[ALBUM]->get_action (ACTION_RELOCATE_ALBUM)->set_sensitive (m_view_albums->get_selection()->count_selected_rows() == 1);

          cb_view_size->set_sensitive (true);
          albums_filter->set_sensitive (true);
          notebook_albums->set_current_page (0);
      }

      void
      Albums::on_albums_remove ()
      {
        albums_remove (false);
      }

      void
      Albums::on_sort_entry_changed ()
      {
        m_view_albums->set_sort_filter (albums_filter->get_text());
        track_view_update ();
      }

      void
      Albums::on_show_only_new_toggled ()
      {
        m_view_albums->set_only_new_filter (m_action_show_only_new->get_active());
        check_state_approve_selected ();
        track_view_update ();
      }

      // Bmp::PlaybackSource
      ustring
      Albums::get_uri ()
      {
        return m_view_tracks->playback_uri ();
      }

      GHashTable*
      Albums::get_metadata ()
      {
        return library->get_hashtable (::play->property_stream().get_value());
      }
      
      void
      Albums::send_title ()
      {
        Track track;
        library->get_metadata (m_view_tracks->playback_uri(), track);

        oustring eartist, ealbum, etitle;
        ustring artist, album, title;

        SimpleTrackInfo sti;

        if (track.artist)
          {
            eartist = Markup::escape_text (track.artist.get()); 
            sti.artist = track.artist;
          }

        if (track.album)
          {
            ealbum  = Markup::escape_text (track.album.get()); 
            sti.album = track.album;
          }

        if (track.title)
          {
            etitle  = Markup::escape_text (track.title.get()); 
            sti.title = track.title;
          }

        if (track.asin) sti.asin = track.asin;
        if (track.duration) sti.duration = track.duration;

        if (etitle)
            title.append ("<b>").append(etitle.get()).append("</b>");

        if (ealbum)
            album.append ("<i>").append(ealbum.get()).append("</i>"); 

        if (track.date && (track.date.get() > 0))
          {
            album.append(", ").append((fint % track.date.get()).str()); 
          }

        char *text = g_strdup_printf ("%s (%s%s%s)", (etitle  ? title.c_str()    : ""),
                                                     (eartist ? eartist.get().c_str()  : ""),
                                                     ((!eartist || (!ealbum))) ? "" : ": ",
                                                     (ealbum  ? album.c_str() : ""));

        s_track_info_.emit (ustring(text), sti); 
        g_free (text);
      }

      bool
      Albums::go_next ()
      {
        unsigned int position, size;

        if (m_view_tracks->notify_next (position, size))
          {
            if (position+1 != size)
              m_caps = Caps (m_caps |  PlaybackSource::CAN_GO_NEXT);
            else
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);

            if (position > 0) 
              m_caps = Caps (m_caps |  PlaybackSource::CAN_GO_PREV);
            else
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_PREV);

            s_caps_.emit (m_caps);
            send_title ();    
            return true;
          }
        return false;
      }

      bool
      Albums::go_prev ()
      {
        unsigned int position, size;
        if (m_view_tracks->notify_prev (position, size))
          {
            if (position+1 != size)
              m_caps = Caps (m_caps |  PlaybackSource::CAN_GO_NEXT);
            else
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);

            if (position > 0) 
              m_caps = Caps (m_caps |  PlaybackSource::CAN_GO_PREV);
            else
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_PREV);

            s_caps_.emit (m_caps);
            send_title ();    
            return true;
          }
        return false;
      }

      void
      Albums::stop ()
      {
        m_view_tracks->notify_stop ();

        m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);
        m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_PREV);
        m_caps = Caps (m_caps & ~PlaybackSource::CAN_PAUSE);
        m_caps = Caps (m_caps & ~PlaybackSource::CAN_PROVIDE_METADATA);

        s_caps_.emit (m_caps);
      }

      void
      Albums::play ()
      {
        m_view_albums->save_context ();
        m_view_tracks->save_context ();

        m_saved_context_filter = albums_filter->get_text ();
        
        unsigned int position, size;
        if (m_view_tracks->notify_play (position, size))
          {
            if (position+1 != size)
              m_caps = Caps (m_caps |  PlaybackSource::CAN_GO_NEXT);
            else
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);

            if (position > 0) 
              m_caps = Caps (m_caps |  PlaybackSource::CAN_GO_PREV);
            else
              m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_PREV);
        }

        m_caps = Caps (m_caps | PlaybackSource::CAN_PAUSE);
        m_caps = Caps (m_caps | PlaybackSource::CAN_PROVIDE_METADATA);
        s_caps_.emit (m_caps);
        send_title ();
      }

      void
      Albums::restore_context ()
      {
        if (m_saved_context_filter) albums_filter->set_text (m_saved_context_filter.get());
        m_view_tracks->restore_context ();
        m_view_albums->restore_context ();
      } 

      void
      Albums::on_albums_selection_changed ()
      {
        bool albums_has_selection = (m_view_albums->get_selection()->count_selected_rows() > 0);

        if (albums_has_selection)
          m_caps = Caps (m_caps | PlaybackSource::CAN_PLAY);
        else
          m_caps = Caps (m_caps & ~PlaybackSource::CAN_PLAY);

        s_caps_.emit (m_caps);
      }

      void
      Albums::on_tracks_selection_changed ()
      {
        bool albums_has_selection = (m_view_albums->get_selection()->count_selected_rows() > 0);
        bool tracks_has_selection = (m_view_tracks->get_selection()->count_selected_rows() == 1);

        m_actions[TRACKS]->get_action (ACTION_REMOVE_TRACKS)->set_sensitive
                      ((m_view_tracks->get_selection ()->count_selected_rows() > 0));

        if (albums_has_selection && tracks_has_selection)
          m_caps = Caps (m_caps | PlaybackSource::CAN_PLAY);
        else
          m_caps = Caps (m_caps & ~PlaybackSource::CAN_PLAY);

        s_caps_.emit (m_caps);
      }
  }
}
