//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_UI_PART_ALBUMS_HH 
#define BMP_UI_PART_ALBUMS_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <glibmm/markup.h>
#include <gtkmm.h>
#include <libglademm.h>

#include <mcs/mcs.h>

#ifdef HAVE_HAL
#  include "hal.hh"
#endif

#include "amazon.hh"
#include "library.hh"
#include "playback_source.hh"
#include "vfs.hh"
#include "xds.hh"

using namespace Bmp::Library;

namespace Bmp
{
  class MainUI;

  class AlbumView : public Gtk::TreeView
  {
    public:

      enum ViewSize { NORMAL, SMALL };

      enum AlbumFlags
      {
        A_NONE            = 0,
        A_HAS_COVER       = 1 << 0,
        A_HAS_SORTNAME    = 1 << 1
      };

      class Albums : public Gtk::TreeModel::ColumnRecord
      {
        public:

          Gtk::TreeModelColumn<Glib::RefPtr<Gdk::Pixbuf> >  cover;
          Gtk::TreeModelColumn<Glib::ustring>               title;
          Gtk::TreeModelColumn<AlbumFlags>                  flags;
          Gtk::TreeModelColumn<guint64>                     rowuid;
          Gtk::TreeModelColumn<Library::Album>              album;
          Gtk::TreeModelColumn<std::string>                 key;

          Albums ()
          {
            add (cover);
            add (title);
            add (flags);
            add (rowuid);
            add (album);
            add (key);
          }
      };
      Albums albums; 

    private:

      // Data Store Handling
      int
      default_sort_func (Gtk::TreeModel::iterator const& iter_a, Gtk::TreeModel::iterator const& iter_b);

      bool
      row_is_visible (Gtk::TreeModel::iterator const& iter);

      void
      album_title_cell_data_func (Gtk::CellRenderer* cell_, Gtk::TreeModel::iterator const& m_iter);
      void
      album_new_cell_data_func (Gtk::CellRenderer* cell_, Gtk::TreeModel::iterator const& m_iter);
      void
      album_cover_cell_data_func (Gtk::CellRenderer* cell_, Gtk::TreeModel::iterator const& m_iter);


      Gtk::CellRendererPixbuf   *m_cell_album_new;
      Gtk::CellRendererText     *m_cell_album_text;
      Glib::RefPtr<Gdk::Pixbuf>  m_emblem_new;


      Glib::RefPtr<Gtk::ListStore>        m_store_base;
      Glib::RefPtr<Gtk::TreeModelFilter>  m_store_filter;
      Glib::RefPtr<Gdk::Pixbuf>           m_disc_normal;
      Glib::RefPtr<Gdk::Pixbuf>           m_disc_small;
      Glib::ustring                       m_sort;

      // Viewsize
      ViewSize m_viewsize; 

      // Selection Handling
      typedef std::pair < unsigned int, Gtk::TreeModel::RowReference > SelectionPair; 
      typedef std::list < SelectionPair > SelectionList; 

      bool          m_selection_list_modified;
      bool          m_selection_change_block;

      SelectionList m_selection_list_albums,
                    m_selection_list_albums_saved;

      bool record_selection_change (Glib::RefPtr<Gtk::TreeModel> const&, Gtk::TreeModel::Path const&, bool);

      // Misc
      bool    m_show_only_new;
      guint64 m_uid;

      typedef std::map< guint64 , Gtk::TreeModel::RowReference > NewItemsMap;
      NewItemsMap m_new_items_map;

    public:

      void new_remove_one (guint64 uid);
      void new_approve_one (guint64 uid);
      void new_approve_all ();
      void new_drop_all ();

      void save_context  ()
      {
        if (m_selection_list_modified)
        {
          m_selection_list_albums_saved = m_selection_list_albums;
          m_selection_list_modified = false;
        }
      }

      void restore_context ()
      {
        m_sort.clear();
        m_store_filter->refilter ();
        get_selection()->unselect_all ();
        m_selection_list_albums.clear (); 
        for (SelectionList::const_iterator i = m_selection_list_albums_saved.begin (),
                                           e = m_selection_list_albums_saved.end(); i != e; ++i)
        {
          get_selection()->select (i->second.get_path());
        }
      }

      bool find_matching_album (Library::Album const& album, Gtk::TreeModel::iterator & m_iter);
      bool find_matching_album (Library::Album const& album);

      void set_viewsize (ViewSize viewsize);
      void set_only_new_filter (bool new_items);
      void set_sort_filter (Glib::ustring const& sort_text);
      void restore_selection ();

#ifdef HAVE_HAL
      void albums_modify (Bmp::Library::PUDIVRPV const& pairs);
#else
      void albums_modify (Bmp::VFS::UriList const& locations);
#endif

      bool albums_relocate ();
      void albums_remove  (bool force);

      void update ();
      void update_new ();

      void scan_approved_sel ();

      void get_album_list (Library::AlbumList &list, Glib::ustring& label_text, bool& multiple);

      AlbumView (BaseObjectType                        *cobject,
                 const Glib::RefPtr<Gnome::Glade::Xml> &xml);
      ~AlbumView () {}
  };

  class TrackView : public Gtk::TreeView
  {
    private:

      typedef std::map <Glib::ustring, Gtk::TreeModel::RowReference> RowMapTracks;

      void cell_data_func_tracks (Gtk::CellRenderer* cell, Gtk::TreeModel::iterator const& m_iter);
      bool select_func_tracks (Glib::RefPtr<Gtk::TreeModel> const& model, Gtk::TreeModel::Path const& path, bool selected);

      RowMapTracks m_row_map_tracks;
      void notify_remove (Glib::ustring& uri);

      Gtk::CellRenderer *cell_playing,
                        *cell_rating,
                        *cell_new;

      Glib::RefPtr<Gdk::Pixbuf>     rating[6];
      Glib::RefPtr<Gtk::ListStore>  store_tracks;
      Glib::RefPtr<Gdk::Pixbuf>     playing;

      int           m_sort_column_saved;
      Gtk::SortType m_sort_order_saved;

      boost::optional<Gtk::TreeModel::RowReference> playing_row;
      boost::optional<Glib::ustring>                playing_uri;

      // External Callbacks
      void check_update (Glib::ustring const& location, Bmp::Library::TrackUpdated track);

    protected:

      virtual bool on_event (GdkEvent * ev);

    public:

      class Tracks : public Gtk::TreeModel::ColumnRecord
      {
        public:

          Gtk::TreeModelColumn<int>           track;
          Gtk::TreeModelColumn<Glib::ustring> title;
          Gtk::TreeModelColumn<int>           rating;
          Gtk::TreeModelColumn<Glib::ustring> artist;
          Gtk::TreeModelColumn<Glib::ustring> album;
          Gtk::TreeModelColumn<Glib::ustring> genre;
          Gtk::TreeModelColumn<int>           bitrate;
          Gtk::TreeModelColumn<int>           samplerate;
          Gtk::TreeModelColumn<Glib::ustring> location;
          Gtk::TreeModelColumn<bool>          present;
#ifdef HAVE_HAL
          Gtk::TreeModelColumn<Glib::ustring> volume_udi;
          Gtk::TreeModelColumn<Glib::ustring> volume_relative_path;
#endif
          Gtk::TreeModelColumn<bool>          new_item;

          Tracks ()
          {
            add (track);
            add (title);
            add (rating);
            add (artist);
            add (album);
            add (genre);
            add (bitrate);
            add (samplerate);
            add (location);
            add (present);
#ifdef HAVE_HAL
            add (volume_udi);
            add (volume_relative_path);
#endif
            add (new_item);
          }
      };
      Tracks tracks;

      void clear ();
      void unsort ();
      bool remove ();

      bool all_tracks_present ();

      void add_tracks (Bmp::DB::RowV const& vector);
      void get_files  (Util::FileList& list);

#ifdef HAVE_HAL
      void get_tracks (Bmp::Library::PUDIVRPV& pairs);
#else
      void get_tracks (Bmp::VFS::UriList& locations);
#endif

      void save_context  ()
      {
        store_tracks->get_sort_column_id (m_sort_column_saved, m_sort_order_saved);
      }

      void restore_context ()
      {
        store_tracks->clear();
        store_tracks->set_sort_column (m_sort_column_saved, m_sort_order_saved);
      }
  
      // PlaybackSource
      Glib::ustring playback_uri ();
      bool notify_play (unsigned int& position, unsigned int& size);
      bool notify_next (unsigned int& position, unsigned int& size);
      bool notify_prev (unsigned int& position, unsigned int& size);
      void notify_stop ();
      void notify_next_prev (bool &next, bool &prev);

      TrackView (BaseObjectType                        *cobject,
                 Glib::RefPtr<Gnome::Glade::Xml> const& xml);
      ~TrackView () {}
  };

  namespace UIPart
  {
    class Albums : public Bmp::PlaybackSource
    {
      public:

        typedef sigc::signal<void, Bmp::VFS::UriList&> SignalURIEnqueueRequest;
    
        SignalURIEnqueueRequest&
        signal_uri_enqueue_request() { return s_uri_enqueue_request_; }

      private:

        SignalURIEnqueueRequest s_uri_enqueue_request_;

        typedef std::map < Glib::ustring, Gtk::TreeModel::iterator> IterMap;

#ifdef HAVE_HAL
        void hal_volume_removed  (Library::HAL::Volume volume);
        void hal_volume_added    (Library::HAL::Volume volume);
#endif //HAVE_HAL

        void remove_tracks ();
        void enqueue_tracks ();
        void import_tracks (Gtk::Widget *);
        void import_tracks_as_album (Gtk::Widget *);

        void approve_all ();
        void approve_sel ();
        void drop_items  ();

        void albums_remove (bool force = false);
        void albums_relocate ();

        void albums_update (bool update_new = true);
        void albums_modify ();

        void on_sort_entry_changed ();
        void on_show_only_new_toggled ();

        void on_tracks_selection_changed ();
        void on_albums_selection_changed ();
        void on_albums_remove ();

        void check_state_approve_selected ();
        void check_nextprev_caps ();

        void activate_default (Gtk::TreeModel::Path const& path, Gtk::TreeViewColumn* column);
        void mcs_view_size_changed (MCS_CB_DEFAULT_SIGNATURE);

        void track_view_update ();

        Glib::RefPtr<Gnome::Glade::Xml> m_ref_xml;

        Bmp::AlbumView        *m_view_albums;
        Bmp::TrackView        *m_view_tracks;

        Gtk::Entry            *albums_filter;
        Gtk::Notebook         *notebook_albums;
        Gtk::Notebook         *notebook_add;
        Gtk::ProgressBar      *progress_modify;
        Gtk::Label            *label_current_album;
        Gtk::ComboBox         *cb_view_size;

        boost::optional<Glib::ustring> m_saved_context_filter;

        enum ActionGroups { IMPORT, ALBUM, APPROVE, TRACKS, COLLECTIONS, N_ACTION_GROUPS };
        Glib::RefPtr<Gtk::UIManager>    m_ui_manager;
        Glib::RefPtr<Gtk::ActionGroup>  m_actions[N_ACTION_GROUPS];
        Glib::RefPtr<Gtk::ToggleAction> m_action_show_only_new;
        bool popup_menu (GdkEvent *ev);

        bool m_updated;
        bool m_busy;
        bool m_init;

        void send_title ();

        friend class Bmp::MainUI;

      protected:

        virtual Glib::ustring
        get_uri (); 

        virtual bool
        go_next ();

        virtual bool
        go_prev ();

        virtual void
        stop ();

        virtual void
        play ();

        virtual void
        play_requested () {}

        virtual void
        restore_context ();

        virtual GHashTable*
        get_metadata ();

      public:

        void import_external (VFS::UriList& urilist);

        Albums (Glib::RefPtr<Gnome::Glade::Xml> const& xml);
        ~Albums ();

    };
  }
}
#endif
