//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <libglademm.h>

#include <cstring>
#include <iostream>
#include <sstream>
#include <string>

#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>

// Musicbrainz
#include <musicbrainz/musicbrainz.h>

#include "uri++.hh"
#include "util.hh"
#include "util-file.hh"
#include "mbxml.hh"
#include "network.hh"
#include "audio.hh"

#include "debug.hh"
#include "stock.hh"
#include "paths.hh"
#include "main.hh"

#include "x_library.hh"
#include "x_play.hh"
#include "x_vfs.hh"
#include "x_lastfm.hh"
#include "x_mcsbind.hh"

#include "ui-toolbox.hh"
#include "ui-part-lastfm.hh"

#include "ui-banner-image.hh"

namespace Bmp
{
  namespace UIPart
  {
    LASTFM::LASTFM (Glib::RefPtr<Gnome::Glade::Xml> const& xml)
        : PlaybackSource        (PlaybackSource::NONE),
          m_ref_xml             (xml),
          m_streamlist_created (false)
    {
      m_ref_xml->get_widget ("last-fm-queue-size", last_fm_queue_size);
      m_ref_xml->get_widget ("last-fm-submit-enable", last_fm_enable);
      m_ref_xml->get_widget ("last-fm-netstatus", last_fm_netstatus);

      m_ref_xml->get_widget ("lastfm-coverimage", lastfm_coverimage);
      m_ref_xml->get_widget ("lastfm-radio-artist", lastfm_radio_artist);
      m_ref_xml->get_widget ("lastfm-radio-album", lastfm_radio_album);
      m_ref_xml->get_widget ("lastfm-radio-track", lastfm_radio_track);
      m_ref_xml->get_widget ("lastfm-current-station", lastfm_current_station);
      m_ref_xml->get_widget ("lastfm-coverimage", lastfm_coverimage);

      m_ref_xml->get_widget ("lastfm-love", b_love);
      m_ref_xml->get_widget ("lastfm-skip", b_skip);
      m_ref_xml->get_widget ("lastfm-ban",  b_ban);

      m_ref_xml->get_widget ("lastfm-station", e_station);
      m_ref_xml->get_widget ("lastfm-tune",  b_tune);

      m_ref_xml->get_widget ("hbox-lastfm-submit-error", hbox_lastfm_submit_error);
      m_ref_xml->get_widget ("label-lastfm-queue-submit-error", l_queue_submit_error);

      m_ref_xml->get_widget_derived ("lastfm-stations", stations);

      stations->station_activated().connect
          (sigc::mem_fun (this, &LASTFM::set_uri));

      b_tune->signal_clicked().connect
          (sigc::mem_fun (this, &LASTFM::tune));

      b_love->signal_clicked().connect
          (sigc::mem_fun (this, &LASTFM::love));

      b_skip->signal_clicked().connect
          (sigc::mem_fun (this, &LASTFM::skip));

      b_ban->signal_clicked().connect
          (sigc::mem_fun (this, &LASTFM::ban));

      b_love->set_sensitive (false);
      b_skip->set_sensitive (false);
      b_ban->set_sensitive  (false);

      Glib::RefPtr<Gdk::Pixbuf> left, right, slice;
      left  = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "header-logo-left.png"));
      right = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "header-logo-right.png"));
      slice = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "header-logo-slice.png"));

      Glib::RefPtr<Gdk::Pixbuf> p;

      p = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "lastfm-station.png"));
      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("lastfm-i-station"))->set (p);

      p = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "lastfm-artist.png"));
      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("lastfm-i-artist"))->set (p);

      p = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "lastfm-album.png"));
      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("lastfm-i-album"))->set (p);

      p = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "lastfm-track.png"));
      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("lastfm-i-track"))->set (p);

      p = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "audioscrobbler.png"));
      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("image-audioscrobbler"))->set (p);

      Bmp::UI::BannerImage *banner = Gtk::manage (new Bmp::UI::BannerImage (left, right, slice));
      banner->set_size_request (-1, 60);

      Gtk::Alignment *alignment;
      m_ref_xml->get_widget ("alignment-lastfm-logo", alignment);
      alignment->add (*banner);
      alignment->show_all ();

      mcs_bind->bind_entry (dynamic_cast<Gtk::Entry *>(m_ref_xml->get_widget ("last-fm-username")) , "lastfm", "username");
      mcs_bind->bind_entry (dynamic_cast<Gtk::Entry *>(m_ref_xml->get_widget ("last-fm-password")) , "lastfm", "password");

      Gtk::ToggleButton *toggle_button;
      m_ref_xml->get_widget ("last-fm-submit-enable", toggle_button);
      mcs_bind->bind_toggle_button (toggle_button, "lastfm", "submit-enable");

      m_ref_xml->get_widget ("last-fm-queue-enable", toggle_button);
      mcs_bind->bind_toggle_button (toggle_button, "lastfm", "queue-enable");

      m_ref_xml->get_widget ("last-fm-connection-message", lastfm_connection_message);
      m_ref_xml->get_widget ("last-fm-connection-state", lastfm_connection_state);

      lastfm_connection_state->set (Gtk::StockID (BMP_STOCK_APPROVE_PARTIAL), Gtk::ICON_SIZE_SMALL_TOOLBAR);

      mcs->subscribe ("Last.FM", "lastfm", "submit-enable", sigc::mem_fun (this, &LASTFM::mcs_lastfm_submit_enable_changed));
      mcs->subscribe ("Last.FM", "lastfm", "queue-enable",  sigc::mem_fun (this, &LASTFM::mcs_lastfm_queue_enable_changed));

      const char *last_fm_network[] =
      {
        "lastfm-net-idle.png",
        "lastfm-net-rx.png",
        "lastfm-net-tx.png"
      };

      for (unsigned int n = 0; n < 3; n++)
        {
          network[n] = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, last_fm_network[n]));
        }

      lastfm_service->signal_queue_size ().connect
          (sigc::mem_fun (this, &LASTFM::on_lastfm_queue_size));

      lastfm_service->signal_submit_start ().connect
          (sigc::mem_fun (this, &LASTFM::on_lastfm_submit_start));

      lastfm_service->signal_submit_end ().connect
          (sigc::mem_fun (this, &LASTFM::on_lastfm_submit_end));

      lastfm_service->signal_submit_error ().connect
          (sigc::mem_fun (this, &LASTFM::on_lastfm_submit_error));


      last_fm_netstatus->set (network[LAST_FM_IDLE]);

      ::play->signal_http_status().connect (sigc::mem_fun (this, &LASTFM::play_http_status));
      ::play->signal_lastfm_sync().connect (sigc::mem_fun (this, &LASTFM::play_lastfm_sync));

      m_ref_xml->get_widget ("b_lastfm_reconnect", b_lastfm_reconnect);
      b_lastfm_reconnect->signal_clicked().connect
          (sigc::mem_fun (this, &LASTFM::radio_connect));

      m_caps = PlaybackSource::PHONY_NEXT;
      s_caps_.emit (m_caps);

      radio_connect ();
    }

    bool
    LASTFM::lastfm_set_idle_image ()
    {
        last_fm_netstatus->set (network[LAST_FM_IDLE]);
        return false;
    }

    bool
    LASTFM::lastfm_set_tx_image ()
    {
        last_fm_netstatus->set (network[LAST_FM_TX]);
        return false;
    }

    void
    LASTFM::on_lastfm_queue_size (unsigned int queue_size)
    {
      last_fm_queue_size->set_text ((boost::format ("%u") % queue_size).str ());
    }

    void
    LASTFM::on_lastfm_submit_end ()
    {
      last_fm_netstatus->set (network[LAST_FM_RX]);
      Glib::signal_timeout().connect (sigc::mem_fun (this, &LASTFM::lastfm_set_idle_image), 1000);
    }

    void
    LASTFM::on_lastfm_submit_error (Glib::ustring message)
    {
      hbox_lastfm_submit_error->show_all();
      l_queue_submit_error->set_markup ("<small>" + message + "</small>");
      mcs->key_set<bool>("lastfm", "submit-enable", false);
    }

    void
    LASTFM::on_lastfm_submit_start ()
    {
      Glib::signal_timeout().connect (sigc::mem_fun (this, &LASTFM::lastfm_set_tx_image), 1000);
    }

    void
    LASTFM::mcs_lastfm_submit_enable_changed (MCS_CB_DEFAULT_SIGNATURE)
    {
      bool active = boost::get<bool>(value); 

      last_fm_enable->set_active (active);
      last_fm_netstatus->set_sensitive (active);

      if (active)
        {
          hbox_lastfm_submit_error->hide();
          l_queue_submit_error->set_text ("");
        }
    }

    void
    LASTFM::mcs_lastfm_queue_enable_changed (MCS_CB_DEFAULT_SIGNATURE)
    {
      Gtk::Widget *widget = 0;
      bool         active;

      if (Bmp::Network::check_connected ())
      {
        const char* widget_names[] =
        {
            "last-fm-submit-enable",
            "last-fm-netstatus",
        };

        active = mcs->key_get<bool>("lastfm", "queue-enable");
        for (unsigned int n = 0; n < G_N_ELEMENTS(widget_names); n++)
        {
            widget = m_ref_xml->get_widget (widget_names[n]);
            widget->set_sensitive (active);
        }

        last_fm_netstatus->set (network[LAST_FM_IDLE]);
        lastfm_service->emit_queue_size ();
      }
      else
      {
        const char* widget_names[] =
          {
            "last-fm-netstatus"
          };

        active = mcs->key_get<bool>("lastfm", "queue-enable");
        for (unsigned int n = 0; n < G_N_ELEMENTS(widget_names); n++)
          {
            widget = m_ref_xml->get_widget (widget_names[n]);
            widget->set_sensitive (active);
          }

        last_fm_netstatus->set (network[LAST_FM_IDLE]);
        lastfm_service->emit_queue_size ();
      }
    }

#define CATCH_DEFAULT_ERRORS \
        catch (LastFM::TimeOutError& cxe) \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "TimeOutError: %s", cxe.what());                \
            set_error ((boost::format (_("Timeout: %s")) % cxe.what()).str());                         \
            return; \
          } \
        catch (LastFM::ConnectionError& cxe) \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "ConnectionError: %s", cxe.what());             \
            set_error ((boost::format (_("Connection Error: %s")) % cxe.what()).str());                         \
            return; \
          } \
        catch (LastFM::ResourceNotFoundError& cxe) \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "ResourceNotFoundError: %s", cxe.what());         \
            set_error ((boost::format (_("Resource Not Found: %s")) % cxe.what()).str());                         \
            return; \
          } \
        catch (LastFM::AuthenticationError& cxe) \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "AuthenticationError: %s", cxe.what());         \
            set_error ((boost::format (_("Authentication Error: %s")) % cxe.what()).str());                         \
            return; \
          } \
        catch (LastFM::ResponseError& cxe) \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "ResponseError: %s", cxe.what());         \
            set_error ((boost::format (_("Invalid HTTP Response: %s")) % cxe.what()).str());                         \
            return; \
          } \
        catch (LastFM::RequestError& cxe) \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "ResponseError: %s", cxe.what());         \
            set_error ((boost::format (_("Response Error: %s")) % cxe.what()).str());                         \
            return; \
          } \
        catch (LastFM::LastFMNotConnectedError& cxe)  \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "ConnectionError: %s", cxe.what()); \
            set_error ((boost::format (_("Not Handshaked/Session Expired: %s")) % cxe.what()).str()); \
            return; \
          } \
        catch (LastFM::LastFMBadauthError& cxe)  \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "ConnectionError: %s", cxe.what()); \
            set_error ((boost::format (_("Couldn't create Session: %s")) % cxe.what()).str()); \
            return; \
          } \
        catch (LastFM::NetworkError& cxe) \
          { \
            g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "NetworkError: %s", cxe.what());                \
            set_error ((boost::format (_("Unknown Network Error: %s")) % cxe.what()).str());                         \
            return; \
          }

    void
    LASTFM::radio_connect ()
    {
        try {
          lastfm_radio->connect ();
          }
        CATCH_DEFAULT_ERRORS

        clear_error ();
        e_station->set_sensitive (true);
        b_tune->set_sensitive (true);

        if (lastfm_radio->connected())
          {
            lastfm_radio->session (m_session);
            if (!m_streamlist_created)
              {
                stations->set_user (mcs->key_get<std::string>("lastfm","username"), m_session.subscriber);
                m_streamlist_created = true;
              }
            m_caps = Caps (m_caps | PlaybackSource::CAN_PLAY);
            s_caps_.emit (m_caps);
          }
    } 

    // Bmp::PlaybackSource

    LASTFM::~LASTFM ()
    {
    }

    void
    LASTFM::clear_error ()
    {
      lastfm_connection_state->set (Gtk::StockID (BMP_STOCK_APPROVE), Gtk::ICON_SIZE_SMALL_TOOLBAR);
      lastfm_connection_message->set_text (_("OK!"));
      b_lastfm_reconnect->set_sensitive (false);
    }

    void
    LASTFM::set_error (Glib::ustring const& message)
    {
      lastfm_connection_state->set (Gtk::StockID (BMP_STOCK_ERROR), Gtk::ICON_SIZE_SMALL_TOOLBAR);
      lastfm_connection_message->set_text (message);
      b_lastfm_reconnect->set_sensitive (true);
    }

    Glib::ustring
    LASTFM::get_uri ()
    {
      return m_session.streamurl; 
    }

    bool
    LASTFM::go_next ()
    {
      skip ();
      return false;
    }

    bool
    LASTFM::go_prev ()
    {
      return false;
    }

    void
    LASTFM::stop ()
    {
      ::play->property_lastfm_mode() = false;
      play_conn.disconnect();
      m_artist.reset ();
      m_album.reset ();
      m_track.reset ();
      b_love->set_sensitive (false);
      b_skip->set_sensitive (false);
      b_ban->set_sensitive (false);
      lastfm_radio_artist->set_text ("");
      lastfm_radio_album->set_text  ("");
      lastfm_radio_track->set_text  ("");
      lastfm_current_station->set_text ("");
      lastfm_coverimage->clear ();
      m_caps = Caps (m_caps & ~PlaybackSource::CAN_PROVIDE_METADATA);
      m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);
      s_caps_.emit (m_caps);
      m_metadata.streaming = false;
    }

    void
    LASTFM::play ()
    {
      ::play->property_lastfm_mode() = true;
      s_caps_.emit (m_caps);
    }

    void
    LASTFM::restore_context ()
    {
    }

    GHashTable*
    LASTFM::get_metadata ()
    {
      GHashTable *table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
      GValue     *value = 0;

      value = g_new0 (GValue,1);
      g_value_init (value, G_TYPE_STRING);
      g_value_set_string (value, m_metadata.artist.c_str()); 
      g_hash_table_insert (table, g_strdup(MetadatumId(Library::DATUM_ARTIST).c_str()), value);

      value = g_new0 (GValue,1);
      g_value_init (value, G_TYPE_STRING);
      g_value_set_string (value, m_metadata.track.c_str()); 
      g_hash_table_insert (table, g_strdup(MetadatumId(Library::DATUM_TITLE).c_str()), value);

      value = g_new0 (GValue,1);
      g_value_init (value, G_TYPE_STRING);
      g_value_set_string (value, m_metadata.album.c_str()); 
      g_hash_table_insert (table, g_strdup(MetadatumId(Library::DATUM_ALBUM).c_str()), value);

      return table;
    }

    bool
    LASTFM::set_metadata_idler ()
    {
      if (m_timer.elapsed() >= (m_metadata.duration+2))
      {
        lastfm_radio->metadata (m_metadata);
        m_timer.reset();
        send_title_set_metadata ();
      
        if (m_metadata.streaming)
          {
            m_caps = Caps (m_caps | PlaybackSource::CAN_GO_NEXT);
            s_caps_.emit (m_caps);
          }
        else
          {
            m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);
            s_caps_.emit (m_caps);
          }
      }
      return true;
    }

    void
    LASTFM::play_lastfm_sync ()
    {
      lastfm_radio->metadata (m_metadata);
      m_timer.reset ();
      send_title_set_metadata ();
      play_conn.disconnect ();
      play_conn = Glib::signal_timeout().connect (sigc::mem_fun (this, &Bmp::UIPart::LASTFM::set_metadata_idler), 1000);
    }

    void
    LASTFM::play_http_status (int status)
    {
      switch (status)
      {
        case 200: break;
        case 401: 
          {
            lastfm_connection_state->set (Gtk::StockID (BMP_STOCK_ERROR), Gtk::ICON_SIZE_SMALL_TOOLBAR);
            lastfm_connection_message->set_text ("Session expired. Please reconnect.");
            stop ();
            ::play->property_status() = PLAYSTATUS_STOPPED;
            b_lastfm_reconnect->set_sensitive (true);
            break;
          }

        case 503:
          {
            lastfm_connection_state->set (Gtk::StockID (BMP_STOCK_ERROR), Gtk::ICON_SIZE_SMALL_TOOLBAR);
            lastfm_connection_message->set_text (_("The radio server is too busy at the moment, please try again in a few minutes."));
            stop ();
            ::play->property_status() = PLAYSTATUS_STOPPED;
            b_lastfm_reconnect->set_sensitive (true);
            break;
          }
  
        case 666:
          {
            lastfm_connection_state->set (Gtk::StockID (BMP_STOCK_ERROR), Gtk::ICON_SIZE_SMALL_TOOLBAR);
            lastfm_connection_message->set_text (_("Radio server is down for maintenance, please try again in a few minutes."));
            stop ();
            ::play->property_status() = PLAYSTATUS_STOPPED;
            b_lastfm_reconnect->set_sensitive (true);
            break;
          }

        case 667: 
          {
            lastfm_connection_state->set (Gtk::StockID (BMP_STOCK_ERROR), Gtk::ICON_SIZE_SMALL_TOOLBAR);
            lastfm_connection_message->set_text (_("Not enough content left to play this station; please tune in to a different one."));
            stop ();
            ::play->property_status() = PLAYSTATUS_STOPPED;
            ::play->property_lastfm_mode() = false;
            play_conn.disconnect();
            m_artist.reset ();
            m_album.reset ();
            m_track.reset ();
            b_love->set_sensitive (false);
            b_skip->set_sensitive (false);
            b_ban->set_sensitive (false);
            lastfm_radio_artist->set_text ("");
            lastfm_radio_album->set_text  ("");
            lastfm_radio_track->set_text  ("");
            lastfm_coverimage->clear ();
            e_station->select_region (0, e_station->get_text_length());
            e_station->grab_focus ();
            m_caps = Caps (m_caps & ~PlaybackSource::CAN_PLAY);
            m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);
            s_caps_.emit (m_caps);
            break;
          }
      }
    }

    void
    LASTFM::tune ()
    {
      try {
        m_caps = Caps (m_caps | PlaybackSource::CAN_PLAY);
        std::string station = e_station->get_text().c_str();
        m_timer.stop();
        lastfm_radio->tune (station);
        clear_error();
        if (!m_metadata.streaming)
          {
              if (!lastfm_radio->connected()) radio_connect ();
              play ();
              s_playback_request_.emit ();
          }
        }
      catch (LastFM::LastFMStreamTuningError& cxe)
        {
          g_log (BMP_LASTFM_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "StreamTuningError: %s", cxe.what());
          set_error (_("Couldn't tune into station"));
          ::play->property_status() = PLAYSTATUS_STOPPED;
        }

      CATCH_DEFAULT_ERRORS

    }

    void
    LASTFM::love ()
    {
      try {
          b_love->set_sensitive (false);
          lastfm_radio->command (Bmp::LastFM::Radio::LOVE);
          clear_error ();
        }

      CATCH_DEFAULT_ERRORS
    }

    void
    LASTFM::skip ()
    {
      try {
          b_skip->set_sensitive (false);
          m_caps = Caps (m_caps & ~PlaybackSource::CAN_GO_NEXT);
          s_caps_.emit (m_caps);
          m_timer.stop ();
          lastfm_radio->command (Bmp::LastFM::Radio::SKIP);
          clear_error ();
        }

      CATCH_DEFAULT_ERRORS

    }

    void
    LASTFM::ban ()
    {
      try {
          b_ban->set_sensitive (false);
          m_timer.stop ();
          lastfm_radio->command (Bmp::LastFM::Radio::BAN);
          clear_error ();
        }

      CATCH_DEFAULT_ERRORS

    }

    ////////////////////////////////////////////  

    void
    LASTFM::send_title_set_metadata ()
    {
      if (!m_metadata.streaming) return;

      if ((m_artist && m_album && m_track) &&
            ((m_artist.get() == m_metadata.artist) &&
            (m_album.get() == m_metadata.album) &&
            (m_track.get() == m_metadata.track))) return;

      b_love->set_sensitive (true);
      b_skip->set_sensitive (true);
      b_ban->set_sensitive (true);

      m_caps = Caps (m_caps | PlaybackSource::CAN_PROVIDE_METADATA);
      m_caps = Caps (m_caps | PlaybackSource::CAN_GO_NEXT);
      s_caps_.emit (m_caps);

      lastfm_radio_artist->set_text (m_metadata.artist);
      lastfm_radio_album->set_text  (m_metadata.album);
      lastfm_radio_track->set_text  (m_metadata.track);
      lastfm_current_station->set_text (m_metadata.station);

      m_artist  = m_metadata.artist;
      m_album   = m_metadata.album;
      m_track   = m_metadata.track;

      Bmp::VFS::Handle handle (m_metadata.albumcover_large);
      vfs->read (handle, VFS::TRANSPORT);
      Glib::RefPtr<Gdk::PixbufLoader> loader = Gdk::PixbufLoader::create(); 
      loader->write (reinterpret_cast<const guint8*>(handle.get_buffer()), handle.get_buffer_size());
      loader->close ();
      Glib::RefPtr<Gdk::Pixbuf> p = loader->get_pixbuf();

      if (p)
        {
          lastfm_coverimage->set (p->scale_simple ( 160, 160, Gdk::INTERP_HYPER ));
        }
      else
        {
          lastfm_coverimage->clear();
        }

      Bmp::SimpleTrackInfo sti;

      sti.artist = m_metadata.artist;
      sti.album = m_metadata.album;
      sti.title = m_metadata.track;
      sti.duration = m_metadata.duration;
      if (p) sti.image = p->scale_simple ( 128, 128, Gdk::INTERP_HYPER );

      Glib::ustring text ("<b>");
      text.append(Glib::Markup::escape_text(m_metadata.track)).append("</b> (");
      text.append(Glib::Markup::escape_text(m_metadata.artist)).append(": <i>");
      text.append(Glib::Markup::escape_text(m_metadata.album)).append("</i>)");
      s_track_info_.emit (text, sti); 
      loader.clear();
    }

    void
    LASTFM::set_uri (Glib::ustring const& uri)
    {
      using namespace Glib;

      URI u (uri);
      u.unescape();
      e_station->set_text (ustring(u));
      tune ();
    }

    LASTFM::TreeViewStations::TreeViewStations (BaseObjectType                       *cobject,
                                                const Glib::RefPtr<Gnome::Glade::Xml> &xml)
                                            : Gtk::TreeView (cobject)
    {
      station = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_LASTFM, "lastfm-station.png"));

      store = Gtk::ListStore::create (columns);
      Gtk::TreeViewColumn *column = 0;
      Gtk::CellRendererPixbuf *cell_pixbuf = 0;
      Gtk::CellRendererText *cell_text = 0;

      cell_pixbuf = Gtk::manage (new Gtk::CellRendererPixbuf ());
      cell_pixbuf->property_ypad () = 2;
      cell_pixbuf->property_xpad () = 2;
      append_column ("", *cell_pixbuf);
      column = get_column (0);
      column->set_sizing (Gtk::TREE_VIEW_COLUMN_FIXED);
      column->set_resizable (false);
      column->set_expand (false);
      column->set_fixed_width (26);
      column->set_cell_data_func (*cell_pixbuf, sigc::mem_fun (this, &Bmp::UIPart::LASTFM::TreeViewStations::cell_data_func));


      cell_text = Gtk::manage (new Gtk::CellRendererText ());
      cell_text->property_ypad () = 2;
      append_column ("", *cell_text); 
      column = get_column (1);
      column->add_attribute (*cell_text, "markup", 1);
      column->set_resizable (false);
      column->set_expand (false);

      get_selection()->set_mode (Gtk::SELECTION_SINGLE);  
      set_model (store);
    }

    LASTFM::TreeViewStations::~TreeViewStations () {}

    void
    LASTFM::TreeViewStations::set_user (Glib::ustring const& username, bool subscriber)
    {
      using boost::algorithm::split;
      using boost::algorithm::is_any_of;
      using namespace std;

      static boost::format neighbours ("lastfm://user/%s/neighbours");      
      static boost::format personal   ("lastfm://user/%s/personal");      
      static boost::format loved      ("lastfm://user/%s/loved");      
      static boost::format globaltag  ("lastfm://globaltags/%s");      

      Gtk::TreeModel::iterator m_iter; 
      store->clear ();

      if (subscriber)
      {
        m_iter = store->append ();
        (*m_iter)[columns.name] = _("personal radio");
        (*m_iter)[columns.station]  = (personal % username.c_str()).str();

        m_iter = store->append ();
        (*m_iter)[columns.name] = _("loved tracks radio");
        (*m_iter)[columns.station]  = (loved % username.c_str()).str();
      }

      m_iter = store->append ();
      (*m_iter)[columns.name] = _("neighbour radio");
      (*m_iter)[columns.station]  = (neighbours % username.c_str()).str();

      try {
          LastFM::Tags tags;
          LastFM::get_tags (username, LastFM::TAGS_TOPTAGS, tags); 
          for (LastFM::Tags::const_iterator t = tags.begin(), e = tags.end(); t != e; ++t)
          {
            LastFM::Tag const& tag = (*t);
            m_iter = store->append ();
            (*m_iter)[columns.name] = Glib::Markup::escape_text (tag.name);
            (*m_iter)[columns.station] = (globaltag % tag.name).str();
          }
        }
      catch  (...) {}
    }

    void
    LASTFM::TreeViewStations::cell_data_func (Gtk::CellRenderer* cell, Gtk::TreeModel::iterator const& m_iter)
    {
      dynamic_cast<Gtk::CellRendererPixbuf *>(cell)->property_pixbuf() = station;
    }

    void
    LASTFM::TreeViewStations::on_row_activated (Gtk::TreeModel::Path const& path, Gtk::TreeViewColumn * column)
    {
      Gtk::TreeModel::iterator m_iter = store->get_iter (path);
      station_activated_.emit ((*m_iter)[columns.station]); 
    }

  } // !UIPart::
} // !Bmp:: 
