//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_PLAYLIST_HH
#define BMP_PLAYLIST_HH

#include <gtkmm.h>
#include <libglademm.h>
#include <boost/shared_ptr.hpp>

#include "playback_source.hh"
#include "dialog-export.hh"
#include "vfs.hh"
#include "flow.hh"

namespace Bmp
{
  namespace UIPart
  {
    class Playlist : public Bmp::PlaybackSource 
    {
      public:

        Playlist (Glib::RefPtr<Gnome::Glade::Xml> const& xml);
        ~Playlist ();

        void add_uris (VFS::UriList& list, bool playback);

      protected:
       
        virtual Glib::ustring
        get_uri ();

        virtual bool
        go_next ();

        virtual bool
        go_prev ();

        virtual void
        stop ();
  
        virtual void
        play ();

        virtual void
        play_requested ();

        virtual void
        restore_context ();

        virtual GHashTable*
        get_metadata ();

     private:

        typedef std::deque<RGUID> Queue; 
        Queue history;
        boost::optional<Queue::size_type> mark;

        Bmp::ExportDialog *export_dialog;

        Glib::RefPtr<Gtk::UIManager>   m_ui_manager;
        Glib::RefPtr<Gtk::ActionGroup> m_actions;

        bool popup_menu (GdkEvent *ev);
        void activate_default (Gtk::TreeModel::Path const& path, Gtk::TreeViewColumn* column);

        class PlaylistFCDialog : public Gtk::FileChooserDialog
        {
          public:
            PlaylistFCDialog ();    
            virtual ~PlaylistFCDialog ();

          protected:
            virtual void on_hide (); 

          private:
            bool audio_files_filter (Gtk::FileFilter::Info const& info);
        };

        class Track: public Gtk::TreeModel::ColumnRecord
        {
          public:

            Gtk::TreeModelColumn<Glib::ustring> artist;
            Gtk::TreeModelColumn<Glib::ustring> album;
            Gtk::TreeModelColumn<unsigned int>  tracknumber;
            Gtk::TreeModelColumn<Glib::ustring> title;
            Gtk::TreeModelColumn<Glib::ustring> genre;
            Gtk::TreeModelColumn<unsigned int>  date;
            Gtk::TreeModelColumn<unsigned int>  duration;
            Gtk::TreeModelColumn<Glib::ustring> location;

            Track ()
            {
              add (artist);
              add (album);
              add (tracknumber);
              add (title);
              add (genre);
              add (date);
              add (duration);
              add (location);
            }
        };
        Track track;

        void cell_data_func_playing (Gtk::CellRenderer* cell, Gtk::TreeModel::iterator const& iter);
        void cell_data_func_numbers (Gtk::CellRenderer* cell, Gtk::TreeModel::iterator const& iter);
        void selection_changed ();
        void send_title ();
  
        void append_track (Glib::ustring const& uri);
        void add (bool play);
        void clear ();
        void clear_selected ();
        void export_playlist ();

        Gtk::TreeView                    *playlist_view;
        Glib::RefPtr<Gtk::ListStore>      playlist_store;
        Glib::RefPtr<Gnome::Glade::Xml>   m_ref_xml;

        Gtk::CellRendererPixbuf          *cell_playing;
        Glib::RefPtr<Gdk::Pixbuf>         m_playing;

        boost::optional<Gtk::TreeModel::RowReference> m_playing_row;

    }; // !Playlist
  } // !UIPart::
} // !Bmp::

#endif // !BMP_PLAYLIST_HH

